using dg1d
import Base: Filesystem
using CairoMakie
CairoMakie.activate!()


function readdata(dir)
  outputfile = joinpath(dir, "output.h5")
  us = Vector{Float64}[]
  ts = Float64[]
  x = h5open(outputfile,"r") do file
    groups = keys(file)
    filter!(g->g!="mesh", groups)
    groups = parse.(Int64, groups)
    sort!(groups)
    for g in groups
      push!(us, vec(read(file[string(g)]["dynamic_variables.u"])))
      push!(ts, read(file[string(g)]["t"]))
    end
    x = read(file["mesh"]["mesh"])
  end

  Npts, K = size(x)

  x = vec(x)

  return Npts, K, x, ts, us
end


function analyze(dir)

  tomlfile = get_parameter_file(dir)
  parameters = FlatDict(TOML.parsefile(joinpath(dir, tomlfile)))

  mesh = make_Mesh(trim_headers(parameters, "Mesh"))

  equation = ScalarEquation.make_Equation(trim_headers(parameters, "Equation"))
  @unpack v = equation

  id_parameters = trim_headers(parameters, "Initialdata", "Mesh")
  id_fn = ScalarEquation.make_function_initialdata_equation(id_parameters)
  analytic_u(x, t) = id_fn(periodize(x - v * t, mesh.L))

  # read data
  Npts, K, x, ts, us = readdata(dir)

  # select samples to plot
  nsamples = 4
  nstep = 25
  ntotal = length(ts)
  rng = collect(1:nstep:nsamples*nstep)
  plot_ts = ts[rng]
  plot_us = us[rng]
  plot_ana_us = [ analytic_u.(x, t) for t in plot_ts ]
  
  # prepare plot
  fig = Figure(resolution=(1600,1200))
  colors = to_colormap(:rainbow, length(rng))
  # dim 'older' results
  transparency(idx, n) = exp((idx-n)*2/n)

  # plot
  subfig1 = fig[2,1]
  ax1 = Axis(subfig1, ylabel="u", xlabel="x")
  subfig2 = fig[2,2]
  ax2 = Axis(subfig2, ylabel="|u - u_ana|", xlabel="x")
  handles, labels = [], String[]
  x = reshape(x, (Npts, K))
  diffu = zeros(Npts, K)
  for (i, (u, ana_u, t)) in enumerate(zip(plot_us, plot_ana_us, plot_ts))
    u = reshape(u, (Npts, K))
    ana_u = reshape(ana_u, (Npts, K))
    # diffu .= abs.(u - ana_u)
    transp = transparency(i, nsamples+1)
    for k = 1:K
      sa = scatterlines!(subfig1, x[:,k], u[:,k], markersize=3, color=(colors[i],transp))
      scatterlines!(subfig2, x[:,k], ana_u[:,k], markersize=3, color=(colors[i],transp))
      if k == K
        push!(handles, sa)
        push!(labels, "t=$t")
      end
    end
  end

  Legend(fig[1,1:2], handles, labels, orientation=:horizontal, tellwidth=true)

  save(joinpath(dir, "evolution.png"), fig)

  return
end


#######################################################################
#                                Main                                 #
#######################################################################


projectdir = Filesystem.abspath(Filesystem.dirname(@__FILE__))
dg1d.analyze_projects(analyze, projectdir)
