export TOV
module TOV


using dg1d
using LaTeXStrings
using Plots
using Printf


# array indices of state vector
const idx_p, idx_m, idx_ϕ, idx_R, idx_r = 1, 2, 3, 4, 5


function TOV_rhs!(du, u, r, eos::Polytrope)

  if r == 0.0
    du .= 0.0
  else
    p, m = u[idx_p], u[idx_m]

    ϵ = try
      rho = eos(Density, Pressure, p)
      eps = eos(InternalEnergy, Density, rho)
      eos(TotalEnergy, Density, rho, InternalEnergy, eps)
    catch err
      if err isa DomainError
        0.0
      else
        rethrow(e)
      end
    end

    # dp/dr
    du[idx_p] = - (ϵ + p) * (m + 4.0 * π * r^3 * p) / (r * (r - 2.0 * m))
    # dm/dr
    du[idx_m] = 4.0 * π * r^2 * ϵ
    # dϕ/dr
    du[idx_ϕ] = - du[idx_p] / (ϵ + p)
    # dR/dr
    du[idx_R] = (1.0 - sqrt(1.0 - 2.0 * m / r)) / sqrt(r^2 - 2.0 * m * r)
  end

  return
end


function solve()

  # equation of state
  Γ = 2
  K = 100
  # ρc = 8.00 * 1e-3 # central rest mass density
  ρc = 1.28 * 1e-3 # central rest mass density
  eos = Polytrope(K = K, Gamma = Γ)
  pc = eos(Pressure, Density, ρc, InternalEnergy, nothing)

  # initial values
  # Note: We put ϕ(0) = 0 as initial condition for the potential in the interior
  # However, because we also fix ϕ(r → ∞) = 0 in the exterior region where
  # the schwarzschild solution should hold, the potentials will not be continues
  # at the star's surface. But because dϕ/dr does not on ϕ, the solution is
  # determined up to an additative constant, e.g. ϕ_1 = ϕ + c is also a valid
  # solution. This trick allows us to integrate ϕ together with the TOV equations
  # simulatenously and later adjust the potential in the interior to match
  # the Schwarzschild solution.
  # The same reasoning is also explained in Baumgarte, Shaprio book and
  # in Bugner's thesis.
  u0 = zeros(4);
  u0[idx_p] = pc
  u0[idx_m] = 0.0
  u0[idx_ϕ] = 0.0
  u0[idx_R] = 0.0


  dr = 1e-2 # inital step width for r
  r_range = [0.0, 30.0] # max integration range
  ODE!(du, u, r) = TOV_rhs!(du, u, r, eos)


  # post hook setup to 
  #   - store solution
  #   - terminate ODE solver if pressure drops below p_threshold
  # solution = [u0..., r_range[1]]
  storage_increment = 1000
  n_variables = 5
  solution = zeros(Float64, storage_increment, n_variables)
  solution[1,:] .= [ u0..., r_range[1] ]
  counter = 1
  p_threshold = 1e-14
  function post_hook(u, r)
    counter += 1
    if counter >= size(solution)[1]
      zs = zeros(Float64, storage_increment, n_variables)
      solution = vcat(solution, zs)
    end
    solution[counter, :] .= [ u..., r ]
    p = u[idx_p]
    return (p > p_threshold) # continue as long as this is true
  end

  # solve ODE
  pblm = dg1d.ODEProblem(:rk4)
  dg1d.evolve_adaptive_ERK(pblm, ODE!, u0, dr, r_range[2], post_hook=post_hook)

  # chop off trailing zeros
  solution = solution[1:counter,:]

  M      = solution[end, idx_m]
  r_surf = solution[end, idx_r]
  R_surf = solution[end, idx_R]
  @info "M = $M"
  @info "r = $r_surf"
  @info "R = $R_surf"

  # potential at star's surface
  Δϕ_surf = solution[end,idx_ϕ] - (0.5 * log(1.0 - 2.0 * M / r_surf))
  # adjust interior potential to align with exterior at star's surface
  solution[:,idx_ϕ] .-= Δϕ_surf

  # fix integration constant between isotropic and schwarzschild coordinate
  Rs = R_surf
  rs = r_surf
  r = solution[:,idx_r]
  R = solution[:,idx_R]
  @. solution[:,idx_R] = 0.5 * r / rs * (sqrt(rs^2 - 2 * M * rs) + rs - M) * exp(R - Rs)

  # save
  output_dir = dg1d.get_output_dir()
  filename = joinpath(output_dir, "TOV.h5")
  h5open(filename, "w") do file
    @info "Saving results to '$filename' ..."
    g = create_group(file, "TOV")
    attributes(g)["Γ"]  = Γ
    attributes(g)["K"]  = K
    attributes(g)["ρc"] = ρc
    attributes(g)["M"]  = M
    g["ϕ"] = solution[:,idx_ϕ]
    g["p"] = solution[:,idx_p]
    g["m"] = solution[:,idx_m]
    g["R"] = solution[:,idx_R]
    g["r"] = solution[:,idx_r]
  end

  return
end



function read_data(filename)
  return h5open(filename, "r") do file
    g = file["TOV"]
    Γ = read_attribute(g, "Γ")
    K = read_attribute(g, "K")
    ρc = read_attribute(g, "ρc")
    M = read_attribute(g, "M")
    ϕ = read(g, "ϕ")
    p = read(g, "p")
    m = read(g, "m")
    R = read(g, "R")
    r = read(g, "r")
    (; Γ, K, ρc, M, ϕ, p, m, R, r)
  end
end


function plot_potentials()

  output_dir = dg1d.get_output_dir()
  filename = joinpath(output_dir, "TOV.h5")
  Γ, K, ρc, M, ϕ, p, m, R, r = read_data(filename)

  eos = Polytrope(K = K, Gamma = Γ)
  ρ   = @. eos(Density, InternalEnergy, nothing, Pressure, p)
  ϵ   = @. eos(InternalEnergy, Density, ρ, Pressure, p)

  r_ext = collect(range(r[end], 15.0, length=100))
  M = m[end]
  ϕ_ext = @. 0.5 * log(1.0 - 2.0 * M / r_ext)

  ϕ_int = ϕ
  Λ_int = @. - 0.5 * log(1.0 - 2.0 * m / r)
  Λ_ext = - ϕ_ext

  Plots.gr()
  plt_metric = Plots.plot()
  # Plots.plot(r, 100.0 * ρ, label="100 * ρ(r)")
  # Plots.plot!(plt, r, 100.0 * ϵ, label="100 * ϵ(r)")
  # Plots.plot!(plt, r, 100.0 * p, label="100 * p(r)")
  # Plots.plot!(plt, r, ϕ_int, label="ϕ_int(r)")
  # Plots.plot!(plt, r_ext, ϕ_ext, label="ϕ_ext(r)")
  # Plots.plot!(plt, r, Λ_int, label="Λ_int(r)")
  # Plots.plot!(plt, r_ext, Λ_ext, label="Λ_ext(r)")
  # Plots.plot!(plt, r, @.(r/R), label="R")
  # R_ext = @. 1/2 * (sqrt(r_ext^2 - 2 * M * r_ext) + r_ext - M)
  # Plots.plot!(plt, r_ext, R_ext, label="R")
  # Plots.plot!(plt, r, m, label="m")

  #### potentials are continuous and differential at surface
  Plots.plot!(plt_metric, r, @.( exp(2.0*ϕ_int) ), label=L"e^{2 \Phi_{int}}")
  Plots.plot!(plt_metric, r, @.( exp(-2.0*Λ_int) ), label=L"e^{2 \Lambda_{int}}")
  Plots.plot!(plt_metric, r_ext, @.( exp(2.0*ϕ_ext) ), label=L"e^{2 \Phi_{ext}}")
  Plots.plot!(plt_metric, r_ext, @.( exp(-2.0*Λ_ext) ), label=L"e^{2 \Lambda_{ext}}")
  rs = r[end]
  Rs = R[end]
  Plots.plot!(plt_metric, [rs, rs], [0,1], line=(:dash, :black), label=L"r=r_S")

  plt_eos = Plots.plot()
  Plots.plot!(plt_eos, r, p, label=L"p")
  Plots.plot!(plt_eos, r, ρ, label=L"\rho")
  # Plots.plot!(plt_eos, r, ϵ, label="ϵ")

  plt = Plots.plot(plt_metric, plt_eos, layout=(2,1),
                   xlabel="r",
                   plot_title="TOV star, $(@sprintf "ρc = %.5f, M = %.5f, R = %.5f" ρc M Rs)",
                   legendfontsize = 20, guidefontsize = 20,
                   tickfontsize = 20, titlefontsize = 20,
                   size=(1200,1200), xlims=(0.0, 15.0))
  Plots.gui(plt)


end


end # module
