#######################################################################
#                            Flux methods                             #
#######################################################################



@with_signature [legacy=false] function flux(equation::EulerEquation)
  @accepts rho, q, E
  @unpack eos = equation
  eps = E / rho - q^2 / (2 * rho^2)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho = q
  flx_q   = q^2 / rho + p
  flx_E   = q / rho * (E + p)
  @returns flx_rho, flx_q, flx_E
end


@with_signature [legacy=false] function maxspeed(equation::EulerEquation)
  @accepts rho, q, E
  @unpack eos = equation
  eps = E / rho - q^2 / (2 * rho^2)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


@with_signature [legacy=false] function pressure(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, q, E
  @unpack eos = equation
  eps = E / rho - q^2 / (2 * rho^2)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end

@with_signature [legacy=false] function entropy_variables(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, q, p
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  flx_S = q / rho * S
  @returns S, flx_S
end


@with_signature [legacy=false] function llf(equation::EulerEquation)
  @accepts        rho, flx_rho, q, flx_q, E, flx_E, v
  @accepts [bdry] bdry_rho, bdry_q, bdry_E, bdry_v
  @accepts [bdry] nx

  vmax = absmax(v, bdry_v)

  @unpack eos = equation
  bdry_eps = bdry_E / bdry_rho - bdry_q^2 / (2 * bdry_rho^2)
  bdry_p = eos(Pressure, Density, bdry_rho, InternalEnergy, bdry_eps)
  bdry_flx_rho = bdry_q
  bdry_flx_q   = bdry_q^2 / bdry_rho + bdry_p
  bdry_flx_E   = bdry_q / bdry_rho * (bdry_E + bdry_p)

  nflx      = nx*flx_rho
  bdry_nflx = nx*bdry_flx_rho
  nflx_rho  = LLF(nflx, bdry_nflx, rho, bdry_rho, vmax)

  nflx      = nx*flx_q
  bdry_nflx = nx*bdry_flx_q
  nflx_q    = LLF(nflx, bdry_nflx, q, bdry_q, vmax)

  nflx      = nx*flx_E
  bdry_nflx = nx*bdry_flx_E
  nflx_E    = LLF(nflx, bdry_nflx, E, bdry_E, vmax)

  @returns [bdry] nflx_rho, nflx_q, nflx_E
end


@with_signature [legacy=false] function bdryllf(equation::EulerEquation)
  @accepts        rho, flx_rho, q, flx_q, E, flx_E, v
  @accepts        init_rho, init_q, init_E, init_v
  @accepts [bdry] nx

  # enforce dirichlet/neumann conditions following stability analysis
  # see dg-notes/tex/evm.pdf
  # Hesthaven, Numerical Methods for Conservation Laws uses same logic

  # Dirichlet conditions
  bdry_rho = -rho + 2*init_rho
  bdry_q   = -q   + 2*init_q
  bdry_E   = -E   + 2*init_E

  # Neumann conditions
  # bdry_rho = rho
  # bdry_q   = q
  # bdry_E   = E

  @unpack eos  = equation
  bdry_eps     = bdry_E / bdry_rho - bdry_q^2 / (2 * bdry_rho^2)
  bdry_p       = eos(Pressure, Density, bdry_rho, InternalEnergy, bdry_eps)
  bdry_flx_rho = bdry_q
  bdry_flx_q   = bdry_q^2 / bdry_rho + bdry_p
  bdry_flx_E   = bdry_q / bdry_rho * (bdry_E + bdry_p)

  vmax = absmax(v, init_v)

  nflx      = nx*flx_rho
  bdry_nflx = nx*bdry_flx_rho
  nflx_rho  = LLF(nflx, bdry_nflx, rho, bdry_rho, vmax)

  nflx      = nx*flx_q
  bdry_nflx = nx*bdry_flx_q
  nflx_q    = LLF(nflx, bdry_nflx, q, bdry_q, vmax)

  nflx      = nx*flx_E
  bdry_nflx = nx*bdry_flx_E
  nflx_E    = LLF(nflx, bdry_nflx, E, bdry_E, vmax)

  @returns [bdry] nflx_rho, nflx_q, nflx_E
end


@with_signature [legacy=false] function av_flux(eq::EulerEquation)
  @accepts flx_rho, ldg_rho, flx_q, ldg_q, flx_E, ldg_E, smoothed_mu

  flx_rho += smoothed_mu * ldg_rho
  flx_q   += smoothed_mu * ldg_q
  flx_E   += smoothed_mu * ldg_E

  @returns flx_rho, flx_q, flx_E
end
@with_signature [legacy=false] function av_nflux(eq::EulerEquation)
  @accepts        ldg_rho, ldg_q, ldg_E, smoothed_mu
  @accepts [bdry] bdry_ldg_rho, bdry_ldg_q, bdry_ldg_E, bdry_smoothed_mu
  @accepts [bdry] nflx_rho, nflx_q, nflx_E, nx

  nflx       = nx*ldg_rho
  bdry_nflx  = nx*bdry_ldg_rho
  nflx_rho  += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  nflx       = nx*ldg_q
  bdry_nflx  = nx*bdry_ldg_q
  nflx_q    += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  nflx       = nx*ldg_E
  bdry_nflx  = nx*bdry_ldg_E
  nflx_E    += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  @returns [bdry] nflx_rho, nflx_q, nflx_E
end


@with_signature [legacy=false] function ldg_nflux(eq::EulerEquation)
  @accepts        rho, q, E
  @accepts [bdry] bdry_rho, bdry_q, bdry_E
  @accepts [bdry] nx

  nflx_rho = 0.5 * nx * (rho + bdry_rho)
  nflx_q   = 0.5 * nx * (q + bdry_q)
  nflx_E   = 0.5 * nx * (E + bdry_E)

  @returns [bdry] nflx_rho, nflx_q, nflx_E
end
@with_signature [legacy=false] function ldg_bdryllf(eq::EulerEquation)
  @accepts        rho, q, E
  @accepts [bdry] nx

  # set exterior state to 0 according to stability analysis
  # see dg-notes/tex/evm.pdf for reasoning
  nflx_rho = nx * rho
  nflx_q   = nx * q
  nflx_E   = nx * E

  @returns [bdry] nflx_rho, nflx_q, nflx_E
end


#######################################################################
#                                 2D                                  #
#######################################################################


@with_signature [legacy=false] function maxspeed(equation::EulerEquation2d)
  # Keeping this in here, because I am not sure what is causing the big mismatches
  # in ref tests when switching between these two versions.
  # @accepts rho, qx, qy, p
  # @unpack eos = equation
  # # we don't recompute the pressure, because an approximate maximum speed is good enough
  # q = sqrt(qx^2+qy^2)
  # v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @accepts rho, qx, qy, E
  @unpack eos = equation
  q2 = qx^2 + qy^2
  eps = E / rho - q2 / (2 * rho^2) # internal energy
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  q = sqrt(q2)
  v = abs(q / p) + sqrt(eos.gamma * p / rho)
  @returns v
end


@with_signature [legacy=false] function entropy_variables(equation::EulerEquation2d{EquationOfState.IdealGas})
  @accepts rho, qx, qy, p
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  flx_S_x = qx / rho * S
  flx_S_y = qy / rho * S
  @returns S, flx_S_x, flx_S_y
end


@with_signature [legacy=false] function pressure(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  @unpack eos = equation
  q2 = qx^2 + qy^2
  eps = E / rho - q2 / (2 * rho^2) # internal energy
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature [legacy=false] function flux(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  @unpack eos = equation
  q2 = qx^2 + qy^2
  eps = E / rho - q2 / (2 * rho^2) # internal energy
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho_x = qx
  flx_rho_y = qy
  flx_qx_x = qx^2/rho + p
  flx_qx_y = qx*qy / rho
  flx_qy_y = qy^2/rho + p
  flx_E_x  = qx / rho * (E + p)
  flx_E_y  = qy / rho * (E + p)
  @returns flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y, flx_qy_y, flx_E_x, flx_E_y
end


@with_signature [legacy=false] function llf(equation::EulerEquation2d)
  @accepts        rho, flx_rho_x, flx_rho_y,
                  qx, flx_qx_x, flx_qx_y,
                  qy, #=flx_qy_x,=# flx_qy_y,
                  E, flx_E_x, flx_E_y, v
  @accepts [bdry] bdry_rho, bdry_qx, bdry_qy, bdry_E, bdry_v
  @accepts [bdry] nx, ny
  @unpack eos = equation

  vmax = absmax(v, bdry_v)

  bdry_q2 = bdry_qx^2 + bdry_qy^2
  bdry_eps = bdry_E / bdry_rho - bdry_q2 / (2 * bdry_rho^2) # internal energy
  bdry_p = eos(Pressure, Density, bdry_rho, InternalEnergy, bdry_eps)

  bdry_flx_rho_x = bdry_qx
  bdry_flx_rho_y = bdry_qy
  bdry_flx_qx_x  = bdry_qx^2/bdry_rho + bdry_p
  bdry_flx_qx_y  = bdry_qx*bdry_qy / bdry_rho
  bdry_flx_qy_y  = bdry_qy^2/bdry_rho + bdry_p
  bdry_flx_E_x   = bdry_qx / bdry_rho * (bdry_E + bdry_p)
  bdry_flx_E_y   = bdry_qy / bdry_rho * (bdry_E + bdry_p)

  nflx          = nx*flx_rho_x + ny*flx_rho_y
  bdry_nflx     = nx*bdry_flx_rho_x + ny*bdry_flx_rho_y
  nflx_rho      = LLF(nflx, bdry_nflx, rho, bdry_rho, vmax)

  nflx          = nx*flx_qx_x + ny*flx_qx_y
  bdry_nflx     = nx*bdry_flx_qx_x + ny*bdry_flx_qx_y
  nflx_qx       = LLF(nflx, bdry_nflx, qx, bdry_qx, vmax)

  flx_qy_x      = flx_qx_y
  bdry_flx_qy_x = bdry_flx_qx_y
  nflx          = nx*flx_qy_x + ny*flx_qy_y
  bdry_nflx     = nx*bdry_flx_qy_x + ny*bdry_flx_qy_y
  nflx_qy       = LLF(nflx, bdry_nflx, qy, bdry_qy, vmax)

  nflx          = nx*flx_E_x + ny*flx_E_y
  bdry_nflx     = nx*bdry_flx_E_x + ny*bdry_flx_E_y
  nflx_E        = LLF(nflx, bdry_nflx, E, bdry_E, vmax)

  @returns [bdry] nflx_rho, nflx_qx, nflx_qy, nflx_E
end


@with_signature [legacy=false] function ldg_nflux(eq::EulerEquation2d)
  @accepts        rho, qx, qy, E
  @accepts [bdry] bdry_rho, bdry_qx, bdry_qy, bdry_E
  @accepts [bdry] nx, ny

  avg_rho = 0.5 * (rho + bdry_rho)
  avg_qx  = 0.5 * (qx + bdry_qx)
  avg_qy  = 0.5 * (qy + bdry_qy)
  avg_E   = 0.5 * (E + bdry_E)

  nflx_rho_x = nx * avg_rho
  nflx_rho_y = ny * avg_rho

  nflx_qx_x  = nx * avg_qx
  nflx_qx_y  = ny * avg_qx

  nflx_qy_x  = nx * avg_qy
  nflx_qy_y  = ny * avg_qy

  nflx_E_x   = nx * avg_E
  nflx_E_y   = ny * avg_E

  @returns [bdry] nflx_rho_x, nflx_rho_y, nflx_qx_x, nflx_qx_y,
                  nflx_qy_x, nflx_qy_y, nflx_E_x, nflx_E_y
end


@with_signature [legacy=false] function av_flux_2d(eq::EulerEquation2d)
  @accepts flx_rho_x, flx_rho_y, ldg_rho_x, ldg_rho_y,
           flx_qx_x, flx_qx_y, ldg_qx_x, ldg_qx_y,
           #=flx_qy_x,=# flx_qy_y, ldg_qy_x, ldg_qy_y,
           flx_E_x, flx_E_y, ldg_E_x, ldg_E_y,
           smoothed_mu

  flx_qy_x = flx_qx_y

  flx_rho_x += smoothed_mu * ldg_rho_x
  flx_rho_y += smoothed_mu * ldg_rho_y

  flx_qx_x  += smoothed_mu * ldg_qx_x
  flx_qx_y  += smoothed_mu * ldg_qx_y

  flx_qy_x  += smoothed_mu * ldg_qy_x
  flx_qy_y  += smoothed_mu * ldg_qy_y

  flx_E_x   += smoothed_mu * ldg_E_x
  flx_E_y   += smoothed_mu * ldg_E_y

  @returns flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y,
           flx_qy_x, flx_qy_y, flx_E_x, flx_E_y
end


@with_signature [legacy=false] function av_nflux_2d(eq::EulerEquation2d)
  @accepts        ldg_rho_x, ldg_rho_y, ldg_qx_x, ldg_qx_y, ldg_qy_x, ldg_qy_y, ldg_E_x, ldg_E_y, smoothed_mu
  @accepts [bdry] bdry_ldg_rho_x, bdry_ldg_rho_y, bdry_ldg_qx_x, bdry_ldg_qx_y, bdry_ldg_qy_x,
                  bdry_ldg_qy_y, bdry_ldg_E_x, bdry_ldg_E_y, bdry_smoothed_mu
  @accepts [bdry] nflx_rho, nflx_qx, nflx_qy, nflx_E
  @accepts [bdry] nx, ny

  nflx       = nx*ldg_rho_x + ny*ldg_rho_y
  bdry_nflx  = nx*bdry_ldg_rho_x + ny*bdry_ldg_rho_y
  nflx_rho  += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  nflx       = nx*ldg_qx_x + ny*ldg_qx_y
  bdry_nflx  = nx*bdry_ldg_qx_x + ny*bdry_ldg_qx_y
  nflx_qx   += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  nflx       = nx*ldg_qy_x + ny*ldg_qy_y
  bdry_nflx  = nx*bdry_ldg_qy_x + ny*bdry_ldg_qy_y
  nflx_qy   += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  nflx       = nx*ldg_E_x + ny*ldg_E_y
  bdry_nflx  = nx*bdry_ldg_E_x + ny*bdry_ldg_E_y
  nflx_E    += 0.5 * (smoothed_mu*nflx + bdry_smoothed_mu*bdry_nflx)

  @returns [bdry] nflx_rho, nflx_qx, nflx_qy, nflx_E
end


@with_signature [legacy=false] function bdryllf(equation::EulerEquation2d)
  @accepts rho, flx_rho_x, flx_rho_y,
           qx,  flx_qx_x, flx_qx_y,
           qy,  #=flx_qy_x,=# flx_qy_y,
           E,   flx_E_x, flx_E_y, v
  @accepts init_rho, init_qx, init_qy, init_E, init_v
  @accepts nx, ny

  TODO()

  lhs_flx_qy_x = lhs_flx_qx_y

  # enforce dirichlet/neumann conditions following stability analysis
  # see dg-notes/tex/evm.pdf
  # Hesthaven, Numerical Methods for Conservation Laws uses same logic

  # Dirichlet conditions
  rhs_rho = -lhs_rho + 2*init_rho
  rhs_qx  = -lhs_qx  + 2*init_qx
  rhs_qy  = -lhs_qy  + 2*init_qy
  rhs_E   = -lhs_E   + 2*init_E

  # Neumann conditions
  # rhs_rho = lhs_rho
  # rhs_qx  = lhs_qx
  # rhs_qy  = lhs_qy
  # rhs_E   = lhs_E

  @unpack eos = equation
  rhs_q = sqrt(rhs_qx^2 + rhs_qy^2)
  rhs_eps = rhs_E / rhs_rho - rhs_q^2 / (2 * rhs_rho^2)
  rhs_p = eos(Pressure, Density, rhs_rho, InternalEnergy, rhs_eps)
  rhs_flx_rho_x = rhs_qx
  rhs_flx_rho_y = rhs_qy
  rhs_flx_qx_x  = rhs_qx^2 / rhs_rho + rhs_p
  rhs_flx_qx_y  = rhs_qx*rhs_qy / rhs_rho
  rhs_flx_qy_x  = rhs_flx_qx_y
  rhs_flx_qy_y  = rhs_qy^2 / rhs_rho + rhs_p
  rhs_flx_E_x   = rhs_qx / rhs_rho * (rhs_E + rhs_p)
  rhs_flx_E_y   = rhs_qy / rhs_rho * (rhs_E + rhs_p)

  vmax = absmax(lhs_v, init_v)

  lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
  rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)

  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  nflx_qx  = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)

  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  nflx_qy  = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)

  lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
  rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)

  @returns nflx_rho, nflx_qx, nflx_qy, nflx_E
end
