initialdata!(env, P, prms) = initialdata!(env, P, prms, env.mesh)

function initialdata!(env, P::Project, prms, mesh::Mesh1d)
  initialdata_equation_of_state!(env, P, P.equation.eos, prms, mesh)
  # initialdata_entropy_variables!(env, P, prms)
  initialdata_hrsc!(env, P, P.hrsc, prms, mesh)
  initialdata_tci!(env, P, P.tci, prms)
  initialdata_bdrycond!(env, P, P.bdrycond, mesh)
  # initialdata_smooth!(env, P, mesh)
end

function initialdata!(env, P::Project2d, prms, mesh::Mesh2d)
  initialdata_equation_of_state!(env, P, P.equation.eos, prms, mesh)
  initialdata_hrsc!(env, P, P.hrsc, prms, mesh)
end


function initialdata_equation_of_state!(env, P::Project, eos::EquationOfState.IdealGas, prms, ::Mesh1d)

  @unpack id = prms

  @unpack cache, mesh = env

  (xmin, xmax), = extrema(mesh)
  γ = eos.gamma
  x = vec(mesh.x)

  E(p, rho, v) = p / (γ - 1.0) + 0.5 * rho * v^2

  tuple_data = if id == "smooth_isentropic_flow"
    # Glaubitz, PhD thesis 2020, p. 219
    @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
    @toggled_assert γ == 3.0
    ρ0 = @. 1.0 + 0.5 * sin(π * x)
    v0 = @. 0.0 * x
    p0 = @. ρ0^γ
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  elseif id == "sod_shock_tube"
    @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
    @toggled_assert γ == 1.4
    ρ0 = [ (xi < 0.0) ? 1.0 : 0.125  for xi in x ]
    v0 = [ (xi < 0.0) ? 0.0 : 0.0    for xi in x ]
    p0 = [ (xi < 0.0) ? 1.0 : 0.1    for xi in x ]
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  elseif id == "shu_osher_shock_tube_v1"
    # taken from Glaubitz 2020, PhD thesis
    # results do not agree, I believe that ID is wrong
    @toggled_assert isapprox(xmin, 0.0) && isapprox(xmax, 4.0)
    @toggled_assert γ == 1.4
    ρ0 = [ (xi <= 2.5) ? 3.857143 : 1.0 + 0.2 * sin(10*π*xi) for xi in x ]
    v0 = [ (xi <= 2.5) ? 2.629369 : 0.0    for xi in x ]
    p0 = [ (xi <= 2.5) ? 10.333333 : 1.0   for xi in x ]
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  elseif id == "shu_osher_shock_tube_v2"
    # taken from Yu, Hesthaven, 2019
    @toggled_assert isapprox(xmin, -5.0) && isapprox(xmax, 5.0)
    @toggled_assert γ == 1.4
    ρ0 = [ (xi <= -4.0) ? 3.857143 : 1.0 + 0.2 * sin(5*xi) for xi in x ]
    v0 = [ (xi <= -4.0) ? 2.629369 : 0.0    for xi in x ]
    p0 = [ (xi <= -4.0) ? 10.333333 : 1.0    for xi in x ]
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  else
    error("Initial data of type $id not implemented!")
  end

  # pack data
  @unpack rho, q, E = get_dynamic_variables(cache)
  @unpack p         = get_static_variables(cache)
  rho .= tuple_data[1]
  q   .= tuple_data[2]
  E   .= tuple_data[3]
  p   .= tuple_data[4]

  return
end


function initialdata_equation_of_state!(env, P, eos::EquationOfState.IdealGas, prms, ::Mesh2d)

  @unpack id = prms
  @unpack cache, mesh = env
  @unpack x, y = get_static_variables(cache)
  @unpack Npts = mesh.element
  Kx, Ky = mesh.tree.dims

  (xmin, xmax), (ymin, ymax) = extrema(mesh)
  γ = eos.gamma

  E(p, rho, v) = p / (γ - 1.0) + 0.5 * rho * v^2
  tuple_data = if id == "smooth_isentropic_flow"
    # Glaubitz, PhD thesis 2020, p. 219
    @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
    @toggled_assert γ == 3.0
    ρ0 = @. 1.0 + 0.5 * sin(π * x)
    vx0 = @. 0.0 * x
    vy0 = @. 0.0 * x
    p0 = @. ρ0^γ
    qx0 = @. ρ0 * vx0
    qy0 = @. ρ0 * vy0
    v0 = @. sqrt(vx0^2 + vy0^2)
    E0 = @. E(p0, ρ0, v0)
    (ρ0, qx0, qy0, E0, p0)
  elseif id == "kelvin_helmholtz"
    @toggled_assert isapprox(xmin, 0.0) && isapprox(xmax, 1.0) && isapprox(ymin, 0.0) && isapprox(ymax, 1.0)
    @toggled_assert isapprox(γ, 1.4)

    # initial data taken from
    # 'Convergence of discontinuous Galerkin schemes for the Euler equations via dissipative weak solutions'
    # Mária Lukáˇcová-Medvid’ová, Philipp Öffner
    # not quite working yet, because we are not enforcing positivity of density
    # the above paper uses the positivity preserving limiter of Shu and Zhang

    # setup am, bm coeffs
    rng = Random.MersenneTwister(123)
    M, d = 10, 0.01
    as = randn(rng, M)
    as ./= sum(as)
    bs = randn(rng, M)
    @. bs = pi * (2*bs - 1)
    function Y(x)
      sum = 0.0
      for m = 1:M
        sum += d * as[m] * cos(bs[m] + 2 * pi * m * x)
      end
      return sum
    end
    # ylow, yup = [ 0.25 for _ = 1:Kx*Npts ], [ 0.75 for _ = 1:Kx*Npts ]
    ρ0, vx0, vy0, p0 = [ similar(x) for _ = 1:4 ]
    CI = CartesianIndices((Npts*Kx,Npts*Ky))
    for (i,(xi,yi)) in enumerate(zip(x,y))
      ylow = 0.25 + Y(xi)
      yup  = 0.75 + Y(xi)
      if yi >= ylow && yi <= yup
        ρ0[i]  = 2.0
        vx0[i] = -0.5
        vy0[i] = 0.0
        p0[i]  = 2.5
      else
        ρ0[i]  = 1.0
        vx0[i] = 0.5
        vy0[i] = 0.0
        p0[i]  = 2.5
      end
    end
    qx0 = @. ρ0 * vx0
    qy0 = @. ρ0 * vy0
    v0 = @. sqrt(vx0^2 + vy0^2)
    E0 = @. E(p0, ρ0, v0)
    (ρ0, qx0, qy0, E0, p0)
  else
    error("Initial data of type $id not implemented!")
  end

  # pack data
  @unpack rho, qx, qy, E = get_dynamic_variables(cache)
  @unpack p              = get_static_variables(cache)
  rho .= tuple_data[1]
  qx  .= tuple_data[2]
  qy  .= tuple_data[3]
  E   .= tuple_data[4]
  p   .= tuple_data[5]

  return
end


# function initialdata_entropy_variables!(env, P::Project, prms)
#   @unpack cache = env
#   @unpack equation = P
#   @unpack t, tm1 = get_global_variables(cache)
#   @unpack S, Sm1, flx_S, flx_Sm1 = get_static_variables(cache)
#   tm1 .= 0.0
#   t   .= 0.0
#   broadcast_volume!(entropy, equation, cache)
#   broadcast_volume!(entropy_flux, equation, cache)
#   @. Sm1 = S
#   @. flx_Sm1 = flx_S
# end


function initialdata_hrsc!(env, P, hrsc::Maybe{HRSC.AbstractHRSC}, prms, mesh) end

function initialdata_hrsc!(env, P, hrsc::HRSC.SmoothedArtificialViscosity, prms, mesh::Mesh1d)
  @unpack cache       = env
  @unpack mu          = get_cell_variables(mesh.cache)
  @unpack smoothed_mu = get_static_variables(mesh.cache)
  initialdata_hrsc!(env, P, hrsc.av, prms, mesh)
  hrsc.smoother(smoothed_mu, mu, env.mesh, mesh.tree.periodic[1])
end
function initialdata_hrsc!(env, P, hrsc::HRSC.SmoothedArtificialViscosity, prms, mesh::Mesh2d)
  @unpack cache       = env
  @unpack mu          = get_cell_variables(mesh.cache)
  @unpack smoothed_mu = get_static_variables(mesh.cache)
  initialdata_hrsc!(env, P, hrsc.av, prms, mesh)
  hrsc.smoother(smoothed_mu, mu, env.mesh, true, true)
end

function initialdata_hrsc!(env, P::Project, hrsc::HRSC.EntropyViscosity, prms, ::Mesh1d)
  @unpack mu                     = get_cell_variables(env.cache)
  @unpack S, Sm1, flx_S, flx_Sm1 = get_static_variables(env.cache)
  @unpack t, tm1 = get_global_variables(env.cache)
  tm1 .= 0.0
  t   .= 0.0
  dg1d.new_broadcast_volume!(entropy_variables, P.equation, env.mesh)
  @. Sm1 = S
  @. flx_Sm1 = flx_S

  # EV method needs starting aid in the form of 'artificial' AV from a different method
  prms = parameters(:HRSC)
  prms["method"] = "av"
  prms["av_method"] = "mda"
  dg1d.check_parameters(:HRSC, prms)
  mda_hrsc = HRSC.make_HRSC(env.mesh, prms)

  @unpack mu, max_v   = get_cell_variables(env.cache)
  @unpack v           = get_static_variables(env.cache)
  @unpack rho         = get_dynamic_variables(env.cache)
  dg1d.new_broadcast_volume!(maxspeed, P.equation, env.mesh)
  Npts, K = layout(env.mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      rho, max_v,
      mda_hrsc.av)
end
# function initialdata_hrsc!(env, P, hrsc::HRSC.EntropyViscosity, prms, mesh::Mesh1d)
#   @unpack smoothed_mu, S, Sm1, flx_S, flx_Sm1 = get_static_variables(mesh.cache)
#   @unpack mu = get_cell_variables(mesh.cache)
#   @. mu = 0.0
#   @. smoothed_mu = S = Sm1 = flx_S = flx_Sm1 = 0.0
# end

function initialdata_hrsc!(env, P::Project2d, hrsc::HRSC.EntropyViscosity, prms, mesh::Mesh2d)
  @unpack smoothed_mu, S, Sm1, flx_S_x, flx_S_y,
          flxm1_S_x, flxm1_S_y = get_static_variables(mesh.cache)
  @unpack mu = get_cell_variables(mesh.cache)
  @. mu = 0.0
  @. smoothed_mu = S = Sm1 = flx_S_x =flx_S_y = flxm1_S_x = flxm1_S_y = 0.0
end


function initialdata_tci!(env, P::Project, tci::Nothing, prms) end


function initialdata_tci!(env, P::Project, tci::TCI.EntropyProduction, prms)
  @unpack cache = env
  @unpack EP = get_static_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs = get_bdry_variables(cache)
  @. EP = 0.0
  @. EPjump_lhs = 0.0
  @. EPjump_rhs = 0.0
end


function initialdata_tci!(env, P::Project, tci::TCI.AbstractTCI, prms)
  @unpack flag = get_cell_variables(env.cache)
  @. flag = 0.0
end


function initialdata_bdrycond!(env, P::Project, bdrycond::dg1d.DirichletBC2, mesh::Mesh1d)
  @unpack rho, q, E = get_dynamic_variables(mesh.cache)
  @unpack flx_rho, flx_q, flx_E, v = get_static_variables(mesh.cache)
  @unpack init_flx_rho, init_flx_q, init_flx_E, init_v = get_static_variables(mesh.cache)
  @unpack init_rho, init_q, init_E = get_static_variables(mesh.cache)
  dg1d.new_broadcast_volume!(maxspeed, P.equation, mesh)
  dg1d.new_broadcast_volume!(flux, P.equation, mesh)
  init_rho     .= rho
  init_q       .= q
  init_E       .= E
  init_flx_rho .= flx_rho
  init_flx_q   .= flx_q
  init_flx_E   .= flx_E
  init_v       .= v
end
function initialdata_bdrycond!(env, P::Project, bdrycond::dg1d.DirichletBC2, mesh::Mesh2d)
  @unpack rho, qx, qy, E                                = get_dynamic_variables(mesh.cache)
  @unpack init_rho, init_qx, init_qy, init_E, v, init_v = get_static_variables(mesh.cache)
  TODO()
  dg1d.new_broadcast_volume!(maxspeed, P.equation, mesh)
  dg1d.new_broadcast_volume!(flux, P.equation, mesh)
  init_rho .= rho
  init_qx  .= qx
  init_qy  .= qy
  init_E   .= E
  init_v   .= v
end


function initialdata_smooth!(env, P, mesh)
  @unpack rho, q, E = get_dynamic_variables(mesh.cache)
  prms = parameters(:HRSC)
  prms["method"] = "filter"
  dg1d.check_parameters(:HRSC, prms)
  bernstein = HRSC.make_HRSC(mesh, prms)
end
