### TODO Move include to GRHD.jl or merge with this file?
include("cons2prim_implementations.jl")


module Formulae
  ### Auxilary functions

  # Lorentz factor
  Wprim(B, ρ, vr, ϵ, p) = 1.0 / sqrt(1.0 - B^2 * vr^2)
  Wcons(B, D, Sr, τ, p) = (τ + p + D) / sqrt((τ + p + D)^2 - (Sr / B)^2)

  ### Primitive variables
  # in terms of conservative ones

  ρ(B, D, Sr, τ, p)  = D / Wcons(B, D, Sr, τ, p)
  vr(B, D, Sr, τ, p) = Sr / (B^2 * (τ + D + p))
  ϵ(B, D, Sr, τ, p) = (sqrt((τ + p + D)^2 - (Sr/B)^2) - Wcons(B, D, Sr, τ, p) * p) / D - 1

  ### Conservative variables
  # in terms of primitive variables

  D(B, ρ, vr, ϵ, p) = ρ * Wprim(B, ρ, vr, ϵ, p)
  Sr(B, ρ, vr, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(B, ρ, vr, ϵ, p)^2 * B^2 * vr
  τ(B, ρ, vr, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(B, ρ, vr, ϵ, p)^2 - p - ρ * Wprim(B, ρ, vr, ϵ, p)

  ### Fluxes
  # in terms of primitive variables

  fD(B, ρ, vr, ϵ, p) = D(B, ρ, vr, ϵ, p) * vr
  fSr(B, ρ, vr, ϵ, p) = Sr(B, ρ, vr, ϵ, p) * vr + p
  fτ(B, ρ, vr, ϵ, p) = (τ(B, ρ, vr, ϵ, p) + p) * vr

  #### Sources

  function sD(r, A, A_r, B, B_r, ρ, vr, ϵ, p)
    val = - A * B^3 * D(B, ρ, vr, ϵ, p) * vr
    return val
  end

  function sSr(r, A, A_r, B, B_r, ρ, vr, ϵ, p)
    h = 1 + ϵ + p / ρ
    Sr = Formulae.Sr(B, ρ, vr, ϵ, p)
    W = Wprim(B, ρ, vr, ϵ, p)
    v2 = B^2 * vr^2
    ρhW2 = ρ * h * W^2
    val = - (ρhW2 - p) * r * B^3 * A_r
    val += A * B^4 * (r * B_r + B / 3) * (ρhW2 * vr^2 + 3 * p  / B^2)
    val += - 4 / 3 * A * B^3 * Sr * vr
    return val
  end

  function sτ(r, A, A_r, B, B_r, ρ, vr, ϵ, p)
    τ = Formulae.τ(B, ρ, vr, ϵ, p)
    h = 1 + ϵ + p / ρ
    W = Wprim(B, ρ, vr, ϵ, p)
    ρhW2 = ρ * h * W^2
    val = - r * B^3 * ρhW2 * vr * A_r - A * B^3 * (τ + p) * vr
    return val
  end
end


@with_signature function flux(equation::Equation)
  @accepts State(sD, sSr, stau), p, rho, vr, eps, A, B, r
  rAB3     = r * A * B^3
  flx_sD, flx_sSr, flx_stau = 0, 0, 0
  try
    flx_sD   = rAB3 * Formulae.fD(B, rho, vr, eps, p)
    flx_sSr  = rAB3 * Formulae.fSr(B, rho, vr, eps, p)
    flx_stau = rAB3 * Formulae.fτ(B, rho, vr, eps, p)
  catch e
    println("failed @ r = $r")
    rethrow(e)
  end
  @returns flx_sD, flx_sSr, flx_stau
end


@with_signature function speed(equation::Equation)
  @accepts State(sD, sSr, stau), p, rho, vr, eps, A, B, r
  # TODO Implement characteristics and derive maxiumum wave speeds.
  # Might be as simple as adding a missing factor for the auxiliary metric to the
  # eigenvalues, since ts is the only thing the generalized Valencia formulation adds
  # to the principal part.
  # for now we use speed of light
  max_v = 0.999
  @returns max_v
end


# @with_signature function impose_atmosphere(equation::Equation)
#   @accepts rho, p, eps, v, D, S, tau
#   @unpack atmosphere = equation
#
#   ρatm = atmosphere.rho
#   patm = 0.0
#   ϵatm = tau / ρatm
#   sm = 1.0
#   Watm = 1.0
#   vatm = 1.0
#
#   if rho < atmosphere.rho * atmosphere.threshold
#     rho = rhoatm
#     p   = patm
#     eps = epsatm
#     v   = vatm
#     D   = Formulae.D(rhoatm, vatm, epsatm, patm)
#     S   = Formulae.S(rhoatm, vatm, epsatm, patm)
#     tau = Formulae.τ(rhoatm, vatm, epsatm, patm)
#   end
#
#   @returns rho, p, eps, v, D, S, tau
# end


@with_signature function primitives(equation::Equation)
  @accepts D, Sr, tau, p, B
  rho = Formulae.ρ(B, D, Sr, tau, p)
  vr  = Formulae.vr(B, D, Sr, tau, p)
  eps = Formulae.ϵ(B, D, Sr, tau, p)
  @returns rho, vr, eps
end


@with_signature function prim2cons(equation::Equation)
  @accepts rho, vr, eps, p, B
  D   = Formulae.D(B, rho, vr, eps, p)
  Sr  = Formulae.Sr(B, rho, vr, eps, p)
  tau = Formulae.τ(B, rho, vr, eps, p)
  @returns D, Sr, tau
end


@with_signature function entropy_variables(equation::Equation{EquationOfState.IdealGas})
  @accepts rho, p, E, flx_E
  TODO(entropy_variables)
  # @unpack eos = equation
  # γ     = eos.gamma
  # E     = rho / (γ - 1) * log(p / rho^γ)
  # flx_E = q * E / rho
  @returns E, flx_E
end


@with_signature function sources(equation::Equation)
  @accepts rho, vr, p, eps, r, A, A_r, B, B_r
  src_sD   = Formulae.sD(r, A, A_r, B, B_r, rho, vr, eps, p)
  src_sSr  = Formulae.sSr(r, A, A_r, B, B_r, rho, vr, eps, p)
  src_stau = Formulae.sτ(r, A, A_r, B, B_r, rho, vr, eps, p)
  @returns src_sD, src_sSr, src_stau
end


@with_signature function cons2scaledcons(equations::Equation)
  @accepts D, Sr, tau, B, r
  sD   = r * B^3 * D
  sSr  = r * B^3 * Sr
  stau = r * B^3 * tau
  @returns sD, sSr, stau
end


@with_signature function scaledcons2cons(equations::Equation)
  @accepts sD, sSr, stau, B, r
  D   = sD / (r * B^3)
  Sr  = sSr / (r * B^3)
  tau = stau / (r * B^3)
  @returns D, Sr, tau
end


"""
Postprocess points where we divided by zero in `scaledcons2cons`.
Do so by interpolating value where x = 0 using symmetry/anti-symmetry
"""
function postprocess_scaledcons2cons!(cache::Cache, equations::Equation, mesh::Mesh)

  @unpack x = mesh
  Npts, K = layout(mesh)

  if isodd(Npts) && isodd(K)
    error("Cannot yet handle case where r = 0 falls into symmetric cell with even polynomial!")
  end

  # find points where x coordinate is zero
  idxs = findall(xi->abs(xi) < 1e-14, mesh.x)

  @unpack D, Sr, tau    = get_static_variables(cache)

  for con in (D, tau) # symmetric variables need to be interpolated
    mat_con  = reshape(view(con, :), (Npts, K))
    for idx in idxs
      i, j = Tuple(idx)
      mat_con[i,j] = 0 # remove nan, because we are computing inplace
      Di = mesh.element.D[i,:] # derivative weights for ith point
      Dii = Di[i]
      mat_con[i,j] = - dot(Di, mat_con[:,j]) / Dii
    end
  end
  for con in (Sr,) # asymmetric variables need to be set to zero
    for idx in idxs
      i, j = Tuple(idx)
      con[i+(j-1)*Npts] = 0
    end
  end
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_flux(equations::Equation)
  @accepts State(sD, sSr, stau)
  flx_ldg_sD, flx_ldg_sSr, flx_ldg_stau = sD, sSr, stau
  @returns flx_ldg_sD, flx_ldg_sSr, flx_ldg_stau
end


@with_signature function ldg_speed(equations::Equation)
  @accepts State(sD, sSr, stau)
  v_ldg_sD, v_ldg_sSr, v_ldg_stau = 1, 1, 1
  @returns v_ldg_sD, v_ldg_sSr, v_ldg_stau
end


@with_signature function av_flux(equations::Equation)
  @accepts State(ldg_sD, ldg_sSr, ldg_stau), smoothed_mu
  flx_ldg_sD, flx_ldg_sSr, flx_ldg_stau =
    smoothed_mu * ldg_sD, smoothed_mu * ldg_sSr, smoothed_mu * ldg_stau
  @returns flx_ldg_sD, flx_ldg_sSr, flx_ldg_stau
end


@with_signature function av_speed(equations::Equation)
  @accepts State(ldg_sD, ldg_sSr, ldg_stau), smoothed_mu
  flx_ldg_sD, flx_ldg_sSr, flx_ldg_stau = smoothed_mu, smoothed_mu, smoothed_mu
  @returns v_flx_ldg_sD, v_flx_ldg_sSr, v_flx_ldg_stau
end
