#######################################################################
#                              Timestep                               #
#######################################################################


function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh = env
  @unpack equation = P
  @unpack max_v = get_static_variables(cache)

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  vmax = dg1d.absolute_maximum(max_v)
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N^2 * vmax)
  return dt
end


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(speed, equation, cache)
  max_v = get_variable(cache, :max_v)
  vmax = dg1d.absolute_maximum(view(max_v,:))
  vmax_limit = 0.9999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(env, P::Project, hrsc::Nothing, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, rsolver = P

  @unpack sD, sSr, stau             = get_dynamic_variables(cache)
  @unpack flx_sD, flx_sSr, flx_stau = get_static_variables(cache)
  @unpack rhs_sD, rhs_sSr, rhs_stau = get_rhs_variables(cache)
  @unpack lhs_numflx_sD,   rhs_numflx_sD,
          lhs_numflx_sSr,  rhs_numflx_sSr,
          lhs_numflx_stau, rhs_numflx_stau = get_bdry_variables(cache)

  broadcast_volume!(scaledcons2cons, equation, cache)
  postprocess_scaledcons2cons!(cache, equation, mesh)
  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(sources, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  @unpack p, vr, rho, eps = get_static_variables(cache)

  compute_rhs_weak_form!(rhs_sD,   flx_sD,    lhs_numflx_sD,   rhs_numflx_sD,   mesh)
  compute_rhs_weak_form!(rhs_sSr,  flx_sSr,   lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
  compute_rhs_weak_form!(rhs_stau, flx_stau,  lhs_numflx_stau, rhs_numflx_stau, mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractReconstruction,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, rsolver = P

  @unpack sD, sSr, stau                    = get_dynamic_variables(cache)
  @unpack flx_sD, flx_sSr, flx_stau        = get_static_variables(cache)
  @unpack rhs_sD, rhs_sSr, rhs_stau        = get_rhs_variables(cache)
  @unpack lhs_numflx_sD,   rhs_numflx_sD,
          lhs_numflx_sSr,  rhs_numflx_sSr,
          lhs_numflx_stau, rhs_numflx_stau = get_bdry_variables(cache)
  @unpack flag                             = get_cell_variables(cache)

  HRSC.reconstruct!(sD,    flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)
  HRSC.reconstruct!(sSr,   flag, hrsc, isperiodic=isperiodic(bdryconds), limit_with_boundaries=false)
  HRSC.reconstruct!(stau,  flag, hrsc, isperiodic=isperiodic(bdryconds), m=1e-10, limit_with_boundaries=false)

  broadcast_volume!(scaledcons2cons, equation, cache)
  postprocess_scaledcons2cons!(cache, equation, mesh)
  broadcast_volume!(cons2prim, equation, cache)
  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(sources, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)

  compute_rhs_weak_form!(rhs_sD,   flx_sD,    lhs_numflx_sD,   rhs_numflx_sD,   mesh)
  compute_rhs_weak_form!(rhs_sSr,  flx_sSr,   lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
  compute_rhs_weak_form!(rhs_stau, flx_stau,  lhs_numflx_stau, rhs_numflx_stau, mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, rsolver, ldg_rsolver, av_rsolver = P

  @unpack sD, sSr, stau                          = get_dynamic_variables(cache)
  @unpack flx_sD, flx_sSr, flx_stau,
          ldg_sD, ldg_sSr, ldg_stau,
          flx_ldg_sD, flx_ldg_sSr, flx_ldg_stau  = get_static_variables(cache)
  @unpack rhs_sD, rhs_sSr, rhs_stau              = get_rhs_variables(cache)
  @unpack lhs_numflx_sD,   rhs_numflx_sD,
          lhs_numflx_sSr,  rhs_numflx_sSr,
          lhs_numflx_stau, rhs_numflx_stau,
          lhs_numflx_ldg_sD,   rhs_numflx_ldg_sD,
          lhs_numflx_ldg_sSr,  rhs_numflx_ldg_sSr,
          lhs_numflx_ldg_stau, rhs_numflx_ldg_stau = get_bdry_variables(cache)

  broadcast_volume!(scaledcons2cons, equation, cache)
  postprocess_scaledcons2cons!(cache, equation, mesh)
  # broadcast_volume!(conservative_fixing, equation, cache)
  broadcast_volume!(cons2prim, equation, cache)
  # broadcast_volume!(atmosphering, equation, cache)
  broadcast_volume!(cons2scaledcons, equation, cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(mean_flux, ldg_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_sD,   flx_ldg_sD,   lhs_numflx_ldg_sD,   rhs_numflx_ldg_sD,   mesh)
  compute_rhs_weak_form!(ldg_sSr,  flx_ldg_sSr,  lhs_numflx_ldg_sSr,  rhs_numflx_ldg_sSr,  mesh)
  compute_rhs_weak_form!(ldg_stau, flx_ldg_stau, lhs_numflx_ldg_stau, rhs_numflx_ldg_stau, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)

  # broadcast_volume!(scaledcons2cons, equation, cache)
  # postprocess_scaledcons2cons!(cache, equation, mesh)
  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(sources, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  @. flx_sD          += flx_ldg_sD
  @. flx_sSr         += flx_ldg_sSr
  @. flx_stau        += flx_ldg_stau
  @. lhs_numflx_sD   += lhs_numflx_ldg_sD
  @. lhs_numflx_sSr  += lhs_numflx_ldg_sSr
  @. lhs_numflx_stau += lhs_numflx_ldg_stau
  @. rhs_numflx_sD   += rhs_numflx_ldg_sD
  @. rhs_numflx_sSr  += rhs_numflx_ldg_sSr
  @. rhs_numflx_stau += rhs_numflx_ldg_stau

  compute_rhs_weak_form!(rhs_sD,   flx_sD,   lhs_numflx_sD,   rhs_numflx_sD,   mesh)
  compute_rhs_weak_form!(rhs_sSr,  flx_sSr,  lhs_numflx_sSr,  rhs_numflx_sSr,  mesh)
  compute_rhs_weak_form!(rhs_stau, flx_stau, lhs_numflx_stau, rhs_numflx_stau, mesh)

  return true
end
