"""
    MDHViscosity

Highest Modal Decay Artificial Viscosity model, cf. Yu, Hesthaven, 2019.
"""
Base.@kwdef struct MDHViscosity <: AbstractArtificialViscosity
  mesh::Mesh
  cmax::Float64 = 1.0
  ca::Float64   = 0.5
  ck::Float64   = 0.1
end


function compute_viscosity!(
    mu,
    u, v,
    av::MDHViscosity)

  @unpack mesh = av
  @unpack element, K, L = mesh
  @unpack V, invV, N = element
  @unpack cmax, ca, ck = av

  h = L / (N * K)

  u = reshape(u, layout(mesh))

  # truncated expansion
  u_modal = invV * u
  diff_u_modal = deepcopy(u_modal)
  diff_u_modal[1:end-1,:] .= 0.0
  diff_u = V * diff_u_modal

  # diff_u = u .- u_trunc
  norm2_diff_u = dg1d.cellwise_inner_product(diff_u, diff_u, mesh)
  norm2_u      = dg1d.cellwise_inner_product(u, u, mesh)

  S = abs.(norm2_diff_u ./ norm2_u)
  # if |u| is close to zero then assume no energy in highest mode
  S[abs.(norm2_u) .< 1e-12] .= 1e-20
  # if |diff_u| = 0 then set to finite positive value, needed for log below
  S[abs.(S) .== 0.0] .= 1e-20

  if any(isinf.(S)) || any(isnan.(S))
    error("Encountered Inf or NaN in estimation of highest mode energy!")
  end

  # shock sensor
  s = @. log10(S)
  s0 = - (ca + 4.0 * log10(N))

  # (3.3)
  # maxabs_v = vec(maximum(abs.(v), dims=1))
  # mu = @. cmax * h / N * maxabs_v
  maxabs_v = extrema(abs.(v))[2]
  fill!(mu, cmax * h / N * maxabs_v)

  for k in 1:K
    sk = s[k]
    # (3.6)
    if sk < s0 - ck
      mu[k] *= 0.0
    elseif s0 - ck <= sk <= s0 + ck
      mu[k] *= 0.5 * sin(1.0 + π * 0.5 * (sk - s0) / ck)
    else
      mu[k] *= 1.0
    end
  end

  return true
end
