initialdata!(env, P, prms) = initialdata!(env, P, prms, env.mesh)

function initialdata!(env, P::Project, prms, mesh::Mesh1d)
  initialdata_equation_of_state!(env, P, P.equation.eos, prms, mesh)
  # initialdata_entropy_variables!(env, P, prms)
  initialdata_hrsc!(env, P, P.hrsc, prms, mesh)
  initialdata_tci!(env, P, P.tci, prms, mesh)
  return
end

function initialdata!(env, P::Project2d, prms, mesh::Mesh2d)
  initialdata_equation_of_state!(env, P, P.equation.eos, prms, mesh)
  initialdata_hrsc!(env, P, P.hrsc, prms, mesh)
  initialdata_bdrycond!(env, P, P.bdrycond, mesh)
end


#######################################################################
#                   Initial data equation of state                    #
#######################################################################


function initialdata_equation_of_state!(env, P::Project, eos::EquationOfState.IdealGas, prms, mesh::Mesh1d)

  @unpack id = prms
  @unpack cache, mesh = env
  x = vec(mesh.x)

  ρ0, v0, p0 = initialdata_equation_of_state(Val(Symbol(id)), eos, x)

  ϵ0 = @. eos(InternalEnergy, Density, ρ0, Pressure, p0)

  @unpack D, S, tau = get_dynamic_variables(cache)
  @unpack p         = get_static_variables(cache)
  @. p   = p0
  @. D   = Formulae.D(ρ0, v0, ϵ0, p0)
  @. S   = Formulae.S(ρ0, v0, ϵ0, p0)
  @. tau = Formulae.τ(ρ0, v0, ϵ0, p0)

  broadcast_volume!(cons2prim, P.equation, cache)

  return
end


function initialdata_equation_of_state(::Val{type}, eos, x) where type
  error("initialdata_equation_of_state: Unknown initial data type '$type'")
end


function initialdata_equation_of_state(::Val{:smooth}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)
  @unpack gamma = eos

  # from Bugner, PhD thesis 2018
  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert gamma == 5/3
  ρ0 = @. 1.0 + 0.2 * sin(2.0 * π * x)
  v0 = @. 0.2 * x^0
  p0 = @. 1.0 * x^0

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_relativistic_shocktube}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  # inspired by Bugner, PhD thesis 2018; Marti, Mueller, JCP 1996
  @toggled_assert isapprox(xmin, -2.0) && isapprox(xmax, 2.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 1.0
  ρ0 = [ (xi < -x0 || xi > x0) ? 10.0  : 1.0   for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0   : 0.0   for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 13.33 : 1e-7  for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_relativistic_shocktube_2}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  # inspired by Bugner, PhD thesis 2018; Marti, Mueller, JCP 1996
  @toggled_assert isapprox(xmin, -2.0) && isapprox(xmax, 2.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 1.0
  ρ0 = [ (xi < -x0 || xi > x0) ? 1.0    : 1.0   for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0    : 0.0   for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 1000.0 : 0.01  for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_shocktube_pressure_jump}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)
  γ          = eos.gamma

  # selfmade, less extreme case than periodic_relativistic_shocktube with only a jump
  # in pressure
  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert γ == 5/3
  x0 = 0.5
  ρ0 = [ (xi < -x0 || xi > x0) ? 10.0  :  10.0 for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0   :  0.0  for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 13.33 :  0.1  for xi in x ]
  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_shocktube_less_extreme}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  # selfmade, less extreme case than periodic_relativistic_shocktube
  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert gamma == 5/3
  x0 = 0.5
  ρ0 = [ (xi < -x0 || xi > x0) ? 10.0  :  0.5   for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0   :  0.0   for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 8.0   :  1.5 for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:periodic_sod_shocktube}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  @toggled_assert isapprox(xmin, -2.0) && isapprox(xmax, 2.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 1.0
  ρ0 = [ (xi < -x0 || xi > x0) ? 1.0  :  0.125  for xi in x ]
  v0 = [ (xi < -x0 || xi > x0) ? 0.0  :  0.0    for xi in x ]
  p0 = [ (xi < -x0 || xi > x0) ? 1.0  :  0.1    for xi in x ]

  (ρ0, v0, p0)
end


function initialdata_equation_of_state(
    ::Val{:sod_shocktube}, eos::EquationOfState.IdealGas, x)

  xmin, xmax = extrema(x)

  @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
  @toggled_assert eos.gamma == 5/3
  x0 = 0.0
  ρ0 = [ (xi < x0) ? 1.0  :  0.125  for xi in x ]
  v0 = [ (xi < x0) ? 0.0  :  0.0    for xi in x ]
  p0 = [ (xi < x0) ? 1.0  :  0.1    for xi in x ]

  (ρ0, v0, p0)
end


#######################################################################
#                                 2d                                  #
#######################################################################



function initialdata_equation_of_state!(env, P, eos::EquationOfState.IdealGas, prms, mesh::Mesh2d)

  @unpack id = prms
  @unpack cache, mesh = env
  @unpack x, y = get_static_variables(cache)

  (xmin, xmax), (ymin, ymax) = extrema(mesh)
  Lx = xmax-xmin
  Ly = xmax-xmin
  @unpack gamma = eos

  ρ0, vx0, vy0, p0 = if id == "smooth_diagonal"
    @unpack id_smooth_flow_speed = prms
    @toggled_assert gamma == 5/3

    # 'angled' coordinate
    L = sqrt(Lx^2+Ly^2)
    alpha = atan(Ly/Lx)
    z   = @. 2 * (cos(alpha) * x + sin(alpha) * y) / L

    v = id_smooth_flow_speed
    ρ0  = @. 1.0 + 0.2 * sin(2.0 * π * z)
    vx0 = @. cos(alpha) * v * z^0
    vy0 = @. sin(alpha) * v * z^0
    p0  = @. 1.0 * z^0

    ρ0, vx0, vy0, p0
  elseif id == "shocktube_2d"
    @toggled_assert xmin == -1.0 && xmax == 1.0 && ymin == -1.0 && ymax == 1.0
    # TODO What EoS is this test for?
    ρ0, vx0, vy0, p0 = [ similar(x) for _ = 1:4 ]
    for (i,(xi,yi)) in enumerate(zip(x,y))
      ρ, vx, vy, p = if xi >= 0 && yi >= 0
        0.03515, 0.0, 0.0, 0.163
      elseif xi < 0 && yi >= 0
        0.1, 0.7, 0.0, 1.0
      elseif xi < 0 && yi < 0
        0.5, 0.0, 0.0, 1.0
      else # xi >= 0 && yi < 0
        0.1, 0.0, 0.7, 1.0
      end
      ρ0[i]  = ρ
      vx0[i] = vx
      vy0[i] = vy
      p0[i]  = p
    end
    ρ0, vx0, vy0, p0
  elseif id == "vortex"
    TODO()
  else
    error("Initial data of type $id not implemented!")
  end

  ϵ0 = @. eos(InternalEnergy, Density, ρ0, Pressure, p0)

  @unpack D, Sx, Sy, tau = get_dynamic_variables(cache)
  @unpack p              = get_static_variables(cache)
  @. p   = p0
  @. D   = Formulae2d.D(ρ0,  vx0, vy0, ϵ0, p0)
  @. Sx  = Formulae2d.Sx(ρ0, vx0, vy0, ϵ0, p0)
  @. Sy  = Formulae2d.Sy(ρ0, vx0, vy0, ϵ0, p0)
  @. tau = Formulae2d.τ(ρ0,  vx0, vy0, ϵ0, p0)

  broadcast_volume_2!(cons2prim, P.equation, mesh)

  @unpack vx, vy = get_static_variables(cache)

  return
end


#######################################################################
#                   Initial data entropy variables                    #
#######################################################################


function initialdata_entropy_variables!(env, P::Project, prms)
  @TODO
  @unpack cache = env
  @unpack t, tm1 = get_global_variables(cache)
  @unpack E, Em1, flx_E, flx_Em1 = get_static_variables(cache)
  tm1 .= 0.0
  t   .= 0.0
  compute_entropy_variables!(P, P.equation.eos)
  @. Em1 = E
  @. flx_Em1 = flx_E
end


#######################################################################
#                          Initial data HRSC                          #
#######################################################################


function initialdata_hrsc!(env, P, hrsc::Maybe{HRSC.AbstractHRSC}, prms, mesh) end


function initialdata_hrsc!(env, P, hrsc::HRSC.AbstractArtificialViscosity, prms, mesh)
  @unpack mu = get_cell_variables(env.cache)
  @. mu = 0.0
end


function initialdata_hrsc!(env, P::Project, hrsc::HRSC.SmoothedArtificialViscosity, prms, mesh)
  initialdata_hrsc!(env, P, hrsc.av, prms)
  @unpack smoothed_mu = get_static_variables(env.cache)
  # TODO Entropy vars should also be initialized
  @. smoothed_mu = 0.0
end

function initialdata_hrsc!(env, P::Project2d, hrsc::HRSC.AbstractArtificialViscosity, prms, mesh::Mesh2d)
  @unpack smoothed_mu, E, Em1, flx_E_x, flx_E_y,
          flxm1_E_x, flxm1_E_y = get_static_variables(mesh.cache)
  @unpack mu = get_cell_variables(mesh.cache)
  @. mu = 0.0
  @. smoothed_mu = E = Em1 = flx_E_x =flx_E_y = flxm1_E_x = flxm1_E_y = 0.0
end

#######################################################################
#                          Initial data TCI                           #
#######################################################################


function initialdata_tci!(env, P::Project, tci::Nothing, prms, mesh) end


function initialdata_tci!(env, P::Project, tci::TCI.EntropyProduction, prms, mesh)
  @unpack cache = env
  @unpack EP = get_static_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs = get_bdry_variables(cache)
  @. EP = 0.0
  @. EPjump_lhs = 0.0
  @. EPjump_rhs = 0.0
end


function initialdata_tci!(env, P::Project, tci::TCI.AbstractTCI, prms, mesh)
  @unpack flag = get_cell_variables(env.cache)
  @. flag = 0.0
end


#######################################################################
#                              Bdryconds                              #
#######################################################################


function initialdata_bdrycond!(env, P::Project2d, bdrycond, mesh)
  TODO()
end


function initialdata_bdrycond!(env, P::Project2d, bdrycond::dg1d.DirichletBC2, mesh)
  @unpack D, Sx, Sy, tau = get_dynamic_variables(mesh.cache)
  @unpack flx_D_x, flx_D_y, flx_Sx_x, flx_Sx_y, flx_Sy_x, flx_Sy_y,
          flx_tau_x, flx_tau_y, max_v = get_static_variables(mesh.cache)
  @unpack init_flx_D_x, init_flx_D_y, init_flx_Sx_x, init_flx_Sx_y, init_flx_Sy_x, init_flx_Sy_y,
          init_flx_tau_x, init_flx_tau_y, init_max_v = get_static_variables(mesh.cache)
  @unpack init_D, init_Sx, init_Sy, init_tau = get_static_variables(mesh.cache)
  broadcast_volume_2!(cons2prim, P.equation, mesh)
  broadcast_volume_2!(maxspeed, P.equation, mesh)
  broadcast_volume_2!(flux, P.equation, mesh)
  init_D         .= D
  init_Sx        .= Sx
  init_Sy        .= Sy
  init_tau       .= tau
  init_flx_D_x   .= flx_D_x
  init_flx_D_y   .= flx_D_y
  init_flx_Sx_x  .= flx_Sx_x
  init_flx_Sx_y  .= flx_Sx_y
  init_flx_Sy_x  .= flx_Sy_x
  init_flx_Sy_y  .= flx_Sy_y
  init_flx_tau_x .= flx_tau_x
  init_flx_tau_y .= flx_tau_y
  init_max_v     .= max_v
end
