Project(env::Environment, prms) = Project(env, prms, env.mesh)


function Project(env::Environment, prms, mesh::Mesh1d)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = make_Equation(eos, prms["SRHD"], mesh)
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])
  rsolver = ApproxRiemannSolver(flux, speed, equation)
  ldg_rsolver, av_rsolver = if hrsc isa AbstractArtificialViscosity
    ApproxRiemannSolver(ldg_flux, ldg_speed, equation),
    ApproxRiemannSolver(av_flux, av_flux, equation)
  else
    nothing, nothing
  end

  P = Project(equation, rsolver, hrsc, tci, ldg_rsolver, av_rsolver)

  # register variables
  # TODO add a ::Nothing overload for register_variables!
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, P)
  register_variables!(mesh, rsolver)
  !isnothing(ldg_rsolver) && register_variables!(mesh, ldg_rsolver)
  !isnothing(av_rsolver) && register_variables!(mesh, av_rsolver, dont_register=true)
  display(cache)

  # setup boundary conditions
  bdryconds = make_BoundaryConditions(env, equation, rsolver, prms["SRHD"])
  ldg_bdryconds, av_bdryconds = if hrsc isa AbstractArtificialViscosity
    make_BoundaryConditions_LDG(env, equation, ldg_rsolver, prms["SRHD"]),
    make_BoundaryConditions_LDG(env, equation, av_rsolver, prms["SRHD"])
  else
    nothing, nothing
  end

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup initial data
  initialdata!(env, P, prms["SRHD"])

  # setup callbacks
  projectcb = make_callback(env, P, bdryconds)

  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


function Project(env::Environment, prms, mesh::Mesh2d)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  hrsc = HRSC.make_HRSC(mesh, prms["HRSC"])
  equation = make_Equation(eos, prms["SRHD"], mesh)
  bdrycond = dg1d.DirichletBC2()

  P = Project2d(equation, hrsc, bdrycond)

  # register variables
  # TODO add a ::Nothing overload for register_variables!
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(mesh, P)
  _register_variables!(mesh, hrsc)
  _register_variables!(mesh, bdrycond)
  display(cache)

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup initial data
  initialdata!(env, P, prms["SRHD"])

  # setup boundary conditions
  bdryconds_x = nothing
  bdryconds_y = nothing
  ldg_bdryconds, av_bdryconds = nothing, nothing

  projectcb = make_callback(env, P)
  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, ((bdryconds_x, bdryconds_y), ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(mesh::Mesh, P::Project)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:D, :S, :tau,),            # conservatives
    rhs_variablenames     = (:rhs_D, :rhs_S, :rhs_tau), # rhs of conservatives
    static_variablenames  = (:flx_D, :flx_S, :flx_tau,  # conservatives' fluxes
                             :rho, :v, :p, :eps,        # primitives
                             :max_v,                    # maximum charactersitic speed
                             :E, :Em1,                  # entropy
                             :flx_E, :flx_Em1),         # entropy flux
    cell_variablenames    = (:cellmax_v,),
    global_variablenames  = (:t, :tm1) # time steps
  )
  _register_variables!(mesh, P.hrsc)
  _register_variables!(mesh, P.tci)
  @unpack x = get_static_variables(mesh.cache)
  # @. x = env.mesh.x[:]
end


function _register_variables!(mesh, tci_or_hrsc::Nothing) end


_register_variables!(mesh, hrsc::AbstractHRSC) =
  error("_register_variables: Not implemented for hrsc of type '$(typeof(hrsc))'")

_register_variables!(mesh, hrsc::HRSC.AbstractReconstruction) = nothing


function _register_variables!(mesh, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_D, :ldg_S, :ldg_tau,              # local DG variable
                             :flx_ldg_D, :flx_ldg_S, :flx_ldg_tau), # local DG flux
    cell_variablenames    = (:mu,), # 'one viscosity to rule them all'
    )
end


function _register_variables!(mesh, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(mesh, hrsc.av)
  register_variables!(mesh.cache,
    static_variablenames  = (:smoothed_mu,)
  )
end


function _register_variables!(mesh::Mesh2d, hrsc::HRSC.EntropyViscosity)
  register_variables!(mesh.cache,
    static_variablenames = (:E, :Em1, :flx_E_x, :flx_E_y, :flxm1_E_x, :flxm1_E_y,
                            :ldg_D_x, :ldg_D_y, :ldg_Sx_x, :ldg_Sx_y,
                            :ldg_Sy_x,  :ldg_Sy_y,  :ldg_tau_x,  :ldg_tau_y),
    bdry_variablenames   = (:nflx_D_x, :nflx_D_y, :nflx_Sx_x, :nflx_Sx_y,
                            :nflx_Sy_x, :nflx_Sy_y, :nflx_tau_x, :nflx_tau_y),
    cell_variablenames   = (:mu,),
    global_variablenames = (:tm1,)
  )
end


function _register_variables!(mesh, tci::TCI.AbstractTCI)
  register_variables!(mesh.cache, cell_variablenames = (:flag,:D_flag,:S_flag,:tau_flag))
end


function _register_variables!(mesh, tci::TCI.EntropyProduction)
  register_variables!(mesh.cache,
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end


function _register_variables!(mesh, P::Project2d)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:D, :Sx, :Sy, :tau,),            # conservatives
    rhs_variablenames     = (:rhs_D, :rhs_Sx, :rhs_Sy, :rhs_tau), # rhs of conservatives
    static_variablenames  = (:flx_D_x, :flx_D_y,
                             :flx_Sx_x, :flx_Sx_y,
                             :flx_Sy_x, :flx_Sy_y,
                             :flx_tau_x, :flx_tau_y,          # conservatives' fluxes
                             :rho, :vx, :vy, :p, :eps,        # primitives
                             :max_v),                         # maximum charactersitic speed
    bdry_variablenames    = (:nflx_D, :nflx_Sx, :nflx_Sy, :nflx_tau),
    global_variablenames  = (:t,),
    cell_variablenames    = (:cellmax_v,),
  )
end


function _register_variables!(mesh, bdrycond::dg1d.DirichletBC2)
  register_variables!(mesh.cache,
    static_variablenames  = (:init_D, :init_Sx, :init_Sy, :init_tau,
                             :init_flx_D_x, :init_flx_D_y,
                             :init_flx_Sx_x, :init_flx_Sx_y,
                             :init_flx_Sy_x, :init_flx_Sy_y,
                             :init_flx_tau_x, :init_flx_tau_y,
                             :init_max_v)
  )
end
