function make_callback(env, P::Project, bdryconds::BoundaryConditions)
  cbfn_equation(u, t) = callback_equation(u, t, env, P, isperiodic(bdryconds), P.equation, env.mesh)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, env, P, isperiodic(bdryconds), P.tci, env.mesh)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, isperiodic(bdryconds), P.hrsc, env.mesh)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  return callbackset
end


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, env, P, isperiodic, eq::AbstractScalarEq, ::Mesh) end


function callback_equation(state_u, state_t, env, P, isperiodic, eq::AbstractScalarEq, ::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation = P
  wrap_dynamic_variables!(cache, state_u)
  @unpack u                     = get_dynamic_variables(cache)
  @unpack E, Em1, F, Fm1        = get_static_variables(cache)
  @unpack t, tm1                = get_global_variables(cache)

  # backup previous values
  @. Em1  = E
  @. Fm1  = F
  @. tm1  = t
  t      .= state_t
  # compute new values
  broadcast_volume!(entropy, equation, cache)
  broadcast_volume!(entropy_flux, equation, cache)
end


#######################################################################
#                                HRSC                                 #
#######################################################################


callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Nothing, mesh) = nothing


callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.AbstractHRSC, mesh) =
  error("not yet implemented for type $(typeof(hrsc))")


function callback_hrsc(state_u, state_t, env, P, isperiodic,
  hrsc::Union{HRSC.BernsteinReconstruction, HRSC.WENOReconstruction}, mesh) end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.AbstractArtificialViscosity, mesh)
  @unpack cache, mesh = env
  @unpack equation = P
  @unpack v         = get_static_variables(cache)
  @unpack mu, max_v = get_cell_variables(cache)
  @unpack u         = get_dynamic_variables(cache)
  broadcast_volume!(speed, equation, cache)
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      u, max_v,
      hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.TCIViscosity, mesh)
  @unpack cache, mesh = env
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh::Mesh1d)
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack cache = env
  @unpack rsolver = P
  @unpack mu = get_cell_variables(cache)
  @unpack smoothed_mu = get_static_variables(cache)
  hrsc.smoother(smoothed_mu,mu,mesh,isperiodic)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh::Mesh2d)
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack cache = env
  @unpack rsolver = P
  @unpack mu = get_cell_variables(cache)
  @unpack smoothed_mu = get_static_variables(cache)
  hrsc.smoother(smoothed_mu,mu,mesh,true,true)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, mesh::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation    = P
  @unpack u              = get_dynamic_variables(cache)
  @unpack E, F, Em1, Fm1, v = get_static_variables(cache)
  @unpack mu, max_v      = get_cell_variables(cache)
  @unpack t, tm1         = get_global_variables(cache)
  broadcast_volume!(speed, equation, cache)
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end
  HRSC.compute_viscosity!(
      mu,
      u, max_v, E, Em1, F, Fm1, t[1], tm1[1],
      hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, mesh::Mesh2d)
  @unpack cache, mesh  = env
  @unpack equation     = P
  @unpack u            = get_dynamic_variables(cache)
  @unpack E, Em1, v,
          F_x, F_y,
          F_xm1, F_ym1,
          dxdX, dydY   = get_static_variables(cache)
  @unpack t, tm1       = get_global_variables(cache)
  @unpack mu           = get_cell_variables(cache)

  # backup previous values
  @. Em1   = E
  @. F_xm1 = F_x
  @. F_ym1 = F_y
  @. tm1   = t
  @. t     = state_t

  # compute new values
  broadcast_volume_2!(entropy, equation, mesh)
  broadcast_volume_2!(speed, equation, mesh)
  broadcast_volume_2!(entropy_flux, equation, mesh)

  HRSC.compute_viscosity!(
      mu,
      u, v, E, Em1, F_x, F_y, F_xm1, F_ym1, t[1], tm1[1], dxdX, dydY,
      hrsc)
end


#######################################################################
#                                 TCI                                 #
#######################################################################


callback_tci(state_u, state_t, env, P, isperiodic, tci::Nothing, mesh) = nothing


callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.AbstractTCI) =
  error("not yet implemented for type $(typeof(tci))")


  function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.EntropyProduction, ::Mesh1d)
  @unpack cache, mesh            = env
  @unpack u                      = get_dynamic_variables(cache)
  @unpack E, Em1, F, Fm1, EP     = get_static_variables(cache)
  @unpack flag                   = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs = get_bdry_variables(cache)
  @unpack t, tm1                 = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         E, Em1, F, Fm1, t[1], tm1[1],
                         tci)
end


# function callback_equation(state_u, state_t, env, P, isperiodic, eq::TCI.EntropyProduction, ::Mesh2d)
#   wrap_dynamic_variables!(cache, state_u)
#   @unpack cache, mesh = env
#   @unpack equation = P
#
#   @unpack u               = get_dynamic_variables(cache)
#   @unpack E, F_x, F_y,
#              F_xm1, F_ym1 = get_static_variables(cache)
#   @unpack t, tm1          = get_global_variables(cache)
#   @unpack EPjump          = get_bdry_variables(cache)
#
#   # backup previous values
#   @. Em1   = E
#   @. F_xm1 = F_x
#   @. F_ym1 = F_y
#   @. tm1   = t
#   @. t     = state_t
#
#   broadcast_volume_2!(entropy, equation, mesh)
#   broadcast_volume_2!(speed, equation, mesh)
#   broadcast_volume_2!(entropy_flux, equation, mesh)
#
#   TCI.compute_indicator!(flag, EP, EPjump, E, Em1, F_x, F_y, F_xm1, F_ym1, t[1], tm1[1], tci)
# end


function callback_tci(state_u, state_t, env, P, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest}, mesh)
  @unpack cache, mesh            = env
  @unpack u                      = get_dynamic_variables(cache)
  @unpack flag                   = get_cell_variables(cache)
  TCI.compute_indicator!(flag, u, tci)
end
