#######################################################################
#                               Factory                               #
#######################################################################


function make_Equation(prms, dims)
  @unpack equation = prms
  if dims == 1
    if equation == "advection"
      v = prms["advection_v"]
      @toggled_assert length(v) == dims
      return Advection(v isa Vector ? v[1] : v)
    elseif equation == "burgers"
      return Burgers()
    else
    end
  elseif dims == 2
    if equation == "advection"
      v = prms["advection_v"]
      @toggled_assert length(v) == dims
      return Advection2d(v...)
    else
      return Burgers2d()
    end
  end
  error("Unknown equation requested: '$equation'")
end


#######################################################################
#                          AbstractScalarEq                           #
#######################################################################


@with_signature function llf_1d(eq::AbstractScalarEq)
  @accepts Prefix(lhs), flx_u, u, v
  @accepts Prefix(rhs), flx_u, u, v
  @accepts nx
  lhs_nflx = nx*lhs_flx_u
  rhs_nflx = nx*rhs_flx_u
  vmax = absmax(lhs_v, rhs_v)
  nflx_u = LLF(lhs_nflx, rhs_nflx, lhs_u, rhs_u, vmax)
  @returns nflx_u
end
@with_signature function bdryllf_1d(eq::AbstractScalarEq)
  @accepts Prefix(lhs), flx_u, u
  @accepts Prefix(rhs), init_flx_u, init_u
  @accepts nx
  lhs_nflx = nx*lhs_flx_x
  rhs_init_nflx = nx*rhs_init_flx_x
  vmax = absmax(lhs_v, rhs_init_v)
  nflx_u = LLF(lhs_nflx, rhs_init_nflx, lhs_u, rhs_init_u, vmax)
  @returns nflx_u
end


@with_signature function llf_2d(eq::AbstractScalarEq)
  @accepts Prefix(lhs), flx_x, flx_y, u, v
  @accepts Prefix(rhs), flx_x, flx_y, u, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_x + ny*lhs_flx_y
  rhs_nflx = nx*rhs_flx_x + ny*rhs_flx_y
  # TODO This should use absmax
  vmax = max(lhs_v, rhs_v)
  nflx_u = LLF(lhs_nflx, rhs_nflx, lhs_u, rhs_u, vmax)
  @returns nflx_u
end
@with_signature function bdryllf_2d(eq::AbstractScalarEq)
  @accepts Prefix(lhs), flx_x, flx_y, u, v
  @accepts Prefix(rhs), init_flx_x, init_flx_y, init_u, init_v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_x + ny*lhs_flx_y
  rhs_init_nflx = nx*rhs_init_flx_x + ny*rhs_init_flx_y
  # TODO This should use absmax
  vmax = max(lhs_v, rhs_init_v)
  nflx_u = LLF(lhs_nflx, rhs_init_nflx, lhs_u, rhs_init_u, vmax)
  @returns nflx_u
end


#######################################################################
#                              Advection                              #
#######################################################################


@with_signature function flux(eq::Advection)
  @accepts State(u)
  flx_u = eq.v * u
  @returns flx_u
end

@with_signature function speed(eq::Advection)
  @accepts State(u)
  v = eq.v
  @returns v
end

@with_signature function entropy(eq::Advection)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Advection)
  @accepts u
  F = u^2 / 2
  @returns F
end

#######################################################################
#                            Advection 2d                             #
#######################################################################


@with_signature function flux(eq::Advection2d)
  @accepts State(u)
  flx_x, flx_y = eq.v_x * u, eq.v_y * u
  @returns flx_x, flx_y
end

@with_signature function speed(eq::Advection2d)
  @accepts State(u)
  v = sqrt(eq.v_x^2 + eq.v_y^2)
  @returns v
end


@with_signature function entropy(eq::Advection2d)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Advection2d)
  @accepts u
  F_x = eq.v_x * u^2 / 2
  F_y = eq.v_y * u^2 / 2
  @returns F_x, F_y
end


#######################################################################
#                               Burgers                               #
#######################################################################


@with_signature function flux(eq::Burgers)
  @accepts State(u)
  flx_u = u^2 / 2
  @returns flx_u
end

@with_signature function speed(eq::Burgers)
  @accepts State(u)
  v = u
  @returns v
end

@with_signature function entropy(eq::Burgers)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Burgers)
  @accepts u
  F = u^3 * 2 / 3
  @returns F
end


#######################################################################
#                              Burgers2d                              #
#######################################################################


@with_signature function flux(eq::Burgers2d)
  @accepts State(u)
  flx_x, flx_y = u^2/2, u^2/2
  @returns flx_x, flx_y
end

@with_signature function speed(eq::Burgers2d)
  @accepts State(u)
  v = sqrt(2) * u
  @returns v
end

@with_signature function entropy(eq::Burgers2d)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Burgers2d)
  @accepts u
  F_x = u^3 * 2/3
  F_y = u^3 * 2/3
  @returns F_x, F_y
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_flux_1d(eq::AbstractScalarEq)
  @accepts u
  flx_q = u
  @returns flx_q
end
@with_signature function ldg_nflux_1d(eq::AbstractScalarEq)
  @accepts Prefix(lhs), flx_q, q
  @accepts Prefix(rhs), flx_q, q
  @accepts nx
  lhs_nflx = nx*lhs_flx_q
  rhs_nflx = nx*rhs_flx_q
  nflx_q = 0.5 * (lhs_nflx + rhs_nflx)
  @returns nflx_q
end


@with_signature function av_flux_1d(eq::AbstractScalarEq)
  @accepts flx_u, q, smoothed_mu
  flx_u += smoothed_mu * q
  @returns flx_u
end
@with_signature function av_nflux_1d(eq::AbstractScalarEq)
  @accepts Prefix(lhs), q, smoothed_mu
  @accepts Prefix(rhs), q, smoothed_mu
  @accepts nx, nflx_u
  lhs_nflx = nx*lhs_q
  rhs_nflx = nx*rhs_q
  nflx_u += 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)
  @returns nflx_u
end


@with_signature function ldg_flux_2d(eq::AbstractScalarEq)
  @accepts u
  flx_qx_x = u
  flx_qy_y = u
  flx_qx_y = flx_qy_x = 0.0
  @returns flx_qx_x, flx_qy_y, flx_qx_y, flx_qy_x
end
@with_signature function ldg_nflux_qx(eq::AbstractScalarEq)
  @accepts Prefix(lhs), flx_qx_x, flx_qx_y, qx
  @accepts Prefix(rhs), flx_qx_x, flx_qx_y, qx
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  nflx_qx = 0.5 * (lhs_nflx + rhs_nflx)
  @returns nflx_qx
end
@with_signature function ldg_nflux_qy(eq::AbstractScalarEq)
  @accepts Prefix(lhs), flx_qy_x, flx_qy_y, qy
  @accepts Prefix(rhs), flx_qy_x, flx_qy_y, qy
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  nflx_qy = 0.5 * (lhs_nflx + rhs_nflx)
  @returns nflx_qy
end


@with_signature function av_flux_2d(eq::AbstractScalarEq)
  @accepts qx, qy, smoothed_mu
  flx_g_x = smoothed_mu * qx
  flx_g_y = smoothed_mu * qy
  @returns flx_g_x, flx_g_y
end
@with_signature function av_nflux_2d(eq::AbstractScalarEq)
  @accepts Prefix(lhs), qx, qy, smoothed_mu
  @accepts Prefix(rhs), qx, qy, smoothed_mu
  @accepts nx, ny
  lhs_nflx = nx*lhs_qx + ny*lhs_qy
  rhs_nflx = nx*rhs_qx + ny*rhs_qy
  nflx_g = 0.5 * (lhs_smoothed_mu*lhs_nflx + rhs_smoothed_mu*rhs_nflx)
  @returns nflx_g
end
