function initialdata!(env, P::Project, prms)
  initialdata_equation!(env, env.mesh, P, P.equation, prms)
  initialdata_hrsc!(env, env.mesh, P, P.hrsc, prms)
  initialdata_tci!(env, env.mesh, P, P.tci, prms)
  initialdata_bdryconds!(env, env.mesh, P, P.bdrycond, prms)
  return
end


function initialdata_equation!(env, mesh::Mesh1d, P::Project, equation::AbstractScalarEq, prms)

  @unpack cache, mesh = env
  @unpack u = get_dynamic_variables(cache)
  @unpack L, x, xmin, xmax = mesh

  @unpack id = prms
  x = x[:]

  id_fn = if id == "sine"
    @unpack id_sine_period, id_sine_phase = prms
    u .= @. sin.(2 * pi * id_sine_period * x/L + id_sine_phase)
  elseif id == "bump"
    @unpack id_bump_height, id_bump_width = prms
    @toggled_assert id_bump_width < L "Initialdata 'bump': Bump width exceeds x domain '$xrange'"
    xl, xr = xmin+(L-id_bump_width)/2, xmin+(L+id_bump_width)/2
    u .= @. (x < xl || x > xr) ? 0.5 : id_bump_height
  elseif id == "gaussian"
    A = prms["id_gaussian_amplitude"]
    w = prms["id_gaussian_width"]
    @toggled_assert w > 0.0
    x0 = (xmax + xmin) / 2
    u .= @. A * exp(-(x-x0)^2 / w^2)
  elseif id == "smooth bump"
    error("TODO Implement 'smooth bump' initial data")
#   elseif type === :smooth_bump
#     xl = xmin + L * 0.10
#     xr = xmin + L * 0.90
#     gap_x = 0.3 * L
#     u0 = zeros(size(x))
#     for (i, xi) in enumerate(x)
#       if xi < xl || xi > xr
#         u0[i] = 0.0
#       elseif xi > xl+gap_x && xi < xr-gap_x
#         u0[i] = 1.0
#       elseif xi >= xl && xi <= xl+gap_x
#         u0[i] = sin( (xi - xl) / gap_x * 0.5 * π)^2
#       elseif xi >= xr-gap_x && xi <= xr
#         u0[i] = 1.0 - sin( (xi - (xr - gap_x)) / gap_x * 0.5 * π)^2
#       end
#     end
  else
    error("Unknown initialdata requested: '$id'")
  end

  return
end


function initialdata_equation!(env, mesh::Mesh2d, P::Project, equation::AbstractScalarEq, prms)
  @unpack cache, mesh = env
  @unpack u = get_dynamic_variables(cache)
  @unpack x, y = get_static_variables(cache)
  @unpack id = prms

  (xmin, xmax), (ymin, ymax) = mesh.extends

  Lx = xmax - xmin
  Ly = ymax - ymin
  alpha = atan(Ly,Lx)
  LL = sqrt(Lx^2 + Ly^2)

  id_fn = if id == "sine"
    # @unpack id_sine_period, id_sine_phase = prms
    @. u = sin(2.0 * pi * (x/Lx + y/Ly))
  elseif id == "gaussian"
    A = prms["id_gaussian_amplitude"]
    w = prms["id_gaussian_width"]
    @toggled_assert w > 0.0
    x0 = (xmax + xmin) / 2
    y0 = (ymax + ymin) / 2
    @. u .= A * exp(-((x-x0)^2 + (y-y0)^2) / w^2)
  elseif id == "bump"
    @unpack id_bump_height, id_bump_width = prms
    @toggled_assert id_bump_width < Lx "Initialdata 'bump': Bump width exceeds x domain '$xrange'"
    xl, xr = xmin+(Lx-id_bump_width)/2, xmin+(Lx+id_bump_width)/2
    @views @. u[x < xl || x > xr] = 0.0
    @views @. u[x >= xl && x <= xr] = id_bump_height
  elseif id == "smoothed_bump"
    xl = xmin + Lx * 0.30
    xr = xmin + Lx * 0.70
    yd = ymin + Ly * 0.30
    yu = ymin + Ly * 0.70
    gap_x = 0.1 * Lx
    gap_y = 0.1 * Ly
    for (i, (xi, yi)) in enumerate(zip(x,y))
      if xi < xl || xi > xr || yi < yd || yi > yu
        u[i] = 0.0
      elseif xi > xl+gap_x && xi < xr-gap_x && yi > yd+gap_y && yi < yu-gap_y
        u[i] = 1.0
      elseif xi >= xl && xi <= xl+gap_x
        if yi >= yd+gap_y && yi <= yu-gap_y
          # xmin stripe
          u[i] = sin( (xi - xl) / gap_x * 0.5 * π)^2
        elseif yi < yd+gap_y
          # xmin-ymin corner
          u[i] = (sin( (xi - xl) / gap_x * 0.5 * π)^2 *
                  sin( (yi - yd) / gap_y * 0.5 * π)^2)
        else # yi > yu-gap_y
          # xmin-ymax corner
          u[i] = (sin( (xi - xl) / gap_x * 0.5 * π)^2 *
                  sin( (-yi + yu) / gap_y * 0.5 * π)^2)
        end
      elseif xi >= xr-gap_x && xi <= xr
        if yi >= yd+gap_y && yi <= yu-gap_y
          # xmax stripe
          u[i] = sin( (-xi + xr) / gap_x * 0.5 * π)^2
        elseif yi < yd+gap_y
          # xmax-ymin corner
          u[i] = (sin( (-xi + xr) / gap_x * 0.5 * π)^2 *
                  sin( (-yi + yd) / gap_y * 0.5 * π)^2)
        else # yi > yu-gap_y
          # xmax-ymax corner
          u[i] = (sin( (-xi + xr) / gap_x * 0.5 * π)^2 *
                  sin( (-yi + yu) / gap_y * 0.5 * π)^2)
        end
      elseif yi >= yd && yi <= yd+gap_y # && xi >= xl+gap_x && xi <= xr-gap_x
        # ymin stripe
        u[i] = sin( (yi - yd) / gap_y * 0.5 * π)^2
      else # yi >= yu-gap_y && yi <= yu && xi >= xl+gap_x && xi <= xr-gap_x
        # ymax stripe
        u[i] = sin( (-yi + yu) / gap_y * 0.5 * π)^2
      end
    end
  else
    error("Unknown initialdata requested: '$id'")
  end

  return
end


function initialdata_hrsc!(env, mesh, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, prms) end


function initialdata_hrsc!(env, mesh, P::Project, hrsc::HRSC.AbstractArtificialViscosity, prms)
  @unpack mu = get_cell_variables(env.cache)
  @. mu = 0.0
end


function initialdata_hrsc!(env, mesh::Mesh1d, P::Project, hrsc::HRSC.EntropyViscosity, prms)
  @unpack E, Em1, F, Fm1 = get_static_variables(env.cache)
  @unpack mu             = get_cell_variables(env.cache)
  @. E = F = Em1 = Fm1 = mu = 0.0
end


function initialdata_hrsc!(env, mesh::Mesh2d, P::Project, hrsc::HRSC.EntropyViscosity, prms)
  @unpack E, Em1, F, Fm1 = get_static_variables(env.cache)
  @unpack mu             = get_cell_variables(env.cache)
  @. E = E_prev = F_x = F_x_prev = F_y = F_y_prev = mu = 0.0
end


function initialdata_tci!(env, mesh, P::Project, tci::Nothing, prms) end


function initialdata_tci!(env, mesh, P::Project, tci::TCI.AbstractTCI, prms)
  @unpack flag = get_cell_variables(env.cache)
  @. flag = 0.0
end


function initialdata_bdryconds!(env, mesh, P, bdryconds, prms) end
function initialdata_bdryconds!(env, mesh::Mesh2d, P::Project, bdryconds, prms)
  @unpack init_u, init_flx_x, init_flx_y, init_v,
          flx_x, flx_y, v = get_static_variables(mesh.cache)
  @unpack u = get_dynamic_variables(mesh.cache)
  init_u  .= u
  init_flx_x .= flx_x
  init_flx_y .= flx_y
  init_v  .= v
end
