#######################################################################
#                              Timestep                               #
#######################################################################

function timestep(env, P::Project, ::Mesh1d, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh = env
  @unpack equation = P
  @unpack v = get_static_variables(cache)

  broadcast_volume!(speed, equation, cache)
  vmax = dg1d.absolute_maximum(v)
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack CFL, element = mesh
  @unpack N, z = element
  dz = z[2]-z[1]
  dx, = minimum(dg1d.widths(mesh.boxes[1]))
  dl = dx * dz

  dt = CFL * dl / (N * vmax)
  return dt
end


function timestep(env, P::Project, ::Mesh2d, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(speed, equation, cache)

  @unpack v = get_static_variables(cache)
  vmax = dg1d.absolute_maximum(v)

  @unpack CFL, element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = CFL * dl / (N * vmax)

  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(env, P::Project, hrsc, bdryconds...)
  return rhs!(env, env.mesh, P, hrsc, bdryconds...)
# TODO Dispatch should be so simple ...
# function dg1d.rhs!(P::Project)
#   return rhs!(P.mesh, P.eq, P.hrsc, P.bdryconds)
end


function rhs!(env, mesh::Mesh1d, P::Project, hrsc::Nothing,
    bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  @unpack cache, mesh = env
  @unpack equation, rsolver = P

  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf_1d, equation, mesh)
  # dg1d.broadcast_bdry_2!(bdryllf_1d, equation, P.bdrycond, mesh)

  @unpack flx_u  = get_static_variables(cache)
  @unpack u      = get_dynamic_variables(cache)
  @unpack rhs_u  = get_rhs_variables(cache)
  @unpack nflx_u = get_bdry_variables(cache)
  compute_rhs_weak_form!(rhs_u, flx_u, nflx_u, mesh)

  return
end


function rhs!(env, mesh::Mesh2d, P::Project, hrsc::Nothing,
    bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  @unpack cache = env
  @unpack equation, rsolver = P

  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf_2d, equation, mesh)
  dg1d.broadcast_bdry_2!(bdryllf_2d, equation, P.bdrycond, mesh)
  # broadcast_boundaryconditions!(central_flux, bdryconds_x, cache, mesh, 0.0)
  # broadcast_boundaryconditions!(central_flux, bdryconds_y, cache, mesh, 0.0)

  @unpack flx_x, flx_y = get_static_variables(cache)
  @unpack u            = get_dynamic_variables(cache)
  @unpack rhs_u        = get_rhs_variables(cache)
  @unpack nflx_u = get_bdry_variables(cache)
  compute_rhs_weak_form!(rhs_u, flx_x, flx_y, nflx_u, mesh)

  return
end


function rhs!(env, mesh::Mesh1d, P::Project, hrsc::Union{BernsteinReconstruction, WENOReconstruction},
    bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  TODO()

  @unpack cache, mesh       = env
  @unpack equation, rsolver = P
  @unpack flag                           = get_cell_variables(cache)
  @unpack flx_u                          = get_static_variables(cache)
  @unpack u                              = get_dynamic_variables(cache)
  @unpack rhs_u                          = get_rhs_variables(cache)
  @unpack lhs_numflx_u, rhs_numflx_u     = get_bdry_variables(cache)

  HRSC.reconstruct!(u, flag, hrsc, isperiodic=isperiodic(bdryconds))

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  compute_rhs_weak_form!(rhs_u, flx_u, lhs_numflx_u, rhs_numflx_u, mesh)

  return
end


function rhs!(env, mesh::Mesh1d, P::Project, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, rsolver, ldg_rsolver, av_rsolver = P

  @unpack flx_u, flx_q, q = get_static_variables(cache)
  @unpack u               = get_dynamic_variables(cache)
  @unpack rhs_u           = get_rhs_variables(cache)
  @unpack nflx_u, nflx_q  = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume_2!(ldg_flux_1d, equation, mesh)
  broadcast_faces_2!(ldg_nflux_1d, equation, mesh)
  compute_rhs_weak_form!(q, flx_q, nflx_q, mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf_1d, equation, mesh)
  broadcast_volume_2!(av_flux_1d, equation, mesh)
  broadcast_faces_2!(av_nflux_1d, equation, mesh)

  compute_rhs_weak_form!(rhs_u, flx_u, nflx_u, mesh)

  return
end


function rhs!(env, mesh::Mesh2d, P::Project, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache = env
  @unpack equation = P

  @unpack flx_x, flx_y,
          flx_qx_x, flx_qx_y,
          flx_qy_x, flx_qy_y,
          flx_g_x, flx_g_y,
          qx, qy, smoothed_mu = get_static_variables(cache)
  @unpack u                   = get_dynamic_variables(cache)
  @unpack rhs_u               = get_rhs_variables(cache)
  @unpack nflx_u, nflx_g,
          nflx_qx, nflx_qy    = get_bdry_variables(cache)

  ## solve auxiliary equations: qx + ∂x u = 0, qy + ∂y u = 0
  broadcast_volume_2!(ldg_flux_2d, equation, mesh)
  broadcast_faces_2!(ldg_nflux_qx, equation, mesh)
  broadcast_faces_2!(ldg_nflux_qy, equation, mesh)
  # broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(qx, flx_qx_x, flx_qx_y, nflx_qx, mesh)
  compute_rhs_weak_form!(qy, flx_qy_x, flx_qy_y, nflx_qy, mesh)

  # ## compute rhs of regularized equation: ∂t u + ∂x f + ∂y f + ∂x μ qx + ∂y μ qy = 0
  broadcast_volume_2!(av_flux_2d, equation, mesh)
  broadcast_faces_2!(av_nflux_2d, equation, mesh)
  # broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)
  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf_2d, equation, mesh)
  # broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  # add up fluxs and numerical fluxes
  @. flx_x  += flx_g_x
  @. flx_y  += flx_g_y
  @. nflx_u += nflx_g
  compute_rhs_weak_form!(rhs_u, flx_x, flx_y, nflx_u, mesh)

  return
end
