#=
# Improvements:
# - Feature: Make Variables type parameteric so that we can also handle non-float variables, e.g.
# integer arrays for cell indices, etc.
# - Feature: Right now variables have to be unique. We might want to provide a
# prefix option to be able to supply different variables for different instances of objects.
# - Enhancement: Benchmark type instabilities.
=#


#######################################################################
#                                Types                                #
#######################################################################


"""
    Cache

Caching object to bundle variables of different kinds.
"""
struct Cache
  variables_to_fields::Dict{Symbol,Symbol}
  fields_to_variables::Dict{Symbol,Vector{Symbol}}
  dynamic_variables::Variables
  static_variables::Variables
  cell_variables::Variables
  rhs_variables::Variables
  bdry_variables::Variables
  global_variables::Variables
end


function Cache(Nptsbulk, Nptsbdry, Ncells)
  verbose = false
  variables_to_fields = Dict{Symbol, Symbol}()
  fields_to_variables = Dict{Symbol, Vector{Symbol}}()
  static_variables  = Variables(Nptsbulk, verbose=verbose)
  rhs_variables     = Variables(Nptsbulk, verbose=verbose)
  dynamic_variables = Variables(Nptsbulk, verbose=verbose)
  cell_variables    = Variables(Ncells, verbose=verbose)
  bdry_variables    = Variables(Nptsbdry, verbose=verbose)
  global_variables  = Variables(1, verbose=verbose)
  return Cache(variables_to_fields, fields_to_variables,
               dynamic_variables, static_variables, cell_variables,
               rhs_variables, bdry_variables, global_variables)
end


#######################################################################
#                               Methods                               #
#######################################################################


function Base.propertynames(::Cache)
  fnames = fieldnames(Cache)
  filter(f -> !in(f, [ :fields_to_variables, :variables_to_fields]), fnames)
end


"""
    variablegroups(cache::Cache)
    variablenames(c::Cache, group::Symbol)

List of available variables in `cache`.
"""
function variablenames(cache::Cache)
  groups = variablegroups(cache)
  varnames = [ names(getproperty(cache, g)) for g in groups ]
  return vcat(varnames...)
end
variablenames(cache::Cache, group::Symbol) =
  vcat([ vi for vi in names(getproperty(cache, group))]...)


"""
    variablegroups(cache::Cache)

List of available group names in `cache`.
"""
variablegroups(c::Cache) = [ g for g in propertynames(c) ]


"""
    get_dynamic_variables(C::Cache)
    get_static_variables(C::Cache)
    get_rhs_variables(C::Cache)
    get_cell_variables(C::Cache)
    get_bdry_variables(C::Cache)
    get_global_variables(C::Cache)

Getters for `Cache` variable fields.
"""
get_dynamic_variables(C::Cache)      = C.dynamic_variables
get_static_variables(C::Cache)       = C.static_variables
get_rhs_variables(C::Cache)          = C.rhs_variables
get_cell_variables(C::Cache)         = C.cell_variables
get_bdry_variables(C::Cache)         = C.bdry_variables
get_global_variables(C::Cache)       = C.global_variables
function get_variable(C::Cache, name::Symbol)::Vector{Float64}
  which_field = getfield(C, :variables_to_fields)[name]
  field = if which_field === :dynamic_variables
      C.dynamic_variables
  elseif which_field === :rhs_variables
      C.rhs_variables
  elseif which_field === :cell_variables
      C.cell_variables
  elseif which_field === :global_variables
      C.global_variables
  elseif which_field === :bdry_variables
      C.bdry_variables
  elseif which_field === :static_variables
      C.static_variables
  end
  return getproperty(field::Variables, name)
end
function get_variable(C::Cache, varname::Symbol, group::Symbol)::Vector{Float64}
  field = getproperty(C, group)
  return getproperty(field, varname)
end


"""
    wrap_dynamic_variables!(C::Cache, u)
    wrap_rhs_variables!(C::Cache, du)

Wrap fields `dynamic_variables, rhs_variables` around data `u, du` that is not
owned by `C`.
"""
function wrap_dynamic_variables!(C::Cache, u::AbstractVector)
  vars = get_dynamic_variables(C)
  set!(vars, u)
  return
end
function wrap_rhs_variables!(C::Cache, du::AbstractVector)
  vars = get_rhs_variables(C)
  set!(vars, du)
  return
end



function register_variables!(C::Cache;
    dynamic_variablenames = [],
    rhs_variablenames     = [],
    static_variablenames  = [],
    cell_variablenames    = [],
    bdry_variablenames    = [],
    global_variablenames  = [])

  register!(get_dynamic_variables(C),  Symbol.(dynamic_variablenames)...)

  register!(get_rhs_variables(C),      Symbol.(rhs_variablenames)...)

  register!(get_static_variables(C),   Symbol.(static_variablenames)...)
  register!(get_cell_variables(C),     Symbol.(cell_variablenames)...)
  register!(get_bdry_variables(C),     Symbol.(bdry_variablenames)...)
  register!(get_global_variables(C),   Symbol.(global_variablenames)...)

  _update_fields_to_variables!(C)

  return
end


function _update_fields_to_variables!(cache::Cache)
  empty!(cache.variables_to_fields)
  for name in dg1d.names(cache.dynamic_variables)
    push!(cache.variables_to_fields, name => :dynamic_variables)
  end
  for name in dg1d.names(cache.rhs_variables)
    push!(cache.variables_to_fields, name => :rhs_variables)
  end
  for name in dg1d.names(cache.static_variables)
    push!(cache.variables_to_fields, name => :static_variables)
  end
  for name in dg1d.names(cache.cell_variables)
    push!(cache.variables_to_fields, name => :cell_variables)
  end
  for name in dg1d.names(cache.bdry_variables)
    push!(cache.variables_to_fields, name => :bdry_variables)
  end
  for name in dg1d.names(cache.global_variables)
    push!(cache.variables_to_fields, name => :global_variables)
  end
  empty!(cache.fields_to_variables)
  cache.fields_to_variables[:dynamic_variables] = dg1d.names(cache.dynamic_variables)
  cache.fields_to_variables[:rhs_variables]     = dg1d.names(cache.rhs_variables)
  cache.fields_to_variables[:static_variables]  = dg1d.names(cache.static_variables)
  cache.fields_to_variables[:cell_variables]    = dg1d.names(cache.cell_variables)
  cache.fields_to_variables[:bdry_variables]    = dg1d.names(cache.bdry_variables)
  cache.fields_to_variables[:global_variables]  = dg1d.names(cache.global_variables)
  all_variablenames = vcat(values(cache.fields_to_variables)...)
  if length(all_variablenames) != length(unique(all_variablenames))
    error("Cache: Duplicated variable names encountered")
  end
end


function Base.show(io::IO, C::Cache)
  variables_list = collect(keys(C.fields_to_variables))
  N_variables = length(variables_list)
  if N_variables == 0 
    @warn "No variables registered!"
    return
  end
  names_list = collect(values(C.fields_to_variables))
  permute_idx = sortperm(variables_list)
  variables_list, names_list = variables_list[permute_idx], names_list[permute_idx]
  max_rows = maximum(length.(names_list))
  data = Matrix{String}(undef, max_rows, N_variables)
  fill!(data, "")
  for (idx, names) in enumerate(names_list)
    N_names = length(names)
    data[1:N_names,idx] .= string.(names)
  end
  # sort!(data, dims=1)
  pretty_table(data, header=variables_list, crop=:none)
end
