#######################################################################
#                                Types                                #
#######################################################################


"""
    AbstractCallback

Supertype for use of callback interface.
The following methods must be implemented by a `Callback<:AbstractCallback`:
- `(cb::Callback)(u, t, iter)` where `u,t,iter` are the current
  statevector, timestep and iteration index. It should return `true`
  if it was triggered and `false` otherwise.
"""
abstract type AbstractCallback end


function (cb::Type{T})(u,t,iter) where T<:AbstractCallback
  error("AbstractCallback interface not implemented!")
end


#######################################################################
#                           Implementations                           #
#######################################################################


#######################################################################
#                           CallbackTiming                            #
#######################################################################


"""
    CallbackTiming(; every_iteration=0, every_dt=0, every_dt_walltime=0)

Object to time whether callbacks should be run after `every_iteration` iterations,
or after `every_dt` (simulation) time has passed, or after `every_dt_walltime` time (in sec)
has passed.

See also: `istimed!`
"""
mutable struct CallbackTiming
  every_iteration::Int64      # update iteration interval
  every_dt::Float64           # update time interval (in secs)
  every_dt_walltime::Float64  # update walltime interval (in secs)
  next_iter::Int64            # next scheduled update wrt iteration
  next_t::Float64             # next scheduled update wrt time
  next_t_walltime::Float64    # next scheduled update wrt time (in secs)
  iter::Int64                 # iteration count
  t::Float64                  # iteration time
  t_walltime::Float64         # iteration walltime (in secs)

  function CallbackTiming(; every_iteration=0, every_dt=0, every_dt_walltime=0)
    @toggled_assert every_iteration >= 0
    @toggled_assert every_dt >= 0
    @toggled_assert every_dt_walltime >= 0
    t_walltime = time()
    return new(every_iteration, every_dt, every_dt_walltime,
               every_iteration, every_dt, t_walltime + every_dt_walltime,
               0,               0,        t_walltime)
  end
end


function state(cbt::CallbackTiming)
  """
  CallbackTiming
    t                   = $(cbt.t)
    t_walltime          = $(cbt.t_walltime)
    next_t              = $(cbt.next_t)
    next_t_walltime     = $(cbt.next_t_walltime)
    every_dt            = $(cbt.every_dt)
    every_dt_walltime   = $(cbt.every_dt_walltime)
    iter                = $(cbt.iter)
    next_iter           = $(cbt.next_iter)
    every_iteration     = $(cbt.every_iteration)
  """
end


Base.show(io::IO, cbt::CallbackTiming) = print(io, state(cbt))


"""
    istimed!(timing::CallbackTiming, t)

Check if callback `timing` should trigger for given time `t`.
"""
function istimed!(timing::CallbackTiming, t)
  @unpack every_iteration, every_dt, every_dt_walltime = timing

  if every_iteration == 0 && every_dt == 0 && every_dt_walltime == 0
    return false
  end

  if t <= timing.t && timing.iter > 0
    return false
  end

  t_walltime         = time()
  timing.iter       += 1
  timing.t           = t
  timing.t_walltime  = t_walltime

  should_we_update_wrt_time = every_dt > 0 && timing.next_t <= t
  if should_we_update_wrt_time
    timing.next_t     += every_dt
  end

  should_we_update_wrt_walltime = every_dt_walltime > 0 && timing.next_t_walltime <= t_walltime
  if should_we_update_wrt_walltime
    timing.next_t_walltime = t_walltime + every_dt_walltime
  end

  should_we_update_wrt_iter = every_iteration > 0 && timing.next_iter <= timing.iter
  if should_we_update_wrt_iter
    timing.next_iter  += every_iteration
  end

  should_we_update = ( should_we_update_wrt_time ||
                       should_we_update_wrt_iter ||
                       should_we_update_wrt_walltime )

  return should_we_update
end


#######################################################################
#                             CallbackSet                             #
#######################################################################


"""
    CallbackSet(::AbstractCallback...)

Use to queue a set of callbacks for evaluation, e.g.
```julia
CallbackSet(SaveCallback(...), PlotCallback(...), ...)
```
Callbacks are called in the order they are queued.
"""
struct CallbackSet
  callbacks::Vector{AbstractCallback}
  CallbackSet(callbacks...) = new([ cb for cb in callbacks ])
end


function (cbs::CallbackSet)(u, t, i=0; force=false)
  anycalled = false
  @timeit TO "CallbackSet" begin
    for idx = 1:length(cbs.callbacks)
      cb = cbs.callbacks[idx]
      status = cb(u, t, i; force=force)::Bool
      anycalled = anycalled || status
    end
  end
  return anycalled
end


Base.pushfirst!(cbs::CallbackSet, cb::AbstractCallback) = pushfirst!(cbs.callbacks, cb)
Base.pushfirst!(cbs::CallbackSet, ::Nothing) = nothing
Base.push!(cbs::CallbackSet, cb::AbstractCallback) = push!(cbs.callbacks, cb)
Base.push!(cbs::CallbackSet, ::Nothing) = nothing
Base.append!(cbs::CallbackSet, cbs2::CallbackSet) = append!(cbs.callbacks, cbs2.callbacks)


function Base.show(io::IO, cbs::CallbackSet)
  for cb in cbs.callbacks
    print(io, cb)
  end
end


#######################################################################
#                          FunctionCallback                           #
#######################################################################


"""
    FunctionCallback(fn::Function, timing::CallbackTiming)

Make callback for calling `fn(u,t)::Bool` at times controlled by `timing`.
"""
struct FunctionCallback{T_Function<:Function} <: AbstractCallback
  func::T_Function
  timing::CallbackTiming
end


function (cb::FunctionCallback)(u, t, iter; force=false)
  @unpack func, timing = cb
  # check if it is update time
  if force
    func(u,t)
    return true
  end
  if istimed!(timing, t)
    func(u, t)
    return true
  end
  return false
end


function state(fcb::FunctionCallback{T_Fn}) where T_Fn
  """
FunctionCallback
  func                = $(nameof(fcb.func))
  timing              =
$(indent(state(fcb.timing), level=3, fillwith="  "))"""
end


Base.show(io::IO, fcb::FunctionCallback{T_Fn}) where T_Fn = print(io, state(fcb))


#######################################################################
#                            SaveCallback                             #
#######################################################################


"""
    SaveCallback(variables::Symbol..., C::Cache, mesh::Mesh, filename, timing::CallbackTiming)

Make callback for saving `variables...` from `C` to file `filename`.
"""
mutable struct SaveCallback <: AbstractCallback
  _file::HDF5.File
  fcb::FunctionCallback
  filename::String
  variablenames::Vector{Tuple{Symbol,Symbol}}

  function SaveCallback(variables::Vector{Symbol}, C::Cache, mesh::Mesh, filename,
                        timing::CallbackTiming;
                        format1d=nothing, format2d=nothing)::Union{Nothing,SaveCallback}

    if isnothing(format1d) && isnothing(format2d)
      error("No output format requested")
    end
    if !isnothing(format1d) && !(format1d in [ "h5" ])
      error("Unknown 1d output format '$format1d' requested")
    end
    if !isnothing(format2d) && !(format2d in [ "h5", "xdmf", "xdmf+h5", "xdmf+bin", "vtk" ])
      error("Unknown 2d output format '$format2d' requested")
    end

    if length(variables) == 0
      return nothing
    end

    # normalize and sanitize filename
    if length(filename) == 0
      error("SaveCallback: Invalid filename = '$filename' ")
    end

    # gather list of registered variables
    registered_groups = variablegroups(C)
    registered_vars = [ variablenames(C, g) for g in registered_groups ]

    filename = Filesystem.abspath(filename)
    dir = Filesystem.dirname(filename)

    # find groups for requested output varialbes
    save_groups_vars = Tuple{Symbol,Symbol}[]
    for var in variables
      idx = findfirst(reg_vars -> var in reg_vars, registered_vars)
      if isnothing(idx)
        error("SaveCallback: Variable '$var' not registered in cache")
      end
      reg_group = registered_groups[idx]
      push!(save_groups_vars, (reg_group, var))
    end


    if format1d == "h5" || format2d == "h5"

      !isdir(dir) && error("Save directory '$dir' does not exist")

      # open a file
      file = h5open(filename, "w")
      flush(file)

      # output function, to be called by FunctionCallback
      h5group_idx = 0 # use internal iteration index to keep contiguous group name increments
      savefn = let h5group_idx=h5group_idx, save_groups_vars=save_groups_vars
        function savefn(u, t)
          # generate group for new time step
          h5group_idx += 1
          h5group = create_group(file, "$(h5group_idx)")
          h5group["t"] = t
          for (group, var) in save_groups_vars
            name = string(var)
            h5group[name] = get_variable(C, var, group)
          end
          # important: use flush to not loose data
          flush(file)
        end
      end

      fcb = FunctionCallback(savefn, timing)

      # save mesh now
      group = create_group(file, "mesh")
      if mesh isa Mesh1d
        group["x"] = get_static_variables(mesh.cache).x
      else
        group["x"] = get_static_variables(mesh.cache).x
        group["y"] = get_static_variables(mesh.cache).y
      end
      flush(file)

      scb = new(file, fcb, filename, save_groups_vars)

      # utilize a finalizer to close the file upon struct destruction
      # this also requires SaveCallback to be mutable
      # one could also use a RefValue and keep it immutable,
      # see https://discourse.julialang.org/t/finalizer-only-works-with-mutable-structs/28777/11?u=fatteneder
      function close_file(scb)
        @async begin
          flush(scb._file)
          close(scb._file)
        end
      end

      return finalizer(close_file, scb)

    elseif format2d == "xdmf" || format2d == "xdmf+bin" || format2d == "xdmf+h5"

      @assert mesh isa Mesh2d

      bulkgroups = (:dynamic_variables, :static_variables, :rhs_variables)
      for v in variables
        if !any(g -> v in variablenames(C, g), bulkgroups)
          error("No 2d output available for variable '$v'. Only variables inside one of these groups supported: '$bulkgroups'")
        end
      end

      mode = format2d == "xdmf+bin" ? :bin : :h5
      xdmf = XDMFOutput(joinpath(dir, "xdmf"), mode)
      xdmfwrite_mesh(xdmf, mesh)

      savefn = let filename=filename, save_groups_vars=save_groups_vars, mesh=mesh
        function savefn_xdmf(u, t)
          vars = [ get_variable(C, var, group) for (group, var) in save_groups_vars ]
          names = [ string(var) for (_,var) in save_groups_vars ]
          xdmfwrite_scalar(xdmf, mesh, t, vars, names)
        end
      end

      fcb = FunctionCallback(savefn, timing)
      file = h5open(joinpath(dir,"tmp"), "w")
      scb = new(file, fcb, filename, save_groups_vars)
      # immediately close, because file handling is done in XDMFOutput now
      close(file)

      return scb


    elseif format2d === "vtk"

      @assert mesh isa Mesh2d

      bulkgroups = (:dynamic_variables, :static_variables, :rhs_variables)
      for v in variables
        if !any(g -> v in variablenames(C, g), bulkgroups)
          error("No 2d output available for variable '$v'. Only variables inside one of these groups supported: '$bulkgroups'")
        end
      end

      vtk = VTKOutput(joinpath(dir, "vtk"))

      savefn = let filename=filename, save_groups_vars=save_groups_vars, mesh=mesh
        function savefn_vtk(u, t)
          vars = [ get_variable(C, var, group) for (group, var) in save_groups_vars ]
          names = [ string(var) for (_,var) in save_groups_vars ]
          vtkwrite_scalar(vtk, mesh, t, vars, names)
        end
      end

      fcb = FunctionCallback(savefn, timing)
      file = h5open(joinpath(dir,"tmp"), "w")
      scb = new(file, fcb, filename, save_groups_vars)
      # immediately close, because file handling is done in XDMFOutput now
      close(file)

      return scb

    end

  end

end


(scb::SaveCallback)(u, t, iter; force=false) = scb.fcb(u, t, iter; force)


function state(scb::SaveCallback)
  """
  SaveCallback
    filename            = $(scb.filename)
    variablenames       = $(scb.variablenames)
  """
end

Base.show(io::IO, scb::SaveCallback) = print(io, state(scb))


#######################################################################
#                            PlotCallback                             #
#######################################################################


"""
    PlotCallback(plotfn::Function, cbt::CallbackTiming; pause::Real=false)

Make callback object for plotting with `plotfn(u,t)` at times controlled by `cbt`.
Set `pause` to a positive number (in seconds) to insert a break after every plot,
or set it to `true` to halt program.
"""
struct PlotCallback <: AbstractCallback
  fcb::FunctionCallback
  pause::Union{Bool,Float64}

  function PlotCallback(plotfn::Function, cbt::CallbackTiming; pause::Real=false)
    fcb = FunctionCallback(plotfn, cbt)
    if !(pause isa Bool)
      @assert pause > 0 "'pause > 0' required"
      pause = Float64(pause)
    end
    return new(fcb, pause)
  end
end


function (pcb::PlotCallback)(u, t, iter; force=false)
  @unpack fcb, pause = pcb

  status = fcb(u, t, iter; force)

  if status
    if pause isa Bool && pause
      println()
      @info "Press any button to continue"
      readline()
    else
      sleep(pause)
    end
  end

  return status
end


function state(pcb::PlotCallback)
  """
  PlotCallback
    pause               = $(pcb.pause)
  """
end

Base.show(io::IO, pcb::PlotCallback) = print(io, state(pcb))


#######################################################################
#                        TimeAlignedSaveCallback                      #
#######################################################################


"""
  TimeAlignedSaveCallback(times::Vector{T}, timestep_fn::Function, scb::SaveCallback) where T<:Real

Callback for aligning adaptive time stepping `timestep_fn(u, t)` with desired
`times` and save data with `scb`.
"""
struct TimeAlignedSaveCallback <: AbstractCallback
  aligned_timestep_fn::Function
  fcb::FunctionCallback

  function TimeAlignedSaveCallback(times::Vector{T}, timestep_fn::Function, scb::SaveCallback) where T<:Real
    @assert length(times) > 0 "Require at least one time stamp"
    @assert all(times .> 0) && all(diff(times) .> 0) """
    Time stamps must be all positive and increasingly ordered
    """
    @assert length(methods(timestep_fn, [Any,Real,Int])) > 0 """
    `timestep_fn` must take arguments `(u,t,n_attempts)`.
    """

    n_times = length(times)
    index_next_t = 1
    schedule_save = false
    # we wrap `timestep_fn` to adjust proposed step in case it exceeds the next desired time stamp
    function aligned_timestep_fn(u, t, n_attempts)
      dt = timestep_fn(u, t, n_attempts)
      if index_next_t <= n_times
        next_t = times[index_next_t]
        if t+dt > next_t
          dt = next_t - t # align time step
          index_next_t += 1
          schedule_save = true # trigger save through save_on_time below
        end
      end
      return dt
    end

    # use a FunctionCallback with this method to save the result of the aligned time stamp
    # we need to split the logic between these two functions, because the proposed
    # time step is computed before the time step is taken after which saving should happen
    function save_on_time(u, t; force=false)
      if schedule_save
        scb(u, t, 0; force=true)
        schedule_save = false
      end
    end
    fcb = FunctionCallback(save_on_time, CallbackTiming(every_iteration=1))

    return new(aligned_timestep_fn, fcb)
  end
end


(ftscb::TimeAlignedSaveCallback)(u, t, iter; force::Bool = false) = ftscb.fcb(u, t, iter; force)


function state(tascb::TimeAlignedSaveCallback)
  """
  TimeAlignedSaveCallback
    aligned_timestep_fn = $(nameof(tascb.aligned_timestep_fn))
    fcb                 = $(state(tascb.fcb))
  """
end

Base.show(io::IO, tascb::TimeAlignedSaveCallback) = print(io, state(tascb))


#######################################################################
#                         TimerOutputCallback                        #
#######################################################################


# TODO Anything wrapping a FunctionCallback should be parameteric
"""
    TimerOutputCallback(to::TimerOutput, filename, timing::CallbackTiming)
"""
mutable struct TimerOutputCallback <: AbstractCallback
  _file::IOStream
  fcb::FunctionCallback
  filename::String
  to::TimerOutput

  function TimerOutputCallback(to::TimerOutput, filename, timing::CallbackTiming)

    # normalize and sanitize filename
    if length(filename) == 0
      error("TimerOutputCallback: Invalid filename = '$filename' ")
    end
    filename = Filesystem.abspath(filename)
    if !endswith(filename, ".log")
      filename = "$(filename).log"
    end

    # open a file
    file = open(filename, "w")

    function savefn(u, t)
      # overwrite file everytime we log
      seekstart(file)
      TimerOutputs.complement!(to)
      println(file, to)
      flush(file)
    end

    fcb = FunctionCallback(savefn, timing)

    tocb = new(file, fcb, filename, to)

    # utilize a finalizer to close the file upon struct destruction
    # this also requires SaveCallback to be mutable
    # one could also use a RefValue and keep it immutable,
    # see https://discourse.julialang.org/t/finalizer-only-works-with-mutable-structs/28777/11?u=fatteneder
    function close_file(tocb)
      @async begin
        flush(tocb._file)
        close(tocb._file)
      end
    end

    finalizer(close_file, tocb)
  end
end


(tocb::TimerOutputCallback)(u, t, iter; force=false) = tocb.fcb(u, t, iter; force)


function state(tocb::TimerOutputCallback)
  """
  TimerOutputCallback
    filename            = $(tocb.filename)
  """
end

Base.show(io::IO, tocb::TimerOutputCallback) = print(io, state(tocb))


#######################################################################
#                         ProgressCallback                            #
#######################################################################


"""
    ProgressCallback(finaltime, progress_every_dt, to::TimerOutput)

Print a progress report and time output summary every `progress_every_dt` (walltime, in seconds)
to `stdout`.
"""
mutable struct ProgressCallback <: AbstractCallback
  fcb::FunctionCallback

  function ProgressCallback(finaltime, progress_every_dt, to::TimerOutput;
                            verbose=true)

    # need to redirect progress to buffer so to not intere with stdout here
    prog_buffer = IOBuffer(; sizehint=2048)
    prog = Progress(100, # %
                    dt=0.0, # always update
                    showspeed=true, output=prog_buffer)
    # accumulate report here
    report_buffer = IOBuffer(; sizehint=2048)
    # track number of newlines in progress report so that we can update stdout
    nlines = -1
    report_every_dt_walltime = 0.3
    t_prev = time()

    function logfn(u, t)

      prcnt = t / finaltime * 100

      t_now = time()
      if t_prev + report_every_dt_walltime > t_now && !isapprox(prcnt,100.0) # we are done
        return
      end
      t_prev = t_now

      if prcnt < 100.0
        ProgressMeter.update!(prog, trunc(Int64,prcnt))
      else
        ProgressMeter.finish!(prog)
      end

      prog_str = String(take!(prog_buffer))
      # ProgressMeter likes to add a trailing '\n' once finished -- remove it
      prog_str = chomp(prog_str)

      # construct report
      TimerOutputs.complement!(to)
      println(report_buffer, to)
      println(report_buffer, prog_str)

      if verbose
        # move cursor in stdout back up so that report can overwrite previous one
        if nlines >= 0
          print(stdout, "\r\u1b[K\u1b[A" ^ nlines)
        end

        # report stuff
        seekstart(report_buffer)
        for line in eachline(report_buffer)
          println(stdout, line)
        end
      end

      # find number of lines which we need to overwrite on next pass
      seekstart(report_buffer)
      # make sure to overwrite all of previous content
      nlines = max(countlines(report_buffer), nlines)
      seekstart(report_buffer)

      # clear buffer
      take!(report_buffer)

    end

    timing = CallbackTiming(every_iteration=1)
    fcb = FunctionCallback(logfn, timing)

    return new(fcb)
  end
end


(pcb::ProgressCallback)(u, t, iter; force=false) = pcb.fcb(u, t, iter; force)


function state(pcb::ProgressCallback)
  """
  ProgressCallback
  """
end


Base.show(io::IO, pcb::ProgressCallback) = print(io, state(pcb))
