"""
  Evolution(rhs_fn!, u0::Vector{Float64}, timestep, tend, alg::TimeStepAlgorithm;
            cfl = 1.0,
            callback_fullstep::Union{Function,CallbackSet,Nothing}=nothing,
            callback_substep::Union{Function,CallbackSet,Nothing}=nothing)

Use to set up an evolution problem described by
- a `rhs_fn!(u, t, i)` that computes the residual inplace of `u` for time `t`. `i` is
the current iteration index. The use of `i` is deprecated.
- initial data `u0`
- a `timestep`: this can be
  - a numerical value
  - a function `timepstep(u, t, i)` that returns the timestep for a given state `u` and time `t`.
    Again, `i` is the iteration index and its usage is deprecated.
- final time `tend`
- time stepping algorithm `alg`; see the function `algorithm`

Optional values
- `cfl` is a global fudge factor to control the time step size; we require `cfl ≥ 0`
- `callback_fullstep` to run a set of functions after a full time step; possible values
  - `callback_fullstep = nothing` is default
  - `callback_fullstep <: CallbackSet` to run a set of functions in a specified order;
    see the documentation of the callback interface and `CallbackSet`
  - `callback_fullstep <: Function` to run a simple callback; the signature is
    `callback_fullstep(u, t, j)` where `u,t` are the state and time after the time step is taken;
    `j` is the number of retries for this timestep and its usage is deprecated
- `callback_substep` to run a set of function after each explicit time step; see
  `callback_fullstep` for allowed values
"""
function Evolution(rhs_fn!, u0::Vector{Float64}, timestep, tend, alg::TimeStepAlgorithm;
    cfl = 1.0,
    callback_fullstep::Union{Function,CallbackSet,Nothing}=nothing,
    callback_substep::Union{Function,CallbackSet,Nothing}=nothing)

  # enforce interface
  timestep_fn = if timestep isa Number
    (u,t,i) -> timestep
  else
    if !(length(methods(timestep, [Any,Real,Int])) > 0)
      throw(ArgumentError("require tstep to take arguments (u,t,n_attempts)"))
    end
    timestep
  end
  cb_fullstep = if !(callback_fullstep isa Function) &&
      ( isnothing(callback_fullstep) || length(callback_fullstep.callbacks) == 0 )
    (u,t,n; force=false) -> true
  else
    if !(length(methods(callback_fullstep, [Any,Real,Int])) > 0)
      throw(ArgumentError("require callback_fullstep to take arguments (u,t,n_attempts)"))
    end
    callback_fullstep
  end
  cb_substep = if !(callback_substep isa Function) &&
      ( isnothing(callback_substep) || length(callback_substep.callbacks) == 0 )
    (u,t,n; force=false) -> true
  else
    if !(length(methods(callback_substep, [Any,Real,Int])) > 0)
      throw(ArgumentError("require callback_substep to take arguments (u,t,n_attempts)"))
    end
    callback_substep
  end
  if cfl <= 0
    throw(ArgumentError("require cfl ≥ 0"))
  end

  stages = make_RK_stages(size(u0), alg.nstages)
  stepper!(up1, u0, t, dt) = step!(up1, rhs_fn!, u0, t, dt, stages, alg, cb_substep)
  up1 = deepcopy(u0) # we need to initialize memory anyways
                     # and the stepper will swap up1, u before he steps

  return Evolution{typeof(rhs_fn!), typeof(timestep_fn), typeof(cb_fullstep),
                   typeof(cb_substep), typeof(alg), typeof(stepper!)}(
        rhs_fn!, timestep_fn, cb_fullstep, cb_substep, tend, cfl, alg,
        stepper!,
        up1, u0, 0.0, 0, stages)
end


done(evo::Evolution) = evo.t >= evo.tend


function step!(evo::Evolution)

  # swap states so that result from previous step is now initial data
  evo.up1, evo.u = evo.u, evo.up1

  @unpack up1, u, t, tend, timestep, stepper!, cb_fullstep = evo

  # TODO Remove cfl from all projects and enable it here
  # dt = evo.cfl * timestep(u, t, 0)
  dt = timestep(u, t, 0)
  if isnothing(dt)
    println()
    @info """Termination requested by timestep function!"""
    return :timestep_termination
  end

  # shorten time step to hit tend exactly
  if t + dt > tend
    dt = tend - t
  end

  if stepper!(up1, u, t, dt) == false
    @info """Evolution step failed at t = $(t)!"""
    return :stepper_failed
  end

  if isinvalid(up1)
    println()
    @warn """NaN or Inf encountered in evolution at t = $(t)!"""
    return :invalid_state
  end

  evo.t += dt
  evo.it += 1

  if cb_fullstep(up1, evo.t, evo.it) == false
    @info """Callback after step failed at t = $(t)!"""
    return :callback_failed
  end

  return true
end


function evolve(rhs!, u0, dt, tend, alg; kwargs...)
  evo = Evolution(rhs!, u0, dt, tend, alg; kwargs...)
  while !done(evo) && step!(evo) === true; end
  return evo.up1
end


#######################################################################
#                             Deprecated                              #
#######################################################################


struct ODEProblem{T}
  coeffs::T
end

ODEProblem(prms::AbstractDict) = ODEProblem(prms["method"])
ODEProblem(method::String) = ODEProblem(Symbol(method))

function ODEProblem(method::Symbol)

  coeffs = if method == :midpt
    BT_MIDPT()
  elseif  method == :rk4
    BT_RK4()
  elseif  method == :rk4_twothirds
    BT_RK4_twothirds()
  elseif  method == :rk4_ralston
    BT_RK4_Ralston()
  elseif  method == :ssprk3
    BT_SSPRK3()
  elseif  method == :lserk4
    COEFFS_LSERK4()
  elseif  method == :rkf10
    BT_RKF10()
  else
    error("Requested ERK scheme not implemented!")
  end

  return ODEProblem{typeof(coeffs)}(coeffs)
end


#######################################################################
#                          Butcher-Tablaeus                           #
#######################################################################

function algorithm(method::Symbol)::TimeStepAlgorithm

  alg = if method == :midpt
    BT_MIDPT()
  elseif  method == :rk4
    BT_RK4()
  elseif  method == :rk4_twothirds
    BT_RK4_twothirds()
  elseif  method == :rk4_ralston
    BT_RK4_Ralston()
  elseif  method == :ssprk3
    BT_SSPRK3()
  elseif  method == :lserk4
    COEFFS_LSERK4()
  elseif  method == :rkf10
    BT_RKF10()
  else
    error("Requested ERK scheme not implemented!")
  end

  return alg
end


#######################################################################
#                              Steppers                               #
#######################################################################


"""
    step!(u!, F, u0, t, dt, tmp_k, coeffs::COEFFS_LSERK4, callback_substep)

Stepper for Low Storage five-stage fourth-order Explicit Runge Kutta scheme.
Taken from Hesthaven, Warburton (2007), section 3.4.
"""
function step!(u!, F, u0, t, dt, tmp_k, coeffs::COEFFS_LSERK4, callback_substep)

  @unpack nstages, a, b, c = coeffs

  kim1, ki = tmp_k
  N = length(u!)

  for i = 1:nstages

    ti = t + c[i] * dt
    if i == 1
      F(ki, u0, ti)
    else
      F(ki, u!, ti)
    end

    @. ki *= dt
    if i == 1
      @. u! = u0 + b[i] * ki
    else
      ai = a[i]
      @. ki += ai * kim1
      @. u! += b[i] * ki
    end

    success = callback_substep(u!, ti, i+1) # +1 because u! is value for next stage
    !success && return false

    if i < nstages
      @. kim1 = ki
    end

  end # for

  return true
end


"""
    step!(u!, F, u0, t, dt, tmp_k, coeffs, callback_substep)

Stepper for Explicit Runge Kutta scheme.
"""
function step!(u!, F, u0, t, dt, tmp_k, coeffs::TimeStepAlgorithm, callback_substep)
  @unpack nstages, a, b, c = coeffs

  tn = t
  for n = 1:nstages
    tn = t + c[n] * dt
    u! .= u0
    for i = 1:n-1
      u! .+= a[n, i] * tmp_k[i]
    end
    if n > 1
      success = callback_substep(u!, tn, n)
      !success && return false
    end
    F(tmp_k[n], u!, tn)
    tmp_k[n] .*= dt
  end

  u! .= u0
  for n = 1:nstages
    u! .+= b[n] * tmp_k[n]
  end

  success = callback_substep(u!, tn, nstages)

  return true
end


"""
    make_RK_stages(size, n)

Generate `n` stages of length `size`. To be used with with `step!` as temporary storage
for substeps.
"""
function make_RK_stages(size, n)
  return [ zeros(Float64, size) for _ in 1:n ]
end


"""
    isinvalid(u)

Return false if any component of `u` is either `NaN` or `Inf`.
"""
function isinvalid(u)
  for ui in u
    ( isnan(ui) || isinf(ui) ) && return true
  end
  return false
end


"""
    evolve_ERK(pblm::ODEProblem, F, u0, tstep::Union{Function,Real}, tend;
               callback_fullstep::Union{Function,Nothing}=nothing,
               callback_substep::Union{Function,Nothing}=nothing)

Use an Explicit-Runge-Kutta scheme to evolve a function `F` from initial state `u0`
with time step `tstep` till `tend` using the scheme `scheme`.

Provide a method for `tstep(u,t,n_attempts)::Float64` to evolve with a state and time
dependent time step.
`u, t` are the current iterate and time, `n_attempts` is the number of attempts made
to retake the current time step. The latter is only needed to adjust the step in case
a substep failed, see `callback_substep` below.
Return `nothing` to terminate the evolution.

Provide `callback_fullstep(u, t, iter)` to run this function after every full step. Any return
value is ignored.

Provide `callback_substep(u, t, n_stage)::Bool` to run this function after every sub step.
The return value indicates success. Use `tstep(u,t,n_attempts)` to retry the full step
with a smaller step.
"""
function evolve_ERK(
    pblm::ODEProblem,
    F::T_F, u0, tstep::T_tstep, tend;
    callback_fullstep=nothing, callback_substep=nothing) where {T_F,T_tstep}

  # enforce interface
  @assert length(methods(tstep, [Any,Real,Int])) > 0 """
  `tstep` must take arguments `(u,t,n_attempts)`.
  """
  cb_fullstep = if isnothing(callback_fullstep)
    (u,t,n; force=false) -> true
  else
    @assert length(methods(callback_fullstep, [Any,Real,Int])) > 0 """
    `callback_fullstep` must take arguments `(u,t,n_attempts)`.
    """
    callback_fullstep
  end
  cb_substep = if isnothing(callback_substep)
    (u,t,n; force=false) -> true
  else
    @assert length(methods(callback_substep, [Any,Real,Int])) > 0 """
    `callback_substep` must take arguments `(u,t,n_attempts)`.
    """
    callback_substep
  end

  # setup storage
  un = deepcopy(u0)
  unp1 = zeros(Float64, length(un))
  tmp_k = make_RK_stages(size(u0), pblm.coeffs.nstages)
  stepper!(u!, F, u0, t, dt) = step!(u!, F, u0, t, dt, tmp_k, pblm.coeffs, cb_substep)

  tstart = 0.0
  tn = tstart
  iter = 0

  # invoke once before evolution to
  cb_fullstep(un, tn, iter; force=true)

  n_retake_attempts = 0
  dt = -1

  while tn < tend

    new_dt = if tstep isa Function
      new_dt = tstep(un, tn, n_retake_attempts)
      if isnothing(new_dt)
        println()
        @info """Termination requested by `tstep`!"""
        unp1 .= un # restore previous state
        break
      end
      new_dt
    else
      tstep
    end

    # shorten time step to hit tend exactly
    if tn + new_dt > tend
      new_dt = tend - tn
      # in this case it might happen that retaking did not reduce the step width enough
      # we then just try again
      if n_retake_attempts > 0 && new_dt >= dt
        n_retake_attempts += 1
        continue
      end
    end

    # is step admissible?
    if n_retake_attempts > 0 && new_dt >= dt
      println()
      @warn """Attempted to retake a step with too large of a step width:
               (dt, proposed dt) = ($dt, $new_dt)
               Terminating evolution ..."""
      unp1 .= un # restore previous state
      break
    end

    dt = new_dt

    # step now
    success = stepper!(unp1, F, un, tn, dt)
    if !success
      n_retake_attempts += 1
      continue
    else
      n_retake_attempts = 0
    end

    if isinvalid(unp1)
      println()
      @warn """NaN or Inf encountered in evolution at t = $(tn)!
                Terminating evolution ..."""
      unp1 .= un # restore previous state
      break
    end

    un .= unp1
    tn += dt
    iter += 1

    cb_fullstep(un, tn, iter)

  end # while

  return unp1
end


function evolve_ERK(pblm::ODEProblem, F, u0, tstep::Real, tend; kwargs...)
  return evolve_ERK(pblm, F, u0, (u,t,n)->tstep, tend; kwargs...)
end


@doc """
    evolve_adaptive_ERK(pblm::ODEProblem, F, u0, tstep, tend;
                        reltol=1e-10, abstol=1e-10, post_hook=nothing)

Similar to `evolve_ERK`, but iterates every ERK step till the difference between two
consecutive steps obeys the following criterion:
```
# un, unp1 are the results from the nth and (n+1)th iteration
norm(un - unp1, 1) <= reltol * norm(un, 1) + abstol
```
"""
function evolve_adaptive_ERK(
    pblm::ODEProblem, F, u0, tstep, tend;
    reltol=1e-10, abstol=1e-10, post_hook=nothing)

  un = deepcopy(u0); unp1 = similar(un);
  tmp_u1 = similar(unp1); tmp_u2 = similar(unp1);
  cb_substep = (u,t,n) -> true
  tmp_k = make_RK_stages(size(u0), pblm.coeffs.nstages)
  stepper!(u!, F, u0, t, dt) = step!(u!, F, u0, t, dt, tmp_k, pblm.coeffs, cb_substep)

  adaptive_max_its = 20 # maximum number of adaptive iterations

  tstart = 0.0
  tn = tstart
  while tn < tend

    if isa(tstep, Function)
      dt = tstep(un, tn)
    else
      dt = tstep
    end

    if tn + dt > tend
      dt = tend - tn
    end

    adapted_dt = dt
    norm_un = norm(un, 1)
    # estimate error by difference between a single RK step and
    # two consecutive RK step evolutions
    # iterate this until tolerance is reached
    adapted_it = 1
    while true
      stepper!(unp1,    F, un,      tn,                 adapted_dt)
      stepper!(tmp_u1,  F, un,      tn,                 adapted_dt*0.5)
      stepper!(tmp_u2,  F, tmp_u1,  tn+adapted_dt*0.5,  adapted_dt*0.5)

      if isinvalid(unp1) || isinvalid(tmp_u1) || isinvalid(tmp_u2)
        println()
        error("NaN or Inf encountered in evolution at t = $(tn)! Aborting ...")
      end

      @. tmp_u1 = unp1 - tmp_u2 # reuse tmp_u1 storage for error computation
      err = norm(tmp_u1, 1)
      err <= reltol * norm_un + abstol && break

      adapted_dt *= 0.5
      adapted_it += 1
      if adapted_it > adaptive_max_its
      # if log(dt/adapted_dt)/log(2) > adaptive_max_its
        println()
        error("Could not determine solution after $adaptive_max_its refinement iterations! " *
              "Aborting ... ")
      end
    end

    un .= unp1
    tn += adapted_dt

    if !isnothing(post_hook)
      success::Bool = post_hook(un, tn)
      if !success
        println()
        @info "Termination requested by post_hook!"
        break
      end
    end

  end # while

  return unp1
end
