#######################################################################
#                                Types                                #
#######################################################################


abstract type AbstractMesh end


struct Mesh{N_Dim,N_Sides} <: AbstractMesh
  tree::Tree{N_Dim,N_Sides} # abstract representation of mesh
  boxes::Vector{Box{N_Dim}} # physical extends of each cell
  extends::NTuple{N_Dim,Tuple{Float64,Float64}} # total extend of mesh
  element::SpectralElement
  cache::Cache
  bulkfaceindices::Vector{Int64}
  bulkbdryindices::Vector{Int64}
  faceindices::Vector{Int64}
  bdryindices::Vector{Int64}
  offsets::Vector{Int64} # data index offsets for each cell
  # TODO Move to evolution part!
  CFL::Float64
end


const Mesh1d = Mesh{1,2}
const Mesh2d = Mesh{2,4}


#######################################################################
#                            Constructors                             #
#######################################################################


function Mesh1d(; N=5, K=10, range=[-1.0,1.0], CFL=0.5, basis="lgl_lumped", periodic=(true,))

  @toggled_assert N > 0
  @toggled_assert K > 0
  @toggled_assert CFL > 0
  @toggled_assert length(range) == 2
  @toggled_assert range[1] < range[2]
  @toggled_assert basis in ["lgl", "glgl", "lgl_lumped"]

  # TODO Remove this after updating parameter names in SpectralElement
  quadrature_method = Dict("lgl" => :LGL, "glgl" => :GLGL, "lgl_lumped" => :LGL_lumped)[basis]

  tree = Tree1d(K,periodic=tuple(periodic...))
  boxes = place_boxes(tree, Float64.(range))
  element = SpectralElement(N, Symbol(quadrature_method))
  @unpack Npts, z = element
  offsets = [ (i-1)*Npts for i = 1:length(tree) ]
  extends = (Tuple(range),)

  cache = Cache(K*Npts, 2*K, K)
  register_variables!(cache, static_variablenames=[:x,:invdetJ],
                             bdry_variablenames=[:nx])
  @unpack x, invdetJ = get_static_variables(cache)
  @unpack nx = get_bdry_variables(cache)

  for (i,b) in enumerate(boxes)
    (xmin, xmax), = extrema(b)
    xi = @. 0.5 * ((xmax - xmin) * z + (xmax + xmin))
    rng = (i-1)*Npts+1:i*Npts
    @views x[rng] .= xi
    @views invdetJ[rng] .= 2.0 / (xmax - xmin)

    # surface normals
    nx[2*(i-1)+1] = -1.0
    nx[2*(i-1)+2] = 1.0
  end

  bulkfaceindices = Int64[]
  for k in 1:K
    append!(bulkfaceindices, ((k-1)*Npts+1, (k-1)*Npts+Npts) )
  end
  faceindices = collect(1:2*K)
  bulkbdryindices, bdryindices = if periodic[1]
    Int64[], Int64[]
  else
    popfirst!(faceindices)
    pop!(faceindices)
    popfirst!(bulkfaceindices)
    pop!(bulkfaceindices)
    Int64[1, Npts*K], Int64[1, 2*K]
  end

  return Mesh1d(tree, boxes, extends, element, cache,
                bulkfaceindices, bulkbdryindices, faceindices, bdryindices, offsets, CFL)
end


function Mesh2d(; N=[5,5], K=[4,4], range=[-1.0,1.0, -1.0,1.0],
                  CFL=0.5, basis="lgl_lumped", periodic=(true,true))

  @toggled_assert length(N) == 2
  @toggled_assert length(K) == 2
  @toggled_assert length(range) == 4
  Nx, Ny = N
  Kx, Ky = K
  xrange, yrange = range[1:2], range[3:4]

  @toggled_assert Nx > 0
  @toggled_assert Ny > 0
  @toggled_assert Kx > 0
  @toggled_assert Ky > 0
  @toggled_assert CFL > 0
  @toggled_assert xrange[1] < xrange[2]
  @toggled_assert yrange[1] < yrange[2]
  @toggled_assert basis in ["lgl", "glgl", "lgl_lumped"]

  # TODO Remove this after updating parameter names in SpectralElement
  quadrature_method = Dict("lgl" => :LGL, "glgl" => :GLGL, "lgl_lumped" => :LGL_lumped)[basis]

  tree = Tree2d(Kx,Ky,periodic=tuple(periodic...))
  boxes = place_boxes(tree, Float64.(xrange), Float64.(yrange))
  element_x = SpectralElement(Nx, Symbol(quadrature_method))
  element_y = SpectralElement(Ny, Symbol(quadrature_method))
  Nptsx, Nptsy = element_x.Npts, element_y.Npts
  offsets = [ (i-1)*Nptsx*Nptsy for i = 1:length(tree) ]
  extends = (Tuple(xrange),Tuple(yrange))

  Nptsbulk = Kx*Nptsx * Ky*Nptsy
  Nptsbdry = 2*(Nptsx+Nptsy) * Kx*Ky
  Ncells = n_cells(tree)
  cache = Cache(Nptsbulk, Nptsbdry, Ncells)
  register_variables!(cache, static_variablenames=[:x,:y,:dxdX,:dydY,:invdetJ],
                             bdry_variablenames=[:nx,:ny])
  @unpack x, y, dxdX, dydY, invdetJ = get_static_variables(cache)
  # outward facing normals on rectangular grid
  @unpack nx, ny = get_bdry_variables(cache)

  for (i,b) in enumerate(boxes)
    (xmin, xmax), (ymin,ymax) = extrema(b)
    xi = @. 0.5 * ((xmax - xmin) * element_x.z + (xmax + xmin))
    yi = @. 0.5 * ((ymax - ymin) * element_y.z + (ymax + ymin))
    xx = repeat(xi, outer=Nptsy)
    yy = repeat(yi, inner=Nptsx)
    offset = (i-1)*Nptsx*Nptsy
    rng = offset+1:offset+Nptsx*Nptsy
    @views x[rng] .= xx
    @views y[rng] .= yy
    @views dxdX[rng]    .= 2.0 / (xmax - xmin)
    @views dydY[rng]    .= 2.0 / (ymax - ymin)
    @views invdetJ[rng] .= 2.0 / (xmax - xmin) * 2.0 / (ymax - ymin)

    # surface normals
    offset_bdry = (i-1) * 2 * (Nptsx + Nptsy)
    rng_lhs  = offset_bdry+1:offset_bdry+Nptsy
    rng_rhs  = rng_lhs  .+ Nptsy
    rng_down = rng_rhs  .+ Nptsx
    rng_up   = rng_down .+ Nptsx
    @views nx[rng_lhs]  .= -1.0
    @views ny[rng_lhs]  .= 0.0
    @views nx[rng_rhs]  .= 1.0
    @views ny[rng_rhs]  .= 0.0
    @views nx[rng_down] .= 0.0
    @views ny[rng_down] .= -1.0
    @views nx[rng_up]   .= 0.0
    @views ny[rng_up]   .= 1.0
  end

  faceindices, bdryindices = Int64[], Int64[]
  bulkfaceindices, bulkbdryindices = Int64[], Int64[]
  bo = 0 # boundary offset
  for (o,cell) in zip(offsets,tree.cells)
    LI = LinearIndices((Nptsx,Nptsy))
    # TODO This indexing order agrees with how it is currently done in
    # broadcast_faces_2! I didn't realize, but with this convention
    # the x direction is not the 'fastest' to iterate.
    # Synchronize this with however the cell layout is done in the Tree.
    # Ideally the layout should be such that when we have periodic bdry conditions
    # faceindices is just the same as 1:length(mesh).
    if has_neighbor(cell, Cart2d.Xmin)
      append!(bulkfaceindices, LI[1,:] .+ o)
      append!(faceindices, (1:Nptsy) .+ bo)
    else
      append!(bulkbdryindices, LI[1,:] .+ o)
      append!(bdryindices, (1:Nptsy) .+ bo)
    end
    bo += Nptsy
    if has_neighbor(cell, Cart2d.Xmax)
      append!(bulkfaceindices, LI[end,:] .+ o)
      append!(faceindices, (1:Nptsy) .+ bo)
    else
      append!(bulkbdryindices, LI[end,:] .+ o)
      append!(bdryindices, (1:Nptsy) .+ bo)
    end
    bo += Nptsy
    if has_neighbor(cell, Cart2d.Ymin)
      append!(bulkfaceindices, LI[:,1] .+ o)
      append!(faceindices, (1:Nptsx) .+ bo)
    else
      append!(bulkbdryindices, LI[:,1] .+ o)
      append!(bdryindices, (1:Nptsx) .+ bo)
    end
    bo += Nptsx
    if has_neighbor(cell, Cart2d.Ymax)
      append!(bulkfaceindices, LI[:,end] .+ o)
      append!(faceindices, (1:Nptsx) .+ bo)
    else
      append!(bulkbdryindices, LI[:,end] .+ o)
      append!(bdryindices, (1:Nptsx) .+ bo)
    end
    bo += Nptsx
  end

  return Mesh2d(tree, boxes, extends, element_x, cache,
                bulkfaceindices, bulkbdryindices, faceindices, bdryindices, offsets, CFL)
end


# TODO Deprecate and remove this
function Mesh(; N, K, range, CFL, basis, periodic)
  dims = Int(length(range)/2)
  MType = dims == 1 ? Mesh1d : Mesh2d
  return MType(; N, K, range, CFL, basis, periodic)
end


#######################################################################
#                               Getters                               #
#######################################################################


# Overloading getters here to restore previous behavior
# TODO Gradually remove this!
function Base.getproperty(m::Mesh1d, p::Symbol)
  if p === :x1
    cache = getfield(m, :cache)
    return get_static_variables(cache).x
  elseif p === :K
    tree = getfield(m, :tree)
    return n_cells(tree)
  elseif p === :xmin
    return extrema(m)[1][1]
  elseif p === :xmax
    return extrema(m)[1][2]
  elseif p === :L
    return widths(m)[1]
  elseif p === :dl
    return widths(m)[1] / m.K
  elseif p === :x
    x = m.x1
    K = m.K
    Npts = getfield(m, :element).Npts
    return reshape(x, (Npts,K))
  elseif p === :invjac
    Npts, K = layout(m)
    cache = getfield(m, :cache)
    return reshape(get_static_variables(cache).invdetJ, (Npts,K))
  else
    return getfield(m, p)
  end
end

function Base.getproperty(m::Mesh2d, p::Symbol)
  if p === :x
    cache = getfield(m, :cache)
    return get_static_variables(cache).x
  elseif p === :y
    cache = getfield(m, :cache)
    return get_static_variables(cache).y
  else
    return getfield(m, p)
  end
end

#######################################################################
#                               Methods                               #
#######################################################################


function widths(m::Mesh{N}) where N
  return NTuple{N,Float64}(abs(r-l) for (l,r) in m.extends)
end


Base.broadcastable(mesh::Mesh) = Ref(mesh)


"""
    extract_cell_bdrys(mesh::Mesh)

Return coordinates of cell boundaries from a `mesh`.
"""
function extract_cell_bdrys(mesh::Mesh)
  @unpack x, K = mesh
  @unpack Npts = mesh.element
  xbs = Float64[]
  for k = 1:K
    push!(xbs, x[(k-1)*Npts+1])
  end
  push!(xbs, x[end])
  return xbs
end


"""
  grid_average(u, mesh::Mesh)

Return the grid average of all cell averages
"""
function grid_average(u, mesh::Mesh)
  @unpack element, dl, L = mesh
  cellavg = cellwise_average(u, mesh) .* dl
  avg = sum(cellavg) / L
  return avg
end


"""
  cellwise_average(u, mesh::Mesh)

Return the cellwise average of `u` over `mesh.x`.
"""
function cellwise_average(u, mesh::Mesh)
  @unpack element, invjac, K, dl = mesh
  @unpack Npts = element
  ru = vreshape(u, (Npts, K))
  avg = zeros(K)
  @views for i in 1:K
    avg[i] = integrate(ru[:,i], element)
  end
  avg ./= invjac[1,:] * dl # constant Jacobian, constant cell width
  return avg
end


function inner_product(u, v, k, mesh::Mesh)
  @unpack element, invjac, K = mesh
  @unpack Npts = element
  ru = vreshape(u, (Npts,K))
  rv = vreshape(v, (Npts,K))
  @views iprod = inner_product(ru[:,k], rv[:,k], element)
  iprod /= invjac[1,k]
  return iprod
end


function integrate_cell(u, k, mesh::Mesh1d)
  @unpack element = mesh
  @unpack invdetJ = get_static_variables(mesh.cache)
  idxs = cellindices(mesh, k)
  uk = view(u, idxs)
  res = integrate(uk, element) / invjac[first(idxs)]
  return res
end


function integrate_cell(u, k, mesh::Mesh2d)
  @unpack element = mesh
  @unpack dxdX, dydY = get_static_variables(mesh.cache)
  @unpack Npts = element
  idxs = cellindices(mesh, k)
  uk = view(u, idxs)
  dxdXk = view(dxdX, idxs)
  dydYk = view(dydY, idxs)
  buffer = zeros(Float64, Npts)
  for (k,col) in enumerate(eachcol(uk))
    buffer[k] = integrate(col, element) / dxdXk[1,k]
  end
  res = integrate(buffer, element) / dydYk[1,1]
  return res
end


"""
  cellwise_inner_product(u, v, mesh::Mesh)

Return vector of inner products between `u, v` on every cell in `mesh`.
"""
function cellwise_inner_product(u, v, mesh::Mesh)
  @unpack element, K, invjac = mesh
  @unpack Npts = element
  L = layout(mesh)
  ru = vreshape(u, L)
  rv = vreshape(v, L)
  inner_products = [ @views( inner_product(ru[:,k], rv[:,k], element) ) for k = 1:K ]
  inner_products ./= invjac[1,:] # constant Jacobian
  return inner_products
end


"""
  broken_inner_product(u, v, mesh::Mesh)

Return sum of all inner products between `u,v` on every cell in `mesh`.
"""
function broken_inner_product(u, v, mesh::Mesh)
  inner_products = cellwise_inner_product(u, v, mesh)
  return sum(inner_products)
end


function LinearAlgebra.norm(u, mesh::Mesh, order)
  if order == 1
    return norm_L1(u, mesh)
  elseif order == 2
    return norm_L2(u, mesh)
  else
    error("Norm for order = $order not implemented.")
  end
end
norm_L2(u, mesh::Mesh) = sqrt(broken_inner_product(u, u, mesh))
# TODO Refine broken_inner_product and other mesh quadrature methods to also accept
# a function that can be applied somewhere in the underlying integrate function before
# the integrand is evaluated. This should help avoid making a copy of the array here.
norm_L1(u, mesh::Mesh) = broken_inner_product(abs.(u), ones(Float64, size(u)), mesh)


"""
    find_nearest(xi, mesh::Mesh)

Find index of node in `mesh` that lies closest to `xi`.
Returns the index of the outermost point if `xi` falls outside `mesh`.
"""
function find_nearest(xi, mesh::Mesh)
  @unpack x = mesh
  _, idx = findmin(xx -> abs(xx - xi), view(x,:))
  if isnothing(idx)
    return xi < mesh.xmin ? 1 : length(mesh)
  end
  return idx
end


@deprecate locate_point(xi, mesh::Mesh) find_nearest(xi, mesh::Mesh)


"""
    find_cell_index(xi, mesh::Mesh)

Find the cell index into which coordinate `xi` falls into the `mesh`.
Note that it returns the first index found in case `xi` falls onto an interface boundary.
Returns nothing if `xi` lies outside the `mesh`.
"""
function find_cell_index(xi, mesh::Mesh)
  cell_bdrys = extract_cell_bdrys(mesh)
  cell_index = findfirst(idx -> cell_bdrys[idx] <= xi <= cell_bdrys[idx+1], 1:length(cell_bdrys)-1)
  return cell_index
end

function find_cell_index(pointidx::Int64, mesh::Mesh)
  TODO()
end
function find_cell_index(pointidx::Int64, mesh::Mesh2d)
  @toggled_assert 1 <= pointidx <= length(mesh)
  idx = findfirst(>=(pointidx), mesh.offsets)
  return isnothing(idx) ? length(mesh.offsets) : idx
end


Base.length(mesh::Mesh)::Int64 = length(get_static_variables(mesh.cache).x)
Base.extrema(mesh::Mesh) = mesh.extends
npoints(mesh::Mesh)::Int = mesh.element.Npts
ncells(mesh::Mesh) = prod(mesh.tree.dims)
layout(mesh::Mesh)::Tuple{Int,Int} = (n_points(mesh), n_cells(mesh))
dimension(mesh::Mesh{N}) where N = N
cache(mesh::Mesh) = mesh.cache

@deprecate n_points(mesh::Mesh) npoints(mesh)
@deprecate n_cells(mesh::Mesh) ncells(mesh)

"""
    MeshInterpolator


Operator for interpolating data from one mesh to another.

# Examples

```julia
julia> sample_mesh, reference_mesh = Mesh(; N=3, K=4); Mesh(; N=10, K=20);

julia> reference_data = sin.(reference_mesh.x);

julia> meshintrp = MeshInterpolator(sample_mesh, reference_mesh);

julia> interp_data = meshintrp(reference_data)

julia> meshintrp(intrp_data, reference_data) # result is stored inside intrp_data
```

Note that for inplace interpolation `intrp_data` and `reference_data` cannot be aliased.
"""
struct MeshInterpolator
  reference_mesh::Mesh1d
  sample_mesh::Mesh1d
  interpolation_matrices::Vector{Tuple{Int64,Matrix{Float64}}}
  reference_cell_indices::Vector{Int64}
end


function MeshInterpolator(smpl_mesh::Mesh1d, ref_mesh::Mesh1d)

  (smpl_xmin, smpl_xmax), = extrema(smpl_mesh)
  (ref_xmin, ref_xmax),   = extrema(ref_mesh)
  if !(ref_xmin <= smpl_xmin <= ref_xmax) || !(ref_xmin <= smpl_xmax <= ref_xmax)
    error("Sample mesh must fall into reference mesh")
  end

  smpl_x = vreshape(smpl_mesh.x, layout(smpl_mesh))
  ref_x  = vreshape(ref_mesh.x, layout(ref_mesh))

  ref_xbdrys       = extract_cell_bdrys(ref_mesh)
  ref_cell_indices = find_cell_index.(smpl_x, Ref(ref_mesh))

  interpolation_matrices = Tuple{Int64, Matrix{Float64}}[]
  current_idx = 1 # first index of sample node that falls into a new cell in the reference mesh
  while true
    cell_idx  = ref_cell_indices[current_idx]
    # find all sample indices of nodes which fall into the same reference cell
    same_idxs = findall(ref_idx -> ref_idx == cell_idx, ref_cell_indices)
    smpl_idxs = [ current_idx ]
    if length(same_idxs) == 1
      append!(smpl_idxs, same_idxs[2:end])
    end
    # extract the sample nodes
    smpl_xs   = smpl_x[smpl_idxs]
    # generate interpolation matrix
    push!(interpolation_matrices,
          (cell_idx, dg1d.lagrange_interp_matrix(smpl_xs, ref_x[:,cell_idx])))

    current_idx += length(smpl_idxs)
    if current_idx > length(ref_cell_indices)
      break
    end
  end

  return MeshInterpolator(ref_mesh, smpl_mesh, interpolation_matrices, ref_cell_indices[:])
end


function (intrp::MeshInterpolator)(ref_data)
  result = zeros(layout(intrp.sample_mesh))
  intrp(result, ref_data)
  return result
end


function (intrp::MeshInterpolator)(storage, ref_data)

  if length(ref_data) != length(intrp.reference_mesh)
    error("Provided reference data does not match with reference mesh.")
  end
  if length(storage) != length(intrp.sample_mesh)
    error("Provided storage for interpolated result does not match with sample mesh.")
  end

  vec_intrp_data = view(storage, :)
  mat_ref_data   = vreshape(ref_data, layout(intrp.reference_mesh))
  current_idx = 1 # first index of sample node that falls into a new cell in the reference mesh
  @views for (ref_cell_idx, intrp_mat) in intrp.interpolation_matrices
    n_intrp_data = size(intrp_mat)[1]
    idx_range = current_idx:current_idx+(n_intrp_data-1)
    vec_intrp_data[idx_range] .= intrp_mat * mat_ref_data[:, ref_cell_idx]
    current_idx += n_intrp_data
  end

  return
end


struct CellDataIterator{M<:AbstractMesh,A<:AbstractArray}
  mesh::M
  data::A
  function CellDataIterator(m::AbstractMesh, a::AbstractArray)
    @toggled_assert length(m) == length(a)
    return new{typeof(m),typeof(a)}(m,a)
  end
end

Base.length(it::CellDataIterator) = ncells(it.mesh)
Base.size(it::CellDataIterator) = size(it.mesh)

@inline function Base.iterate(it::CellDataIterator{Mesh1d}, state=1)
  state > length(it) && return nothing
  idx = it.mesh.offsets[state]
  Npts = it.mesh.element.Npts
  v = view(it.data, idx+1:idx+Npts)
  return v, state+1
end
@inline function Base.iterate(it::CellDataIterator{Mesh2d}, state=1)
  state > length(it) && return nothing
  idx = it.mesh.offsets[state]
  Npts = it.mesh.element.Npts
  v = view(it.data, idx+1:idx+Npts^2)
  rv = reshape(v, (Npts,Npts))
  return rv, state+1
end

# eachcell(mesh::AbstractMesh) = CellIterator(mesh)
eachcell(mesh::AbstractMesh, data) = CellDataIterator(mesh, data)

@inline function flatcellindices(mesh::Mesh1d, idx)
  Nx = mesh.element.Npts
  offset = mesh.offsets[idx]
  return offset+1:offset+Nx
end
@inline function flatcellindices(mesh::Mesh2d, idx)
  Nx = mesh.element.Npts
  Ny = Nx
  offset = mesh.offsets[idx]
  return offset+1:offset+Nx*Ny
end

@inline function cellindices(mesh::Mesh1d, idx)
  Nx = mesh.element.Npts
  offset = mesh.offsets[idx]
  return offset+1:offset+Nx
end
@inline function cellindices(mesh::Mesh2d, idx)
  Nx = mesh.element.Npts
  Ny = Nx
  offset = mesh.offsets[idx]
  return reshape(offset+1:offset+Nx*Ny, (Nx,Ny))
end

@inline function faceindices(mesh::Mesh1d, idx)
  offset = 2*(idx-1)
  return (offset+1:offset+1,offset+2:offset+2)
end
@inline function faceindices(mesh::Mesh2d, idx)
  Nx = mesh.element.Npts
  Ny = Nx
  offset = (idx-1)*2*(Nx+Ny)
  lhs  = offset+1:offset+Ny
  rhs  = lhs .+ Ny
  down = rhs .+ Nx
  up   = down .+ Nx
  return lhs, rhs, down, up
end


# need dummy type to enable iterating over volume and we don't care about the boundary indices
struct SkipIndex end
Base.eachindex(rng, ::SkipIndex) = eachindex(rng)

# define getindex for SkipIndex itself to return a SkipIndex
# we also define getindex on generic containers and SkipIndex so that
# we can throw a meaningful error in case one attempts to use a bdry variable over volume
Base.getindex(s::SkipIndex, i) = s
Base.getindex(v, s::SkipIndex) =
  error("broadcast_volume!: attempted to use a 'bdry' variable when iterating over volume")


function volumeindices(mesh::AbstractMesh)
  return (1:length(mesh), SkipIndex())
end
function faceindices(mesh::AbstractMesh)
  return (mesh.bulkfaceindices, mesh.faceindices)
end
function bdryindices(mesh::AbstractMesh)
  return (mesh.bulkbdryindices, mesh.bdryindices)
end


interpolate_face_data!(mesh, vars) = interpolate_face_data.(mesh, vars)
function interpolate_face_data!(mesh::Mesh1d, var::AbstractVector, bdrybuf::AbstractVector)
  bo = 0 # boundary data offset
  @inbounds for cell in mesh.tree.cells
    if has_neighbor(cell, Cart1d.Xmin)
      i = cell.neighbors[Cart1d.Xmin]
      idxs = cellindices(mesh,i)
      v = view(var, idxs)
      bdrybuf[bo+1] = v[end]
    end
    bo += 1
    if has_neighbor(cell, Cart1d.Xmax)
      i = cell.neighbors[Cart1d.Xmax]
      idxs = cellindices(mesh,i)
      v = view(var, idxs)
      bdrybuf[bo+1] = v[1]
    end
    bo += 1
  end
end
function interpolate_face_data!(mesh::Mesh2d, var::AbstractVector, bdrybuf::AbstractVector)
  Nptsx, Nptsy = mesh.element.Npts, mesh.element.Npts
  bo = 0 # boundary data offset
  @inbounds for cell in mesh.tree.cells
    if has_neighbor(cell, Cart2d.Xmin)
      i = cell.neighbors[Cart2d.Xmin]
      idxs = cellindices(mesh,i)
      v = view(var, idxs)
      for j in 1:Nptsy
        bdrybuf[bo+j] = v[end,j]
      end
    end
    bo += Nptsy
    if has_neighbor(cell, Cart2d.Xmax)
      i = cell.neighbors[Cart2d.Xmax]
      idxs = cellindices(mesh,i)
      v = view(var, idxs)
      for j in 1:Nptsy
        bdrybuf[bo+j] = v[1,j]
      end
    end
    bo += Nptsy
    if has_neighbor(cell, Cart2d.Ymin)
      i = cell.neighbors[Cart2d.Ymin]
      idxs = cellindices(mesh,i)
      v = view(var, idxs)
      for j in 1:Nptsx
        bdrybuf[bo+j] = v[j,end]
      end
    end
    bo += Nptsx
    if has_neighbor(cell, Cart2d.Ymax)
      i = cell.neighbors[Cart2d.Ymax]
      idxs = cellindices(mesh,i)
      v = view(var, idxs)
      for j in 1:Nptsx
        bdrybuf[bo+j] = v[j,1]
      end
    end
    bo += Nptsx
  end
end
