"""
    LLF(nfl,nfr,ul,ur,v)

Local Lax-Friedrich flux:

  nf* = 1/2 ( f(ul) + f(ur) - v * (ur - ul) )
"""
@inline LLF(nfl,nfr,ul,ur,v) = 1/2 * (nfl + nfr - v * (ur - ul))


"""
    estimate_convergence_rate(n1, n2, f1, f2)

Experimental rate of convergence based on resolutions `n1, n2` and data `f1, f2`.
"""
function estimate_convergence_rate(n1, n2, f1, f2)
  return log10(f1/f2) / log10(n2/n1)
end


"""
    purge_zeros!(M, thrs=1e-13)

Scan `M` for elements smaller than `thrs` and set them to 0.
"""
purge_zeros!(M, thrs=1e-13) = M[abs.(M).<thrs] .= 0.0


"""
    least_sqr_linear(x, y)

Compute a linear least squares fit for data `x, y`.
Returns fit function `f, a, b`, where `f(x) = a + b * x` and `a,b` are the fitting coefficients.
Alogrithm taken from https://mathworld.wolfram.com/LeastSquaresFitting.html
"""
function least_sqr_linear(x, y)

  n = length(x)
  @assert n == length(y) "Mismatch in number of data points x, y!"

  yx = dot(y, x)
  xx = dot(x, x)
  mean_y = sum(y) / n
  mean_x = sum(x) / n

  ss_xx = xx - mean_x^2 * n
  ss_xy = yx - mean_x * mean_y * n

  b = ss_xy / ss_xx
  a = mean_y - b * mean_x

  f(x) = a + b * x
  return f, a, b
end


stepdata(x, x0, fL, fR) = [ (xi < x0) ? fL : fR for xi in x ]


"""
    periodize(i, N)

Wrap `i` around period `N` if `i > N`, otherwise return `i`.
Behaves similar when called with `Int`s or `Real`s.
"""
@inline function periodize(i::T, N::S) where {T<:Int,S<:Int}
  mod = i % N
  return mod <= 0 ? N+mod : mod
end
@inline function periodize(x::T, L::S) where {T<:Real,S<:Real} 
  xround = floor(x/L) * L
  return x <= xround ? x : x - xround
end


@deprecate periodic_index(i, N) periodize(i, N)


@inline absolute_maximum(x) = norm(x, Inf)


"""
    absolute_maximum_index(x::AbstractArray)

Return index and value of the absolute maximum of array `x`.
"""
function absolute_maximum_index(x::AbstractArray)
  max_i = 1
  max_x = abs(x[1])
  for (i,xi) in enumerate(x)
    abs_x = abs(xi)
    if abs_x > max_x
      max_i = i
      max_x = abs_x
    end
  end
  return max_i, max_x
end


"""
    absmax(vs...)

Return absolute maximum of `vs...`.
"""
@inline absmax(vs...) = maximum(abs.(vs))
@inline absmax(v1,v2) = max(abs(v1),abs(v2))
