## Notes
#
# TODO
# - Refactor usage of `array` to `vector` if it is a vector.


#######################################################################
#                                Type                                 #
#######################################################################


# This could be simplified by using a dictionary instead, e.g. datadict::Dict{Symbol,Vector{Float64}}.
# This would also allow/make it more convinient to use string keys to retrieve a variable, 
# e.g. vars["grhd.rho"], etc.
# To implement this overload Base.getindex(A, idx)
# Just noted that getproperty can be overriden for Strings, hence, one can also write
# A."idx" to retrieve a component.
#
# Regarding performance: How does a key-lookup compare to index search in Vector{Symbol}?
# And: If it is slower, does/can this even matter compared to the other parts?
# But this is an implementationd detail which can be benchmarked latter on if needed.
mutable struct _VariablesDict
  npts::Int64
  datadict::OrderedDict{Symbol, Vector{Float64}}
  verbose::Bool # notify when variables are added

  function _VariablesDict(npts::Int64, names::Vector{Symbol}, arrays::Vector{Vector{Float64}};
                          verbose::Bool=false)
    @assert npts > 0 "Number of points must be greater than zero"
    @assert length(names) == length(arrays)
    npts_arrays = [ length(array) for array in arrays ]
    for npts_arr in npts_arrays
      @assert npts == npts_arr "Mismatch between number of points and length of data array!"
    end
    datadict = OrderedDict([ (name, array) for (name, array) in zip(names, arrays) ])
    return new(npts, datadict, verbose)
  end
end


#######################################################################
#                            Constructors                             #
#######################################################################


function _VariablesDict(npts::Int64, names::Symbol...; kwargs...)
  names   = collect(Symbol, names)
  arrays  = [ zeros(Float64, npts) for _ in 1:length(names) ]
  return _VariablesDict(npts, names, arrays; kwargs...)
end


function _VariablesDict(array::Vector{Float64}, npts::Int64, names::Symbol...)
  names   = collect(Symbol, names)
  wrapped_arrays = [ unsafe_wrap(Vector{Float64}, pointer(array, (i-1)*npts+1), npts) 
                     for i in 1:length(names) ]
  return _VariablesDict(npts, names, wrapped_arrays)
end


# _VariablesDict(npts::Int64) = _VariablesDict(npts, String[], Vector{Float64}[])
# not defined yet, because it requires a resize method
# _VariablesDict()            = _VariablesDict(0)


#######################################################################
#                               Methods                               #
#######################################################################


function register!(vars::_VariablesDict, name::Symbol)
  data = zeros(Float64, length(vars))
  add!(vars, name, data)
  return
end


function register!(vars::_VariablesDict, names::Symbol...)
  for name in names
    register!(vars, name)
  end
  return
end


function set!(vars::_VariablesDict, name::Symbol, array::Vector{Float64})
  vars.verbose && @info "Adding variable '$name' ..."
  vars.datadict[name] = array
  return
end


function set!(vars::_VariablesDict, array::Vector{Float64})
  n_vars = n_variables(vars)
  npts::Int64 = length(vars)
  if length(array) % npts * n_vars != 0 
    error("Mismatch between number points, number of variables and length of array")
  end
  for (i, name) in enumerate(names(vars))
    vars.datadict[name] = unsafe_wrap(Vector{Float64}, pointer(array, (i-1)*npts+1), npts)
  end
  return
end


function unset!(vars::_VariablesDict, name::Symbol)
  vars.verbose && @info "Removing variable '$name' ..."
  return Base.delete!(vars.datadict, name)
end


function add!(vars::_VariablesDict, name::Symbol, array::Vector{Float64})
  if haskey(vars.datadict, name)
    error("Variable '$name' already contained. Use set! to overwrite it.")
  end
  set!(vars, name, array)
  return
end


function remove!(vars::_VariablesDict, name::Symbol)
  if !haskey(vars.datadict, name)
    error("Variable '$name' not found. Use unset! to avoid this error.")
  end
  return unset!(vars, name)
end


function Base.getproperty(vars::_VariablesDict, name::Symbol)
  datadict = getfield(vars, :datadict)
  value = get(datadict, name, nothing)
  if !isnothing(value)
    return value
  else
    return getfield(vars, name)
  end
end


function Base.propertynames(vars::_VariablesDict, private::Bool=false)
  return [ fieldnames(_VariablesDict)..., names(vars)... ]
end


function Base.setproperty!(vars::_VariablesDict, name::Symbol, array::Vector{Float64})
  if length(array) != vars.npts
    error("Mismatch between number of points and data array length")
  end
  add!(vars, (name, array))
end


function names(vars::_VariablesDict)::Vector{Symbol}
  return collect(keys(vars.datadict))
end


function fields(vars::_VariablesDict)
  return collect(values(vars.datadict))
end


function Base.pairs(vars::_VariablesDict)
  return zip(names(vars), fields(vars))
end


n_variables(vars::_VariablesDict)::Int64  = length(vars.datadict)
Base.length(vars::_VariablesDict)::Int64  = vars.npts
Base.isinf(vars::_VariablesDict)   = any(map(arr -> any(isinf.(arr)), fields(vars)))
Base.isnan(vars::_VariablesDict)   = any(map(arr -> any(isnan.(arr)), fields(vars)))


# custom display method utilizing pretty_tables
function Base.show(io::IO, vars::_VariablesDict)
  data = hcat(fields(vars)...)
  header = names(vars)
  if length(header) > 0
    println()
    pretty_table(data, header=header)
  end
end


# overload show with MIME type so that display(var) shows the same thing in the REPL
function Base.show(io::IO, mime::MIME"text/plain", var::_VariablesDict)
  show(io, var)
end


## resize method?
# There are various implementations possible:
# - Extend resize vectors naively. This makes not much sense, because the data then has to be
# adjusted somewhere else
# - Provide an 'interpolateion/restriction interface', e.g. 
# resize(vars::Variables, resize_function) where the `resize_function` does interpolation
# or restriction in a clever way to avoid errors and keep data consistent
#
# This feature will be needed when implementing h/p refinement.
