"""
    @with_signature [legacy=true]

This is the current default behavior of `@with_signature` but it is deprecated.
Macro to extract `accepted` and `returned` arguments from a function and make them
programmatically accessible. Functions defined using this macro can be used with
the `broadcast_volume!, broadcast_faces!, broadcast_bdrys` interfaces.


# Example
```
@with_signature function fn(eq::Equation)
  @accepts State(u1, u2), a, b
  # compute values rhs_u1, rhs_u2
  @returns rhs_u1, rhs_u2
end
```
translates into
```
function fn(args::NTuple{4,Float64}, eq::Equation)
  (u1, u2, a, b) = args
  # compute values rhs_u1, rhs_u2
  (rhs_u1, rhs_u2)
end
accepts(::typeof(fn), ::Equation) = (:u1, :u2, :a, :b)
returns(::typeof(fn), ::Equation) = (:rhs_u1, :rhs_u2)
state_indices(::typeof(fn), ::Equation) = (1,2)
has_signature(::typeof(fn), ::Equation) = true
```


---


    @with_signature [legacy=false]

Macro to extract `accepted` and `returned` arguments from a function and make them
programmatically accessible. Functions defined using this macro can be used with
the `broadcast_volume!, broadcast_faces!, broadcast_bdrys` interfaces.

To test a `@with_signature` in standalone mode, see below for an example.

The following options are available to customize `@with_signature` definitions:
- `skip_debug_info::Bool=false`: If `true` the generated code does not contain
  LineNumberNodes which makes it such that when an error occurs in the computation, Julia can't tell
  where it came from in the definition of the function. Only available to easily debug
  if the function parsing works correctly.
- `inbounds::Bool=true`: Insert a `@inbounds` for the grid loop.
- `simd::Bool=true`: Insert a `@simd` for the grid loop.
- `debug::Bool=false`: Disable `inbounds, simd` options.

The following options are available to customize the keyword `@accepts, @returns` definitions:
- `Prefix(name)`: Append `name_` to all names following the keywords.
- `bdry`: Indicate that data from the variables following this option should be loaded
from the boundary indices. Default is to load data from volume indices.


# Example
```
@with_signature [legacy=false] function flux(eq::Equation)
   @accepts fx, fy, u, v
   @accepts [Prefix(out),bdry] fx, fy, u, v
   @accepts [bdry] nx, ny
   nflx     = nx*fx + ny*fy
   out_nflx = nx*out_fx + ny*out_fy
   nflx     = LLF(nflx, out_nflx, u, out_u, absmax(v, out_v))
   @returns [bdry] nflx
end
```
translates (roughly) into
```
function flux(eq::Equation, arguments, returns, volumeidxs, bdryidxs)

  # unpack
  _arg_fx     = arguments[1]
  _arg_fy     = arguments[2]
  _arg_u      = arguments[3]
  _arg_v      = arguments[4]
  _arg_out_fx = arguments[5]
  _arg_out_fy = arguments[6]
  _arg_out_u  = arguments[7]
  _arg_out_v  = arguments[9]
  _arg_out_nx = arguments[10]
  _arg_out_ny = arguments[11]
  _ret_nflx   = returns[1]

  @inbounds for ii in eachindex(volumeidxs, bdryidxs)
    vi = volumeidxs[ii]
    bi = bdryidxs[ii]

    # read
    fx     = _arg_fx[vi]
    fy     = _arg_fy[vi]
    u      = _arg_u[vi]
    v      = _arg_v[vi]
    out_fx = _arg_out_fx[bi]
    out_fy = _arg_out_fy[bi]
    out_u  = _arg_out_u[bi]
    out_v  = _arg_out_v[bi]
    nx     = _arg_out_nx[bi]
    ny     = _arg_out_ny[bi]

    # actual computation
    nflx     = nx*fx + ny*fy
    out_nflx = nx*out_fx + ny*out_fy
    nflx     = LLF(nflx, out_nflx, u, out_u, absmax(v, out_v))

    # write
    _ret_nflx[bi] = nflx
  end
end

# helpers
accepts(::typeof(fn), ::Equation, nothing) = (:fx, :fy, :u, :v, :fx, :fy, :u, :v, :ny, :ny)
returns(::typeof(fn), ::Equation, nothing) = (:nflx,)
has_signature(::typeof(fn), ::Equation, nothing) = true
```

The above can be called as
```
N = 100
args = [ rand(N) for _ = 1:11 ]
rets = [ rand(N) for _ = 1:1  ]
volumeidxs = 1:100
bdryidxs   = 100:-1:1
eq = Equation()
flux(eq, args, rets, volumeidxs, bdryidxs)
```
"""
macro with_signature(ex)
  eex = legacy_with_signature(ex)
  return esc(eex)
end
macro with_signature(ex_option, ex)
  options = parse_options(ex_option)
  eex = options[:legacy] ? legacy_with_signature(ex) : with_signature(ex, options)
  return esc(eex)
end


function parse_options(ex)
  if !_is_vect(ex)
    error("@with_signature: invalid options '$ex', try instead '@with_signature [option_1=true] function flux(eq::Equation) ... end'")
  end
  options = Dict{Symbol,Any}()
  args = ex.args
  for a in args
    matched = MacroTools.@capture(a, lhs_ = rhs_)
    if !matched
      error("@with_signature: invalid options '$ex', try instead '@with_signature [option_1=true] function flux(eq::Equation) ... end'")
    end
    if haskey(options, lhs)
      error("@with_signature: duplicated option found in '$ex'")
    end
    if lhs === :skip_debug_info
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'skip_debug_info', expected a Bool")
      end
      options[:skip_debug_info] = rhs
      continue
    elseif lhs === :legacy
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'legacy', expected a Bool")
      end
      options[:legacy] = rhs
      continue
    elseif lhs === :print_idxs
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'print_idxs', expected a Bool")
      end
      options[:print_idxs] = rhs
      continue
    elseif lhs === :inbounds
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'inbounds', expected a Bool")
      end
      options[:inbounds] = rhs
      continue
    elseif lhs === :simd
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'simd', expected a Bool")
      end
      options[:simd] = rhs
      continue
    elseif lhs === :debug
      if !(rhs isa Bool)
        error("@with_signature: invalid value '$rhs' for option 'debug', expected a Bool")
      end
      options[:debug] = rhs
      continue
    end
    error("@with_signature: unknown option '$a' found in '$ex', skipping ...")
  end
  # set defaults
  get!(options, :skip_debug_info, false)
  get!(options, :legacy, true)
  get!(options, :print_idxs, false)
  get!(options, :inbounds, true)
  get!(options, :simd, true)
  get!(options, :debug, false)
  if options[:debug]
    options[:inbounds] = false
    options[:simd]     = false
  end
  return options
end


"""
See @with_signature
"""
macro accepts(ex); ex; end
macro returns(ex); ex; end


_is_state(ex) = ex isa Expr && ex.head === :call && ex.args[1] === :State
_is_tuple(ex) = ex isa Expr && ex.head === :tuple
_is_vect(ex) = ex isa Expr && ex.head === :vect


# gather list of all variables and a list of indices of variables
# tagged with State and unpack optional Prefix names, e.g.
# accepts(:(State(a, b), c, d, State(e), f))
# ->
# (:a, :b, :c, :d, :e, :f), (1, 2, 5), nothing
# oor
# accepts(:(Prefix(lhs), c, d, State(e), f))
# ->
# (:c, :d, :e, :f), (), :lhs
function legacy_accepts(ex)
  allsyms, i_states, prefix = if ex isa Symbol
    (ex,), (), nothing
  elseif _is_state(ex)
    vars = ex.args[2:end]
    vars, collect(1:length(vars)), nothing
  elseif _is_tuple(ex)
    vars = ex.args[1:end]
    allsyms, i_states = Symbol[], Int64[]
    prefix = if vars[1] isa Expr && vars[1].head === :call && vars[1].args[1] === :Prefix
      if length(vars[1].args) != 2
        error("@accepts: invalid use of 'Prefix' in '$(vars[1])', use as '@accepts Prefix(lhs), a, b'")
      end
      p = vars[1].args[2]
      if !(p isa Symbol)
        error("@accepts: expected a symbol as prefix, found '$(vars[1])'")
      end
      popfirst!(vars)
      p
    else
      nothing
    end
    last_i = 1
    for v in vars
      if _is_state(v)
        vs = v.args[2:end]
        push!(allsyms, vs...)
        append!(i_states, collect(1:length(vs)) .+ (last_i-1))
        last_i += length(vars)-1
      elseif v isa Symbol
        push!(allsyms, v)
      else
        error("@accepts: expected a variable name or a State tag")
      end
    end
    allsyms, i_states, prefix
  else
    error("@accepts: expected '@accepts a, b, ...'")
  end
  tuple(allsyms...), tuple(i_states...), prefix
end


function legacy_returns(ex)
  if ex isa Symbol
    (ex,)
  elseif _is_tuple(ex)
    tuple(ex.args...)
  else
    error("@returns: expected '@returns f_a, f_b, ...'")
  end
end


state(tpl, is::NTuple{N,Int64}) where N = getindex.(Ref(tpl), is)


function legacy_with_signature(ex)
  ex = MacroTools.prewalk(MacroTools.rmlines, ex)
  # display(ex)

  ex.head === :function || error("@with_signature: only works with functions")
  fn_signature = ex.args[1]
  length(fn_signature.args) == 2 || error(
    "@with_signature: function with exactly one argument required")
  fn_name, first_arg = fn_signature.args

  ex.args[2].head === :block || error("@with_singature: failed to extract function body")
  fn_body = ex.args[2].args

  # extract place holders
  length(fn_body) == 0 && error("@with_signature: missing function body")
  accepts_lines = Any[]
  for line in fn_body
    matched_accepts = MacroTools.@capture(line, @accepts accepts_variables__)
    !matched_accepts && break
    accepts_variables = last(accepts_variables)
    push!(accepts_lines, accepts_variables)
  end
  length(accepts_lines) == 0 && error(
    "@with_signature: expected '@accepts a, b, ...' at the start of the function body")
  matched_returns = MacroTools.@capture(fn_body[end], @returns returns_variables_)
  matched_returns || error(
    "@with_signature: expected '@returns a, b, ...' at the end of the function body")

  # @accepts and @returns can only occur right at the start or end
  if any(line -> MacroTools.@capture(line, @accepts _), fn_body[length(accepts_lines)+1:end])
    error("@with_signature: expected '@accepts a, b, ...' at the start of the function body")
  end
  for line in fn_body[1:end-1]
    MacroTools.@capture(line, @returns rest_) && error(
      "@with_signature: @returns can only occur once")
  end

  # gather variables from place holders
  accepts_vars_stateidxs_prefix = legacy_accepts.(accepts_lines)
  returns_variables = legacy_returns(returns_variables)

  # to keep backwards compatibility with State(...), but will remove State syntax the future
  for (_,stateidxs,prefix) in accepts_vars_stateidxs_prefix
    if !isnothing(prefix) && length(stateidxs) > 0
      error("cannot use State(...) and Prefix(...) together in @accepts")
    end
  end
  pos_state_variables = if length(accepts_vars_stateidxs_prefix) > 1
    for (_,stateidxs,_) in accepts_vars_stateidxs_prefix
      if length(stateidxs) > 0
        error("cannot use State(...) with @accepts arguments")
      end
    end
    ()
  else
    accepts_vars_stateidxs_prefix[1][2]
  end

  # replace the @accept lines with tuple unpacking
  allvars = Symbol[]
  for (i,(vars, stateidxs, prefix)) in enumerate(accepts_vars_stateidxs_prefix)
    if !isnothing(prefix)
      vars = Symbol.(prefix,:_,vars)
      append!(allvars, vars)
    end
    unpack = Expr(:(=), Expr(:tuple, vars...), Symbol(:args,i))
    fn_body[i] = unpack
  end

  # error on duplicated variable names (after resolving prefixes)
  if !allunique(allvars)
    uvars = unique(allvars)
    idxdups = findall(v -> count(Base.Fix1(===,v),allvars) > 1, uvars)
    dups = allvars[idxdups]
    error("duplicated arguments in all @accepts found (after resolving prefixes): $(join(dups,","))")
  end
  fn_body[end] = Expr(:tuple, returns_variables...)

  # sanitize function arguments
  (first_arg isa Expr && first_arg.head === :(::)) || error(
    "@with_signature: function argument '$first_arg' must be type annotated")
  dispatch_type = last(first_arg.args)

  # values are passed via an additional argument
  # we type annotate it here with a ::NTuple{N,Float64}
  annotated_variables_arg = Expr[]
  for (i,(vars, stateidxs, prefix)) in enumerate(accepts_vars_stateidxs_prefix)
    nvars = length(vars)
    push!(annotated_variables_arg, :($(Symbol(:args,i))::NTuple{$nvars,Float64}))
  end
  new_fn_signature = Expr(:call, fn_name, annotated_variables_arg..., first_arg)
  ex.args[1] = new_fn_signature

  # unpack the cache variable names
  accepts_variables = Tuple( vars for (vars,_,_) in accepts_vars_stateidxs_prefix )

  thismod = @__MODULE__

  new_ex = quote
    $ex
    $(thismod).signature(::typeof($(fn_name)), ::$(dispatch_type)) =
        ($(accepts_variables), $(returns_variables))
    $(thismod).state_indices(::typeof($(fn_name)), ::$(dispatch_type)) = $(pos_state_variables)
    $(thismod).has_signature(::typeof($(fn_name)), ::$(dispatch_type)) = true
  end

  # display(MacroTools.prewalk(MacroTools.rmlines, new_ex))

  return new_ex
end



# parses a `ex = :([option=1,option2] a, b, c, d)` and returns
# `Dict(option=>1, option=>2], (:c, :d, :e, :f)`
#
# `keyword, line` are only for error messages
function parse_options_variables(keyword, line, ex)

  # unpack
  ex_options, ex_args = if length(ex) == 1
    nothing, ex[1]
  elseif length(ex) == 2
    ex[1], ex[2]
  else
    error("$keyword: invalid syntax in '$line', use as '$keyword [options] a, b, c'")
  end

  # unpack options
  options = Dict{Symbol,Any}()
  if !isnothing(ex_options)
    if !_is_vect(ex_options)
      error("$keyword: invalid syntax for options '$ex_options' found in '$ex', use as '$keyword [options] a, b, c'")
    end
    args = ex_options.args
    for a in args
      if haskey(options, a)
        error("$keyword: duplicated option found in '$ex'")
      end
      if a === :bdry
        options[:bdry] = true
        continue
      end
      matched = MacroTools.@capture(a, Prefix(match_))
      if matched
        options[:prefix] = match
        continue
      end
      error("$keyword: unknown option '$a' found in '$ex_options', skipping ...")
    end
  end
  # set defaults
  get!(options, :bdry, false)
  get!(options, :prefix, nothing)

  vars = if ex_args isa Symbol
    [ex_args]
  elseif _is_tuple(ex_args)
    vars = ex_args.args[1:end]
    for v in vars
      if !(v isa Symbol)
        error("$keyword: invalid variable '$v' found in '$line', only use Symbols with $keyword")
      end
    end
    vars
  else
    error("$keyword: invalid syntax for variables found in '$line', use as '$keyword [options] a, b, c'")
  end

  options, vars
end


parse_accepts(line, ex) = parse_options_variables("@accepts", line, ex)
parse_returns(line, ex) = parse_options_variables("@returns", line, ex)


function with_signature(expr, options=(;))
  ex = MacroTools.prewalk(MacroTools.rmlines, expr)
  # display(ex)

  ex.head === :function || error("@with_signature: only works with functions")
  fn_signature = ex.args[1]
  length(fn_signature.args) == 2 || error(
    "@with_signature: function with exactly one argument required")
  fn_name, first_arg = fn_signature.args

  ex.args[2].head === :block || error("@with_singature: failed to extract function body")
  fn_body = ex.args[2].args

  # extract place holders
  length(fn_body) == 0 && error("@with_signature: missing function body")
  accepts_lines, accepts_exs = Any[], Any[]
  for line in fn_body
    matched_accepts = MacroTools.@capture(line, @accepts accepts_ex__)
    !matched_accepts && break
    push!(accepts_lines, line)
    push!(accepts_exs, accepts_ex)
  end
  length(accepts_lines) == 0 && error(
    "@with_signature: expected '@accepts [options] a, b, ...' at the start of the function body")
  returns_lines, returns_exs = Any[], Any[]
  for line in reverse(fn_body) # search for @returns from bottom to top
    matched_returns = MacroTools.@capture(line, @returns returns_ex__)
    !matched_returns && break
    push!(returns_lines, line)
    push!(returns_exs, returns_ex)
  end
  length(returns_lines) == 0 && error(
    "@with_signature: expected '@returns [options] a, b, ...' at the end of the function body")
  # reverse returns expressions so that we emmit code from top to bottom
  returns_exs   = returns_exs[end:-1:1]
  returns_lines = returns_lines[end:-1:1]

  # @accepts and @returns can only occur right at the start or end
  if any(line -> MacroTools.@capture(line, @accepts _), fn_body[length(accepts_lines)+1:end])
    error("@with_signature: expected '@accepts [options] a, b, ...' at the start of the function body")
  end
  if any(line -> MacroTools.@capture(line, @returns _), fn_body[1:end-length(returns_lines)])
    error("@with_signature: expected '@returns [options] a, b, ...' at the end of the function body")
  end

  # gather variables from place holders
  accepts_options_vars = parse_accepts.(accepts_lines, accepts_exs)
  returns_options_vars = parse_returns.(returns_lines, returns_exs)

  # exctract compute body from initial expression in order to preserve
  # line number nodes for debugging
  compute_body = if !options[:skip_debug_info]
    try
      @assert expr.head === :function
      init_fn_body = expr.args[2].args
      init_fn_body[2+2*length(accepts_lines):end-2*length(returns_lines)]
    catch
      fn_body[1+length(accepts_lines):end-length(returns_lines)]
    end
  else
      fn_body[1+length(accepts_lines):end-length(returns_lines)]
  end
  if length(compute_body) == 0
    error("""@with_signature: nothing to compute in expression
            @with_signature $ex
          """)
  end

  # make gen'ed symbols for construction of the rest of the function body
  acc, ret, ii, vi, bi, vidxs, bidxs =
    gensym.((:accepts, :returns, :ii,:vi,:bi,:vidxs,:bidxs))

  # for each variable in @accepts
  # 1) resolve the prefix,
  # 2) make a gen'ed symbol for array unpacking
  # 3) construct the array unpacking calls
  acc_vars, prefix_acc_vars, gen_acc_vars, bdry_acc = Symbol[], Symbol[], Symbol[], Bool[]
  unpack_acc = Expr[]
  n = 1
  for (i,(opts, vars)) in enumerate(accepts_options_vars)
    append!(acc_vars, vars)
    append!(prefix_acc_vars, isnothing(opts[:prefix]) ? vars : Symbol.(opts[:prefix],:_,vars))
    append!(bdry_acc, fill(opts[:bdry], length(vars)))
    for v in vars
      gv = gensym(v)
      push!(unpack_acc, :($gv = $acc[$n]))
      push!(gen_acc_vars, gv)
      n += 1
    end
  end

  # for each variable in @returns
  # 1) resolve the prefix,
  # 2) make a gen'ed symbol for array unpacking
  # 3) construct the array unpacking calls
  ret_vars, prefix_ret_vars, gen_ret_vars, bdry_ret = Symbol[], Symbol[], Symbol[], Bool[]
  unpack_ret = Expr[]
  n = 1
  for (i,(opts, vars)) in enumerate(returns_options_vars)
    append!(ret_vars, vars)
    append!(prefix_ret_vars, isnothing(opts[:prefix]) ? vars : Symbol.(opts[:prefix],:_,vars))
    append!(bdry_ret, fill(opts[:bdry], length(vars)))
    for v in vars
      gv = gensym(v)
      push!(unpack_ret, :($gv = $ret[$n]))
      push!(gen_ret_vars, gv)
      n += 1
    end
  end

  # error on duplicated variable names (after resolving prefixes)
  if !allunique(prefix_acc_vars)
    uvars = unique(prefix_acc_vars)
    idxdups = findall(v -> count(Base.Fix1(===,v),prefix_acc_vars) > 1, uvars)
    dups = prefix_acc_vars[idxdups]
    error("duplicated arguments in all @accepts found (after resolving prefixes): $(join(dups,", "))")
  end
  if !allunique(prefix_ret_vars)
    uvars = unique(prefix_ret_vars)
    idxdups = findall(v -> count(Base.Fix1(===,v),prefix_ret_vars) > 1, uvars)
    dups = prefix_ret_vars[idxdups]
    error("duplicated arguments in all @returns found (after resolving prefixes): $(join(dups,", "))")
  end

  # generate getindex calls (v = var"##v#359"[i]) for all @accepts variables to load data inside loop
  getvars_lines = [ :( $pv = $gv[$(bdry ? bi : vi)] )
                    for (pv,gv,bdry) in zip(prefix_acc_vars,gen_acc_vars,bdry_acc) ]
  # generate setindex! calls (var"##v#359"[i] = v) for all @returns variables to write data inside loop
  setvars_lines = [ :( $gv[$(bdry ? bi : vi)] = $pv )
                    for (pv,gv,bdry) in zip(prefix_ret_vars,gen_ret_vars,bdry_ret) ]

  # need to use @macroexpand1 @with_signature when debugging to not expand @inbounds macro
  loop_ex = if options[:print_idxs]
    quote
      for $ii in eachindex($vidxs, $bidxs)
        $vi = $vidxs[$ii]
        $bi = $bidxs[$ii]
        $(getvars_lines...)
        $(compute_body...)
        $(setvars_lines...)
        println((;ii=$ii, vi=$vi, bi=$bi))
      end
    end
  else
    quote
      for $ii in eachindex($vidxs, $bidxs)
        $vi = $vidxs[$ii]
        $bi = $bidxs[$ii]
        $(getvars_lines...)
        $(compute_body...)
        $(setvars_lines...)
      end
    end
  end
  loop_ex = MacroTools.prewalk(MacroTools.rmlines, loop_ex)
  loop_ex = loop_ex.args[1] # unwrap a begin block
  loop_body = loop_ex.args[2]

  # prepend macros to for loop
  if options[:simd]
    loop_ex = Expr(:macrocall, Symbol("@simd"), nothing, loop_ex)
  end
  if options[:inbounds]
    loop_ex = Expr(:macrocall, Symbol("@inbounds"), nothing, loop_ex)
  end

  loop_ex = Expr(:block, loop_ex) # wrap into begin

  new_ex = deepcopy(ex)
  new_ex.args[2].args = vcat(unpack_acc, unpack_ret, loop_ex)

  # sanitize function arguments
  (first_arg isa Expr && first_arg.head === :(::)) || error(
    "@with_signature: function argument '$first_arg' must be type annotated")
  dispatch_type = last(first_arg.args)

  new_fn_signature = Expr(:call, fn_name, first_arg, acc, ret, vidxs, bidxs)
  new_ex.args[1] = new_fn_signature

  tpl_acc_vars = tuple(acc_vars...)
  tpl_ret_vars = tuple(ret_vars...)

  thismod = @__MODULE__

  # TODO Remove dummy dispatch later
  # TODO What about docstrings for these methods?
  defs = quote
    $new_ex
    $(thismod).signature(::typeof($(fn_name)), ::$(dispatch_type), ::Nothing) = ($(tpl_acc_vars), $(tpl_ret_vars))
    $(thismod).has_signature(::typeof($(fn_name)), ::Type{<:$(dispatch_type)}, ::Nothing) = true
  end

  # display(MacroTools.prewalk(MacroTools.rmlines, new_ex))

  return defs
end


#######################################################################
#                          Interface methods                          #
#######################################################################


has_signature(fn, dispatch) = false
signature(fn, dispatch) = error(
  "function '$fn' with dispatch '$(typeof(dispatch))' was not defined using @with_signature")
state_indices(fn, dispatch) = error(
  "function '$fn' with dispatch '$(typeof(dispatch))' was not defined using @with_signature")
accepts(fn, dispatch) = signature(fn, dispatch)[1]
returns(fn, dispatch) = signature(fn, dispatch)[2]

accepts(fn, dispatch, ::Nothing) = signature(fn, dispatch, nothing)[1]
returns(fn, dispatch, ::Nothing) = signature(fn, dispatch, nothing)[2]
signature(fn, dispatch, ::Nothing) = error(
  "function '$fn' with dispatch '$(typeof(dispatch))' was not defined using @with_signature")
has_signature(fn, dispatch::Type, ::Nothing) = false
has_signature(fn, dispatch, ::Nothing) = has_signature(fn, typeof(dispatch), nothing)


#######################################################################
#                          broadcast methods                          #
#######################################################################


#######################################################################
#                              version 1                              #
#######################################################################


"""
    broadcast_volume!(f::Function, D::T_Dispatch, C::Cache) where T_Dispatch

Broadcast a `@with_signature` function `f` dispatched on `D` over variables of
cache `C`. The arguments and return values are determined by the `accepts, returns` overloads
of `(f, D)` from the `@with_signature` interface.

See: `@with_signature`.
"""
function broadcast_volume!(f::Function, D::T_Dispatch, C::Cache) where T_Dispatch
  has_signature(f, D) || error("broadcast_volume!: Function '$f' has no signature!")
  accepts_syms, returns_syms = accepts(f, D)[1], returns(f, D)
  n_accepts, n_returns = length(accepts_syms), length(returns_syms)
  v = get_variable(C, first(accepts_syms))
  accepts_vars = NTuple{n_accepts,Vector{Float64}}( get_variable(C, a) for a in accepts_syms )
  returns_vars = NTuple{n_returns,Vector{Float64}}( get_variable(C, r) for r in returns_syms )
  f_args = args -> f(args, D)
  _broadcast_volume!(f_args, C, accepts_vars, returns_vars)
  return
end


function _broadcast_volume!(f, C::Cache,
    accepts_vars::NTuple{N,Vector{Float64}},
    returns_vars::NTuple{M,Vector{Float64}}) where {N,M}

  soa_accepts = StructArray{NTuple{N,Float64}}( accepts_vars )
  soa_returns = StructArray{NTuple{M,Float64}}( returns_vars )
  npts = length(first(accepts_vars))
  for idx = 1:npts
    soa_returns[idx] = f(soa_accepts[idx])
  end
  return
end


@deprecate broadcast_volume!(f, rsolver, cache, mesh) broadcast_faces!(f, rsolver, cache, mesh)


"""
    broadcast_faces!(f, rsolver::ApproxRiemannSolver, cache::Cache, mesh::Mesh)

Broadcast `@with_signature` function `f` with dispatch `rsolver` over `cache` and
faces of `mesh`.
"""
function broadcast_faces!(f, rsolver::AbstractRiemannSolver, cache::Cache, mesh::Mesh)
  has_signature(f, rsolver) || error("broadcast_faces!: Function '$f' has no signature!")
  accepts_vars = accepts(f, rsolver)
  returns_vars = returns(f, rsolver)
  n_accepts    = length(accepts_vars)
  n_returns    = length(returns_vars)
  accepts_data = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                    for a in accepts_vars )
  returns_data = NTuple{n_returns,Vector{Float64}}( get_variable(cache, r)
                                                    for r in returns_vars )
  f_args = (args, normal) -> f(args, rsolver, normal)
  _broadcast_face!(f_args, mesh, accepts_data, returns_data)
  return
end


function broadcast_faces!(f, equation, mesh::Mesh)
  has_signature(f, equation) || error("broadcast_faces!: Function '$f' has no signature!")
  accepts_vars = accepts(f, equation)
  returns_vars = returns(f, equation)
  n_accepts    = length(accepts_vars)
  n_returns    = length(returns_vars)
  accepts_data = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                    for a in accepts_vars )
  returns_data = NTuple{n_returns,Vector{Float64}}( get_variable(cache, r)
                                                    for r in returns_vars )
  f_args = (args, normal) -> f(args, rsolver, normal)
  _broadcast_face!(f_args, mesh, accepts_data, returns_data)
  return
end


function _broadcast_face!(f, mesh::Mesh1d,
    state_data::NTuple{Na,Vector{Float64}},
    nflx_data::NTuple{Nr,Vector{Float64}}) where {Na,Nr}

  soa_state = StructArray{NTuple{Na,Float64}}( state_data )
  soa_nflx = StructArray{NTuple{Nr,Float64}}( nflx_data )
  Npts, K = layout(mesh)
  Idx = LinearIndices((Npts, K))
  n_out = (-1.0,) # 'normal' in negative direction
  for cell = 1:K
    # each loop deals with the lhs' interface of the current cell
    cell_out, cell_in = periodic_index(cell-1, K), cell
    idx_out, idx_in = Idx[end,cell_out], Idx[1,cell_in]
    accepts_out, accepts_in = soa_state[idx_out], soa_state[idx_in]
    nflx_out = f((accepts_out, accepts_in), n_out)
    # n_out * f(u_out, u_in, n_out) = - n_in * f(u_in, u_out, n_in) with n_out = -n_in
    nflx_in  = @. -nflx_out
    face_idx_rhs = cell
    face_idx_lhs = K+periodize(cell-K-1, K)
    soa_nflx[face_idx_rhs] = nflx_out
    soa_nflx[face_idx_lhs] = nflx_in
  end
  return
end


function _broadcast_face!(f, mesh::Mesh2d,
    state_data::NTuple{Na,Vector{Float64}},
    nflx_data::NTuple{Nr,Vector{Float64}}) where {Na,Nr}

  soa_state = StructArray{NTuple{Na,Float64}}( state_data )
  soa_nflx  = StructArray{NTuple{Nr,Float64}}( nflx_data )

  @unpack Npts = mesh.element
  Kx, Ky = mesh.tree.dims
  Nx, Ny = Npts, Npts

  @unpack nx, ny = get_bdry_variables(mesh.cache)

  for cell = 1:ncells(mesh)

    bulk_in = cellindices(mesh, cell)
    rng_lhs, rng_rhs, rng_down, rng_up = faceindices(mesh, cell)

    # xmin face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Xmin]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_lhs)
      idx_out = bulk_out[end,fidx]
      idx_in  = bulk_in[1,fidx]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

    # xmax face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Xmax]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_rhs)
      idx_out = bulk_out[1,fidx]
      idx_in  = bulk_in[end,fidx]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

    # ymin face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Ymin]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_down)
      idx_out = bulk_out[fidx,end]
      idx_in  = bulk_in[fidx,1]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

    # ymax face
    cell_out = mesh.tree.cells[cell].neighbors[Cart2d.Ymax]
    bulk_out = cellindices(mesh, cell_out)
    for (fidx, face_idx) in enumerate(rng_up)
      idx_out = bulk_out[fidx,1]
      idx_in  = bulk_in[fidx,end]
      n_out = (nx[face_idx], ny[face_idx])
      state_out, state_in = soa_state[idx_out], soa_state[idx_in]
      soa_nflx[face_idx] = f((state_out, state_in), n_out)
    end

  end
  return
end


"""
    broadcast_boundaryconditions!(f, bc::BoundaryConditions, cache::Cache, mesh::Mesh, t)

Impose boundary conditions using the flux method `f` and boundary conditions `bc`
for fields in `cache` on `mesh` at time `t`.
"""
function broadcast_boundaryconditions!(f, bc::BoundaryConditions, cache::Cache, mesh::Mesh1d, t=0.0)
  @unpack lhs_bc, rhs_bc, rsolver = bc
  has_signature(f, rsolver) || error("broadcast_volume!: Function '$f' has no signature!")
  bc.isperiodic && return

  accepts_vars = accepts(f, rsolver)
  returns_vars = returns(f, rsolver)
  n_accepts    = length(accepts_vars)
  n_returns    = length(returns_vars)
  accepts_data = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                    for a in accepts_vars )
  returns_data = NTuple{n_returns,Vector{Float64}}( get_variable(cache, r)
                                                    for r in returns_vars )
  f_args = (args, normal) -> f(args, rsolver, normal)
  _broadcast_boundaries!(f_args, mesh, accepts_data, returns_data, lhs_bc, rhs_bc, t)
  return
end


boundary_state(bc::AbstractBC, u_inner, t) = bc(t)
boundary_state(bc::OutflowBC, u_inner, t)  = u_inner


# similar to _broadcast_volume!, but only iterating domain bdrys and broadcasting
# over two returned data arrays
function _broadcast_boundaries!(f, mesh::Mesh,
    state_data::NTuple{Na,Vector{Float64}},
    nflx_data::NTuple{Nr,Vector{Float64}},
    lhs_bc::AbstractBC,
    rhs_bc::AbstractBC,
    t) where {Na,Nr}

  soa_state = StructArray{NTuple{Na,Float64}}( state_data )
  soa_nflx = StructArray{NTuple{Nr,Float64}}( nflx_data )
  Npts, K = layout(mesh)

  # lhs bdry
  cell = 1
  state = soa_state[cell]
  bdry_state = boundary_state(lhs_bc, state, t)
  n_out = (-1.0,)
  nf = f((bdry_state, state), n_out)
  soa_nflx[1] = f((bdry_state, state), n_out)

  # rhs bdry
  cell = K
  state = soa_state[end]
  bdry_state = boundary_state(rhs_bc, state, t)
  n_out = (1.0,)
  soa_nflx[end] = f((bdry_state, state), n_out)

  return
end


#######################################################################
#                              version 2                              #
#######################################################################


function wrap_variables_in_soa(cache, vars1)
  n1 = length(vars1)
  v1 = NTuple{n1,Vector{Float64}}( get_variable(cache, v) for v in vars1 )
  soa_v1 = StructArray{NTuple{n1,Float64}}( v1 )
  return soa_v1
end
function wrap_variables_in_soa(cache, vars1, vars2, vars3)
  n1 = length(vars1)
  v1 = NTuple{n1,Vector{Float64}}( get_variable(cache, v) for v in vars1 )
  soa_v1 = StructArray{NTuple{n1,Float64}}( v1 )
  n2 = length(vars2)
  v2 = NTuple{n2,Vector{Float64}}( get_variable(cache, v) for v in vars2 )
  soa_v2 = StructArray{NTuple{n2,Float64}}( v2 )
  n3 = length(vars3)
  v3 = NTuple{n3,Vector{Float64}}( get_variable(cache, v) for v in vars3 )
  soa_v3 = StructArray{NTuple{n3,Float64}}( v3 )
  return soa_v1, soa_v2, soa_v3
end


function broadcast_volume_2!(f, equation, mesh::AbstractMesh)
  @timeit TO "broadcast_volume_2!" begin
    _broadcast_volume_2!(f, equation, mesh)
  end
end


function _broadcast_volume_2!(f, equation, mesh::Mesh)
  @toggled_assert has_signature(f, equation) "broadcast_volume_2!: '$f' was not defined using @with_signature"
  accepts_vars1, = accepts(f, equation)
  returns_vars = returns(f, equation)
  args = wrap_variables_in_soa(mesh.cache, accepts_vars1)
  rets = wrap_variables_in_soa(mesh.cache, returns_vars)
  wrapped_f =  idx -> rets[idx] = f(args[idx], equation)
  _broadcast_volume_2!(wrapped_f, mesh)
  return
end


function _broadcast_volume_2!(f, m::Mesh)
  for idx = 1:length(m)
    @inline f(idx)
  end
  return
end


function broadcast_faces_2!(f, equation, mesh::AbstractMesh)
  @timeit TO "broadcast_faces_2!" begin
    _broadcast_faces_2!(f, equation, mesh)
  end
end


function _broadcast_faces_2!(f, equation, mesh::Mesh)
  @toggled_assert has_signature(f, equation) "broadcast_faces!: '$f' was not defined using @with_signature"
  vars1, vars2, vars3 = accepts(f, equation)
  returns_vars = returns(f, equation)
  args1, args2, args3 = wrap_variables_in_soa(mesh.cache, vars1, vars2, vars3)
  rets  = wrap_variables_in_soa(mesh.cache, returns_vars)
  wrapped_f = (i,j1,j2,j3) -> rets[i] = f(args1[j1],args2[j2],args3[j3],equation)
  _broadcast_face_2!(wrapped_f, mesh)
  return
end


function _broadcast_face_2!(f, mesh::Mesh1d)

  @unpack Npts = mesh.element
  Kx, = mesh.tree.dims
  Nx = Npts

  @inbounds for (icell,cell_in) = enumerate(mesh.tree.cells)

    bulk_in = cellindices(mesh, icell)
    rng_lhs, rng_rhs = faceindices(mesh, icell)

    # xmin face
    if has_neighbor(cell_in, Cart1d.Xmin)
      cell_out = cell_in.neighbors[Cart1d.Xmin]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_lhs)
        idx_out = bulk_out[end]
        idx_in  = bulk_in[1]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

    # xmax face
    if has_neighbor(cell_in, Cart1d.Xmax)
      cell_out = cell_in.neighbors[Cart1d.Xmax]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_rhs)
        idx_out = bulk_out[1]
        idx_in  = bulk_in[end]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

  end

end


function _broadcast_face_2!(f, mesh::Mesh2d)

  @unpack Npts = mesh.element
  Kx, Ky = mesh.tree.dims
  Nx, Ny = Npts, Npts

  for (icell,cell_in) = enumerate(mesh.tree.cells)

    bulk_in = cellindices(mesh, icell)
    rng_lhs, rng_rhs, rng_down, rng_up = faceindices(mesh, icell)

    if has_neighbor(cell_in, Cart2d.Xmin)
      cell_out = cell_in.neighbors[Cart2d.Xmin]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_lhs)
        idx_out = bulk_out[end,fidx]
        idx_in  = bulk_in[1,fidx]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

    if has_neighbor(cell_in, Cart2d.Xmax)
      cell_out = cell_in.neighbors[Cart2d.Xmax]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_rhs)
        idx_out = bulk_out[1,fidx]
        idx_in  = bulk_in[end,fidx]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

    if has_neighbor(cell_in, Cart2d.Ymin)
      cell_out = cell_in.neighbors[Cart2d.Ymin]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_down)
        idx_out = bulk_out[fidx,end]
        idx_in  = bulk_in[fidx,1]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

    if has_neighbor(cell_in, Cart2d.Ymax)
      cell_out = cell_in.neighbors[Cart2d.Ymax]
      bulk_out = cellindices(mesh, cell_out)
      for (fidx, face_idx) in enumerate(rng_up)
        idx_out = bulk_out[fidx,1]
        idx_in  = bulk_in[fidx,end]
        @inline f(face_idx, idx_in, idx_out, face_idx)
      end
    end

  end

  return
end


function broadcast_bdry_2!(f, equation, bdrycond::AbstractBC, mesh::Mesh)
  @toggled_assert has_signature(f, equation) "broadcast_bdry!: '$f' was not defined using @with_signature"
  vars1, vars2, vars3 = accepts(f, equation)
  returns_vars = returns(f, equation)
  args1, args2, args3 = wrap_variables_in_soa(mesh.cache, vars1, vars2, vars3)
  rets  = wrap_variables_in_soa(mesh.cache, returns_vars)
  wrapped_f = (i,j1,j2,j3) -> rets[i] = f(args1[j1],args2[j2],args3[j3],equation)
  _broadcast_bdry_2!(wrapped_f, bdrycond, mesh)
  return
end


function _broadcast_bdry_2!(f, bdrycond::DirichletBC2, mesh::Mesh1d)

  @unpack Npts = mesh.element
  Kx, = mesh.tree.dims
  Nx = Npts

  @inbounds for bdrycell in mesh.tree.bdrycells

    rng_lhs, rng_rhs = faceindices(mesh, bdrycell.index)
    idxs = cellindices(mesh, bdrycell.index)

    if !has_neighbor(bdrycell, Cart1d.Xmin)

      for (fidx, face_idx) in enumerate(rng_lhs)
        bidx = idxs[1]
        @inline f(face_idx, bidx, bidx, face_idx)
      end

    end

    if !has_neighbor(bdrycell, Cart1d.Xmax)

      for (fidx, face_idx) in enumerate(rng_rhs)
        bidx = idxs[end]
        @inline f(face_idx, bidx, bidx, face_idx)
      end

    end

  end

end


function _broadcast_bdry_2!(f, bdrycond::DirichletBC2, mesh::Mesh2d)

  @unpack Npts = mesh.element
  Kx, Ky = mesh.tree.dims
  Nx, Ny = Npts, Npts

  for bdrycell in mesh.tree.bdrycells

    rng_lhs, rng_rhs, rng_down, rng_up = faceindices(mesh, bdrycell.index)
    idxs = cellindices(mesh, bdrycell.index)

    if !has_neighbor(bdrycell, Cart2d.Xmin)

      for (fidx, face_idx) in enumerate(rng_lhs)
        bidx = idxs[1,fidx]
        @inline f(face_idx, bidx, bidx, face_idx)
      end

    end

    if !has_neighbor(bdrycell, Cart2d.Xmax)

      for (fidx, face_idx) in enumerate(rng_rhs)
        bidx = idxs[end,fidx]
        @inline f(face_idx, bidx, bidx, face_idx)
      end

    end

    if !has_neighbor(bdrycell, Cart2d.Ymin)

      for (fidx, face_idx) in enumerate(rng_down)
        bidx = idxs[fidx,1]
        @inline f(face_idx, bidx, bidx, face_idx)
      end

    end

    if !has_neighbor(bdrycell, Cart2d.Ymax)

      for (fidx, face_idx) in enumerate(rng_up)
        bidx = idxs[fidx,end]
        @inline f(face_idx, bidx, bidx, face_idx)
      end

    end

  end

  return
end


#######################################################################
#                              version 3                              #
#######################################################################


function broadcast_computation!(f, equation, mesh, volidxs, bdryidxs)
  @toggled_assert has_signature(f, equation, nothing) "'$f(::$(typeof(equation)))' was not defined using @with_signature"
  acc_syms = accepts(f, equation, nothing)
  ret_syms = returns(f, equation, nothing)
  acc_vars = NTuple{length(acc_syms),Vector{Float64}}( get_variable(mesh.cache, v) for v in acc_syms )
  ret_vars = NTuple{length(ret_syms),Vector{Float64}}( get_variable(mesh.cache, v) for v in ret_syms )
  f(equation, acc_vars, ret_vars, volidxs, bdryidxs)
  return
end

function new_broadcast_volume!(f, equation, mesh::AbstractMesh)
  @timeit TO "new_broadcast_volume!" begin
    broadcast_computation!(f, equation, mesh, volumeindices(mesh)...)
  end
end
function new_broadcast_faces!(f, equation, mesh::AbstractMesh)
  @timeit TO "new_broadcast_faces!" begin
    broadcast_computation!(f, equation, mesh, faceindices(mesh)...)
  end
end
function new_broadcast_bdry!(f, equation, mesh::AbstractMesh)
  @timeit TO "new_broadcast_bdry!" begin
    broadcast_computation!(f, equation, mesh, bdryindices(mesh)...)
  end
end
