const XDMF_END = """
</Domain>
</Xdmf>
"""

mutable struct XDMFOutput
    dir::String
    data_filename::String
    mesh_filename::String
    meshcount::Int64 # number of meshes written
    tcount::Int64 # number of time steps written
    binoffset::Int64 # binary offset to insert next data set, only for xdmf+bin
    mode::Symbol
end


function XDMFOutput(dir::AbstractString, mode::Symbol)
    isdir(dir) && throw(ArgumentError("directory '$dir' already exists"))
    mkdir(dir)

    data_filename = if mode === :h5
        joinpath(dir, "data.h5")
    elseif mode === :bin
        TODO()
        # Does not work yet, probably something wrong with binary encoding, endianess, offsets, etc.
        # also should try getting it to work without xpath first
        joinpath(dir, "data.bin")
    else
        error("Invalid storage mode requested '$mode', use one of ':bin, :h5'")
    end

    mesh_filename = joinpath(dir, "mesh.xdmf")

    # create empty files
    h5open(data_filename, "w") do file
    end
    open(mesh_filename, "w") do file
    end

    return XDMFOutput(dir, data_filename, mesh_filename, 0, 0, 0, mode)
end


function xdmfwrite_mesh(xdmf::XDMFOutput, mesh::Mesh)

    if xdmf.meshcount >= 1
        throw(ArgumentError("Mesh has already been write"))
    end

    buffer = IOBuffer()
    write_mesh_header(buffer)

    if xdmf.mode === :h5
        # Using "w" here to overwrite in case previous calls failed
        h5open(xdmf.data_filename, "w") do file
            write_mesh_h5(file, buffer, mesh)
        end
    else # xdmf.mode === :bin
        open(joinpath(xdmf.dir,"xyz.bin"), "w") do file
            b64io = Base64EncodePipe(file)
            write_mesh_binary(b64io, buffer, mesh)
        end
    end

    write_mesh_end(buffer)

    # finish up xml files
    # Using "w" here to overwrite in case previous calls failed
    open(xdmf.mesh_filename, "a") do file
        seekstart(buffer)
        write(file, buffer)
    end

    # updating count after all writes succeeded
    xdmf.meshcount += 1

end


function write_mesh_header(buffer)
    write(buffer, """
<?xml version="1.0" ?>
<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
<Xdmf xmlns:xi="http://www.w3.org/2003/XInclude" Version="3.0">
<Domain>
<Grid GridType="Collection" CollectionType="Spatial">
""")
end


function write_mesh_end(buffer)
    write(buffer, "</Grid>\n") # close grid collection
    write(buffer, XDMF_END) # close Domain and Xdmf
end


function write_mesh_h5(h5, buffer, mesh::Mesh2d)

    @unpack x, y = get_static_variables(mesh.cache)
    z = zeros(Float64, length(x))
    itx, ity, itz = eachcell(mesh,x), eachcell(mesh,y), eachcell(mesh,z)
    for (i,(cx,cy,cz)) in enumerate(zip(itx,ity,itz))
        nx, ny, nz = size3d(size(cx))
        g = create_group(h5, "mesh$i")
        # g["x"] = view(cx,:)
        # g["y"] = view(cy,:)
        # g["z"] = view(cz,:)
        xf32 = convert.(Float32, view(cx,:))
        yf32 = convert.(Float32, view(cy,:))
        zf32 = convert.(Float32, view(cz,:))
        g["x"] = xf32
        g["y"] = yf32
        g["z"] = zf32
        precision = 4
        write(buffer, """
<Grid>
<Topology TopologyType="3DSMesh" NumberOfElements="$nz $ny $nx"/>
<Geometry GeometryType="X_Y_Z">
<DataItem Name="x" Dimensions="$nz $ny $nx" Format="HDF" Precision="$precision"> data.h5:/mesh$i/x </DataItem>
<DataItem Name="y" Dimensions="$nz $ny $nx" Format="HDF" Precision="$precision"> data.h5:/mesh$i/y </DataItem>
<DataItem Name="z" Dimensions="$nz $ny $nx" Format="HDF" Precision="$precision"> data.h5:/mesh$i/z </DataItem>
</Geometry>
</Grid>
""")
    end

end


function write_mesh_binary(bin, buffer, mesh::Mesh2d)

    @unpack x, y = get_static_variables(mesh.cache)
    z = zeros(Float64, length(x))
    for (xi,yi,zi) in zip(x,y,z)
        write(bin, xi, yi, zi)
    end
    offset = 0
    for (k,(cx,_)) in enumerate(zip(eachcell(mesh,x),mesh.offsets))
        nx, ny, nz = size3d(size(cx))
        write(buffer, """
<Grid>
<Topology TopologyType="3DSMesh" NumberOfElements="$nz $ny $nx"/>
<Geometry GeometryType="XYZ">
<DataItem Name="xyz" Dimensions="$nz $ny $nx" Format="Binary" Seek="$offset"> xyz.bin </DataItem>
</Geometry>
</Grid>
""")
        offset += nx*ny*nz
    end

end


# function write_mesh(h5, buffer, mesh::Mesh3d)
#
#     x, y, z = mesh.coords
#     itx, ity, itz = eachcell(mesh,x), eachcell(mesh,y), eachcell(mesh,z)
#     for (i,(cx,cy,cz)) in enumerate(zip(itx,ity,itz,))
#         nx, ny, nz = size3d(size(cx))
#         g = create_group(h5, string(i))
#         g["x"] = cx
#         g["y"] = cy
#         g["z"] = cz
#         write(buffer, """
# <Grid>
# <Topology TopologyType="3DSMesh" NumberOfElements="$nz $ny $nx"/>
# <Geometry GeometryType="X_Y_Z">
# <DataItem Name="x" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/x </DataItem>
# <DataItem Name="y" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/y </DataItem>
# <DataItem Name="z" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/z </DataItem>
# </Geometry>
# </Grid>
# """)
#     end
#
# end


size3d(sz::NTuple{1}) = (sz[1],1,1)
size3d(sz::NTuple{2}) = (sz[1],sz[2],1)
# size3d(sz::NTuple{3}) = (sz[1],sz[2],sz[3])


function xdmfwrite_scalar(xdmf::XDMFOutput, mesh::Mesh, t, vars, names)

    buffers = IOBuffer[]
    for _ in vars
        b = IOBuffer()
        write(b, """
<Grid GridType="Collection" CollectionType="Temporal"><Grid GridType="Collection" CollectionType="Spatial">
<Time Value="$t"/>
""")
        push!(buffers, b)
    end

    tcount = xdmf.tcount + 1

    if xdmf.mode === :h5
        h5open(xdmf.data_filename, "r+") do h5
            write_scalar_h5(h5, buffers, mesh, tcount, vars, names)
        end
    else # xdmf.mode === :bin
        open(xdmf.data_filename, "a") do file
            b64io = Base64EncodePipe(file)
            xdmf.binoffset = write_scalar_binary(b64io, xdmf.binoffset, buffers, mesh, tcount, vars, names)
        end
    end

    # finish up xml files
    for (buffer,name) in zip(buffers,names)
        filename = joinpath(xdmf.dir, "$name.xdmf")
        existed = isfile(filename)
        open(filename, "a") do file
            if !existed
                print(file, """
    <?xml version="1.0" ?>
    <!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
    <Xdmf xmlns:xi="http://www.w3.org/2003/XInclude" Version="3.0">
    <Domain>
    """)
            else
                skip(file, -length(XDMF_END))
            end
            seekstart(buffer)
            write(file, buffer)
            write(file, XDMF_END)
        end
    end

    # updating count after all writes succeeded
    xdmf.tcount = tcount

end



function write_scalar_h5(h5, buffers, mesh, tcount, vars, names)

    celliters = [ eachcell(mesh, var) for var in vars ]

    # a group for the new time step
    h5group_t = create_group(h5, string(tcount))
    # sub groups for each cell
    groups = [ create_group(h5group_t, string(i)) for i = 1:ncells(mesh) ]


    for (cellit, name, var, buffer) in zip(celliters, names, vars, buffers)
        for (i,(cellvar,g)) in enumerate(zip(cellit,groups))
            nx, ny, nz = size3d(size(cellvar))
            # g[name] = view(cellvar,:)
            f32 = convert.(Float32, view(cellvar,:))
            g[name] = f32
            precision = 4
            write(buffer, """
<Grid Name="$i"><xi:include href="mesh.xdmf" xpointer="xpointer(/Xdmf/Domain/Grid[1]/Grid[$i]/child::*)"/><Attribute Name="$name" Center="Node"><DataItem Dimensions="$nz $ny $nx" Format="HDF" Precision="$precision"> data.h5:/$tcount/$i/$name </DataItem></Attribute></Grid>
""")
        end
        write(buffer, "</Grid></Grid>")
    end
end


function write_scalar_binary(bin, binoffset, buffers, mesh, tcount, vars, names)

    celliters = [ eachcell(mesh, var) for var in vars ]

    for (cellit, name, var, buffer) in zip(celliters, names, vars, buffers)
        write(bin, var)
        for (i,o) in enumerate(mesh.offsets)
            offset = binoffset + o
            idxs = cellindices(mesh,i)
            nnn = length(idxs)
            # TODO Set endianess
            write(buffer, """
<Grid Name="$i"><xi:include href="mesh.xdmf" xpointer="xpointer(/Xdmf/Domain/Grid[1]/Grid[$i]/child::*)"/><Attribute Name="$name"><DataItem Dimensions="$nnn 1" Format="Binary" Precision="8" Seek="$offset"> data.bin </DataItem></Attribute></Grid>
""")
        end
        write(buffer, "</Grid></Grid>")
    end

    binoffset += length(first(vars))

    return binoffset
end
