@testset "Evolution" begin


import dg1d: evolve


@testset "Convergence rate Runge Kutta schemes" begin

  # Test problem: df/dt = - α u = rhs(u), u(t0) = u0
  t0 = 0.0
  tend = 1.0
  u0 = [0.5]
  α = 0.8
  function rhs!(du, u, t)
    du .= - α * u
  end
  u(t) = u0 * exp(- α * (t - t0))

  u_ana = u(tend)

  schemes = [ :midpt, :rk4,       :rk4_twothirds, :rk4_ralston, :ssprk3, :lserk4, :rkf10 ]
  orders  = [ 2,      4,          4,              4,            3,       4,       10     ]

  dts = [0.1, 0.05, 0.025]

  err = zeros(length(dts), length(schemes))
  for (it, dt) in enumerate(dts)
    for (is, scheme) in enumerate(schemes)
      u_num = evolve(rhs!, deepcopy(u0), # need to reuse u0 between all tests
                     dt, tend, dg1d.algorithm(scheme))
      err[it,is] = norm(u_num - u_ana)
    end
  end

  for (is, scheme) in enumerate(schemes)
    for (it, dt) in enumerate(dts)
      if it < length(dts)
        e1 = err[it,is]
        e2 = err[end,is]
        if abs(e1 - e2) <= 1e-14 # cannot use such small errors to estimate convergence rate
          @test true
          continue
        end
        dt1 = dts[it]
        dt2 = dts[end]
        cvgrates = - log(e1/e2) / log(dt2/dt1)
        @test cvgrates > orders[is]
      end
    end
  end

end


@testset "Adaptive time stepping Runge Kutta scheme" begin

  # Test problem: df/dt = - α u = rhs(u), u(t0) = u0
  t0 = 0.0
  tend = 1.0
  u0 = [0.5]
  α = 15.0 # stiff ODE
  function rhs!(du, u, t)
    du .= - α * u
  end
  u(t) = u0 * exp(- α * (t - t0))

  u_ana = u(tend)

  schemes = [ :rk4, :rk4_twothirds, :rk4_ralston, :ssprk3, :lserk4 ]
  dts = [0.1, 0.05, 0.01]

  # solutions for different dts should all converge to the same result
  # hence, set reltol to zero to avoid influence from initial data
  reltol = 0.0
  abstol = 1e-14

  err = zeros(length(dts), length(schemes))
  for (it, dt) in enumerate(dts)
    for (is, scheme) in enumerate(schemes)
      pblm = dg1d.ODEProblem(scheme)
      u_num = dg1d.evolve_adaptive_ERK(pblm, rhs!, u0, dt, tend,
                                       reltol=reltol, abstol=abstol)
      err[it,is] = norm(u_num - u_ana)
    end
  end

  # for (it, dt) in enumerate(dts)
  #   for (is, scheme) in enumerate(schemes)
  #     u_num = dg1d.evolve_adaptive_ERK(rhs!, u0, dt, tspan, scheme=scheme;
  #                                      reltol=reltol, abstol=abstol)
  #     err[it,is] = norm(u_num - u_ana, 1)
  #   end
  # end
  #
  # for (is, scheme) in enumerate(schemes)
  #   for (it, dt) in enumerate(dts)
  #     if it <= length(dts)
  #       # for now only verify if correct result is obtained for different
  #       # initial time steps
  #       @test err[it,is] <= 1e-10
  #     end
  #   end
  # end

end


@testset "Test interfaces" begin

  # standard problem
  t0 = 0.0
  tend = 1.0
  dt = 0.1
  u0 = [0.5]
  α = 1.0
  function rhs!(du, u, t)
    du .= - α * u
  end

  called = false
  function callback_fullstep(u, t, n; force=false)
    called = true
  end

  alg = dg1d.algorithm(:rk4)
  u_num = evolve(rhs!, u0, dt, tend, alg, callback_fullstep=callback_fullstep)

  @test called == true

  # callbacks have wrong interface
  @test_throws ArgumentError evolve(rhs!, u0, dt, tend, alg,
                                    callback_fullstep=(u,t)->nothing)
  @test_throws ArgumentError evolve(rhs!, u0, dt, tend, alg,
                                    callback_substep=(u,)->nothing)
  @test_throws ArgumentError evolve(rhs!, u0, dt, tend, alg,
                                    callback_fullstep=(u,t::String,n::Int)->nothing)
  @test_throws ArgumentError evolve(rhs!, u0, dt, tend, alg,
                                    callback_substep=(u,t::Vector{Real},n::Int)->nothing)
  @test_throws ArgumentError evolve(rhs!, u0, dt, tend, alg, cfl=-0.1)
end


@testset "Dynamic time step interface" begin

  # standard problem
  t0 = 0.0
  tend = 1.0
  u0 = [0.5]
  α = 1.0
  function rhs!(du, u, t)
    du .= - α * u
  end

  # dynamic time stepping
  dt(u, t, n) = t < 0.5 ? 0.1 : 0.05

  schemes = [ :rk4, :rk4_twothirds, :rk4_ralston, :ssprk3, :lserk4 ]

  for scheme in schemes
    # pblm = dg1d.ODEProblem(scheme)
    # u_num = dg1d.evolve_ERK(pblm, rhs!, u0, dt, tend)
    u_num = evolve(rhs!, u0, dt, tend, dg1d.algorithm(scheme))
  end

end


# @testset "Substep interface" begin
#
#   # standard problem
#   t0 = 0.0
#   tend = 1.0
#   u0 = [0.5]
#   α = 1.0
#   function rhs!(du, u, t)
#     du .= - α * u
#   end
#
#   # To test rejection and time step correction we artificially reject some time steps.
#   n_fails = 5
#   # we trigger a fail once these times are passed
#   ts_fail = [ (tend-t0)/n_fails * (i-1) for i = 1:n_fails ]
#   failed_once = false
#   test_n_failed = 0 # for testing purposes
#   function callback_substep(u, t, n_stage)
#     if test_n_failed < n_fails && t >= ts_fail[test_n_failed+1]
#       test_n_failed += 1
#       failed_once = true
#       return false
#     end
#     return true
#   end
#
#   # once failed we can recover by adjusting the next time step
#   base_dt = 0.05
#   function tstep(u, t, n_fails)
#     if !failed_once
#       # in this example we have to free from the previous fail manually
#       failed_once = true
#     end
#     return base_dt / (n_fails+1)
#   end
#
#   schemes = [ :rk4, :rk4_twothirds, :rk4_ralston, :ssprk3, :lserk4 ]
#   for scheme in schemes
#     test_n_failed = 0 # reset counter
#     u_num = evolve(rhs!, deepcopy(u0), # need to reuse u0 between all tests
#                    tstep, tend, dg1d.algorithm(scheme),
#                    callback_substep=callback_substep)
#     @test test_n_failed == n_fails
#   end
#
# end


end
