@testset "Mesh" begin


@testset "Factories" begin

  # or keyword arguments
  mesh = Mesh1d(N=4, K=10)

  # or just the defaults
  mesh = Mesh2d()

end


@testset "Mesh quadrature" begin

  # ∫(0≤x≤2π) sin(2 π x) dx = 0
  mesh = Mesh1d(N=10, K=100)
  u = @. sin(2.0*π*mesh.x)
  grid_avg = grid_average(u, mesh)
  result = 0.0
  @test isapprox(grid_avg, result, atol=1e-12)

  # ∫(0≤x≤2π) (sin(2 π x) + 1) dx = 1
  u = @. mesh.x + 1.0
  grid_avg = grid_average(u, mesh)
  result = 1.0
  @test isapprox(grid_avg, result, atol=1e-12)

  # ∫(xl≤x≤xr) C dx / dl = c, where dl = xr - xl and xl,xr are the cell boundaries
  C = 2.0
  u = C .* ones(size(mesh.x))
  cell_avg = cellwise_average(u, mesh)
  result = C .* ones(mesh.K)
  @test all(isapprox.(cell_avg, result))

  # ∫(xl≤x≤xr) C^2 dx = C^2 dl, where dl = xr - xl and xl,xr are the cell boundaries
  C = 2.0
  u = C .* ones(size(mesh.x))
  cell_iproduct = cellwise_inner_product(u, u, mesh)
  result = C^2 .* mesh.dl .* ones(mesh.K)
  @test all(isapprox.(cell_iproduct, result))


  # Σ(cells) ∫(xl≤x≤xr) C^2 dx = C^2 L, where dl = xr - xl and xl,xr are the cell boundaries
  # =^= Σ(cells) cellwise_inner_product(u, u, mesh)
  C = 2.0
  u = C .* ones(size(mesh.x))
  broken_iproduct = broken_inner_product(u, u, mesh)
  result = C^2 * mesh.L
  @test isapprox(broken_iproduct, result)

end


@testset "Utilities" begin

  N, K = 5, 10
  mesh = Mesh1d(N=N, K=K)

  #### dimensions, length, layout
  npts, ncells = layout(mesh)
  @test npts == N+1
  @test ncells == K
  len = length(mesh)
  @test len == (N+1)*K

  #### find cells given nodes on mesh
  mesh = Mesh1d(N=2, K=10) # =^= equidistant grid with doubled nodes at interfaces
  @unpack L, K = mesh
  # as a sample we place a node into to center of every cell
  dL = L / K
  smpl_x = [ (idx-1/2) * dL + mesh.xmin  for idx = 1:K ]

  expected = collect(1:K) # expected cell indices
  observed = find_cell_index.(smpl_x, Ref(mesh))
  @test all(observed .== expected)

  # node does not lie on mesh
  smpl_node_outside = -10.0
  observed = find_cell_index(smpl_node_outside, mesh)
  expected = nothing
  @test observed == nothing

  # if node falls onto cell interface, then the first cell index is returned
  smpl_node_interface_lhs = mesh.x[end,1]
  smpl_node_interface_rhs = mesh.x[1,2]
  @test isapprox(smpl_node_interface_lhs, smpl_node_interface_rhs)
  observed_lhs = find_cell_index(smpl_node_interface_lhs, mesh)
  expected_lhs = 1
  observed_rhs = find_cell_index(smpl_node_interface_rhs, mesh)
  expected_rhs = 1
  @test observed_lhs == expected_lhs
  @test observed_rhs == expected_rhs

  #### locate closest point to given coordiante
  smpl_node = mesh.x[1]
  observed  = locate_point(smpl_node, mesh)
  expected  = 1
  @test isapprox(observed, expected)

  # some where between point 4 and 5, but closer to 5
  smpl_node = 0.3 * mesh.x[4] + 0.7 * mesh.x[5]
  observed  = locate_point(smpl_node, mesh)
  expected  = 5
  @test isapprox(observed, expected)

  # edge case when points lie outside of mesh
  smpl_node_lhs, smpl_node_rhs = -10.0, 10.0
  observed_lhs, observed_rhs = locate_point.([smpl_node_lhs, smpl_node_rhs], Ref(mesh))
  expected_lhs, expected_rhs = 1, length(mesh)
  @test observed_lhs == expected_lhs
  @test observed_rhs == expected_rhs

end


@testset "MeshInterpolator" begin

  smpl_mesh = Mesh1d(N=4, K=6)
  ref_mesh  = Mesh1d(N=10, K=20)
  ref_data = sin.(ref_mesh.x)
  meshintrp = MeshInterpolator(smpl_mesh, ref_mesh)
  observed = meshintrp(ref_data)
  expected = sin.(smpl_mesh.x)
  @test all(isapprox.(observed, expected))

  # inplace version
  obvserved = zeros(layout(smpl_mesh))
  meshintrp(observed, ref_data)
  @test all(isapprox.(observed, expected))

  #### edge cases
  # sample mesh lies (partly) outside of reference mesh
  smpl_mesh_outside = Mesh1d(N=4, K=6, range=[-10.0,0.0])
  ref_mesh  = Mesh1d(N=10, K=20)
  @test_throws ErrorException MeshInterpolator(smpl_mesh_outside, ref_mesh)

  # reference data does not match reference mesh
  ref_data_wrong_size = sin.(ref_mesh.x)[1:end-1]
  @test_throws ErrorException meshintrp(ref_data_wrong_size)

  # inplace storage does not match sample mesh
  storge_intrp_data_wrong_size = zeros(length(smpl_mesh)-1)
  @test_throws ErrorException meshintrp(storge_intrp_data_wrong_size, ref_data)

end


end
