@testset "Rootfinders" begin


function make_problems()
  problems = [
              # name       function    derivative   search interval   root
              ("x",           x->(x, 1),            (-1.2, 1.5),      0.0),
              ("x^2-1",       x->(-x^2+1, -2*x),    (-1.3, -0.8),     -1.0),
              ("-x^2+1",      x->(x^2-1, 2*x),      (-1.3, -0.8),     -1.0),
              ("x^2-π",       x->(x^2-π, 2*x),      (1.0, 2.0),       sqrt(π)),
              ("cos(x)-x^3",  x->(cos(x)-x^3,                         # obtained with NewtonSolver
                                  -sin(x)-3*x^2),   (0.0,1.0),        0.8654740331110141)
              ]
  return problems
end


@testset "Bisection" begin

  # test with standard problems
  problems = make_problems()

  method = dg1d.Bisection(maxits=35)
  for p in problems
    name, f_df, xrange, xroot = p
    try
      x, fx, converged, tol = find_root(f_df, xrange, method)
      if !converged
        @warn "'$name' test failed"
      end
      @test converged
      !ismissing(xroot) && @test isapprox(x, xroot, atol=method.abstol)
      @test isapprox(fx,  0.0, atol=tol)
    catch e
      @warn "'$name' test errored"
      rethrow(e)
    end
  end

  # call with invalid interval
  @test_throws ErrorException find_root(x->x, (1.0,-1.0), method)

  # root outside the interval cannot converge
  x, fx, converged, tol = find_root(x->x, (1.0,2.0), method)
  @test !converged

end


@testset "Newton Solver with analytical derivative" begin

  # test with standard problems
  problems = make_problems()

  method = dg1d.NewtonSolver(approx_derivative=false)
  for p in problems
    name, f_df, xrange, xroot = p
    try
      xl, xr = xrange
      x0 = (xl + xr) / 2
      x, fx, converged, tol = find_root(f_df, x0, method)
      if !converged
        @warn "'$name' test failed"
      end
      @test converged
      !ismissing(xroot) && @test isapprox(x, xroot, atol=method.abstol)
      @test isapprox(fx,  0.0, atol=tol)
    catch e
      @warn "'$name' test errored"
      rethrow(e)
    end
  end

  # edge cases
  f_df(x) = tuple(x^2 - π, 2 * x)
  x0 = 0.0 # bad guess
  x, fx, converged, tol = find_root(f_df, x0, method)
  @test converged == false

end


@testset "Newton Solver with numerical derivative" begin

  # test with standard problems
  problems = make_problems()

  method = dg1d.NewtonSolver(approx_derivative=true)
  for p in problems
    name, f_df, xrange, xroot = p
    try
      xl, xr = xrange
      x0 = (xl + xr) / 2
      x, fx, converged, tol = find_root(f_df, x0, method)
      if !converged
        @warn "'$name' test failed"
      end
      @test converged
      !ismissing(xroot) && @test isapprox(x, xroot, atol=method.abstol)
      @test isapprox(fx,  0.0, atol=tol)
    catch e
      @warn "'$name' test errored"
      rethrow(e)
    end
  end

  # edge cases
  f_df(x) = tuple(x^2 - π, 2 * x)
  x0 = 0.0 # bad guess
  x, fx, converged, tol = find_root(f_df, x0, method)
  @test converged == false

end


end
