@testset "Spectral Element" begin


function evaluate_derivatives(el::SpectralElement, f, df)
  f_ana = f.(el.z)
  df_ana = df.(el.z)
  df_num = el.D * f_ana
  return df_ana, df_num
end


@testset "Nodal derivatives" begin

  N = 4
  f(x) = x^3 - 5.0 * x^2 + 1.0
  df(x) = 3.0 * x^2 - 10.0 * x
  # ddf(x) = 6.0 * x - 10.0

  el = SpectralElement(N, :LGL)
  df_ana, df_num = evaluate_derivatives(el, f, df)
  @test isapprox(df_ana, df_num)

  el = SpectralElement(N, :LGL_lumped)
  df_ana, df_num = evaluate_derivatives(el, f, df)
  @test isapprox(df_ana, df_num)

  glgl = SpectralElement(N, :LGL)
  df_ana, df_num = evaluate_derivatives(el, f, df)
  @test isapprox(df_ana, df_num)

end


@testset "Quadrature check" begin

  # integrate f^2 over x ∈ [-1,1]
  f(x) = x^3 - 5.0 * x^2 + 1.0
  f2_ana = 118.0/21.0 # according to wolfram alpha

  # use quadrature order N+1 = 3+1 = 4 here, because :LGL and :LGL_lumped are only exact up
  # to and including 2N-1
  N = 4
  L = 2.0
  quadrs = [ :LGL, :GLGL, :LGL_lumped ]

  for quadr in quadrs
    el = SpectralElement(N, quadr)
    fx = f.(el.z)
    f2 = dg1d.inner_product(fx, fx, el)
    @test isapprox(f2, f2_ana)
  end

  # check that :GLGL quadrature is exact up to and including order 2N
  N = 3
  L = 2.0
  el = SpectralElement(N, :GLGL)
  fx = f.(el.z)
  f2 = dg1d.inner_product(fx, fx, el)
  @test isapprox(f2, f2_ana)

end


@testset "Barycentric interpolation" begin

  # setup
  n_pts = 6
  x, _ = dg1d.LGL.rule(n_pts)
  bw = dg1d.barycentric_weights(x)

  # test function
  f(x) = x
  # f(x) = x^4 - 2*x + 1
  fx = @. f(x)

  # test interpolation on random grid
  nsmpl = 3
  xsmpl = rand(nsmpl) # ∈ [0,1]
  @. xsmpl = 2 * xsmpl - 1 # ∈ [-1,1]
  fxsmpl = @. f(xsmpl)

  # test single point interpolation
  fxintrp = [ dg1d.barycentric_interp(xi, x, bw, fx) for xi in xsmpl ]
  @test all(isapprox.(fxintrp, fxsmpl))

  # test interpolation using matrix multiplication 
  I = dg1d.barycentric_interp_matrix(xsmpl, x)
  fxintrp = I * fx
  @test all(isapprox.(fxintrp, fxsmpl))

  # edge case: interpolate onto nodes themselves
  xsmpl = x
  fxsmpl = fx
  fxintrp = [ dg1d.barycentric_interp(xi, x, bw, fx) for xi in xsmpl ]
  @test all(isapprox.(fxintrp, fxsmpl))

  I = dg1d.barycentric_interp_matrix(xsmpl, x)
  fxintrp = I * fx
  @test all(isapprox.(fxintrp, fxsmpl))

end


@testset "Operators GLGL" begin

  # set up quadrature rule
  N = 5
  N_pts = N + 1
  x, w, v = dg1d.GLGL.rule(N_pts)
  D = dg1d.first_derivative_matrix(x)

  # analytical result
  L_ana = zeros(N_pts, N_pts)
  for i in 1:N_pts
    L_ana[i,i] = 2.0 / (2.0 * i - 1.0)
  end

  L = dg1d.GLGL.linear_dependency_matrix(x, w, v, D)

  @test isapprox(L, L_ana)

end


@testset "Operators LGL" begin

  # set up quadrature rule
  n_pts = 6
  x, w = dg1d.LGL.rule(n_pts)
  quad(f1, f2) = dg1d.LGL.inner_product(w, f1, f2)

  # analytical result
  L_ana = zeros(n_pts, n_pts)
  for i in 1:n_pts
    L_ana[i,i] = 2.0 / (2.0 * i - 1.0)
  end

  L = dg1d.LGL.linear_dependency_matrix_exact(x, w)

  @test isapprox(L, L_ana)

end


@testset "Test orthogonality relation" begin

  # set up quadrature rule
  n_pts = 6
  x, w = dg1d.LGL.rule(n_pts)
  quad(f1, f2) = dg1d.LGL.inner_product(w, f1, f2)

  # analytical result
  L_ana = zeros(n_pts, n_pts)
  for i in 1:n_pts
    L_ana[i,i] = 2.0 / (2.0 * i - 1.0)
  end

  L = dg1d.LGL.linear_dependency_matrix_exact(x, w)

  @test isapprox(L, L_ana)

end


end
