@testset "@with_signature legacy" begin

  struct DummyEquation end
  dummyeq = DummyEquation()

  # general usage
  @with_signature function fn_general(eq::DummyEquation)
    @accepts u1, u2
    rhs_u1, rhs_u2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns rhs_u1, rhs_u2
  end
  @test accepts(fn_general, dummyeq) == ((:u1, :u2),)
  @test returns(fn_general, dummyeq) == (:rhs_u1, :rhs_u2)
  @test state_indices(fn_general, dummyeq) == ()
  @test has_signature(fn_general, dummyeq) == true
  args = (1.0, 2.0)
  obtained = fn_general(args, dummyeq)
  expected = (4.0, 5.0)
  @test obtained == expected

  # @accepts and @returns with only one argument each
  @with_signature function fn_accepts1(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn_accepts1, dummyeq) == ((:u1,),)
  @test returns(fn_accepts1, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_accepts1, dummyeq) == ()
  @test has_signature(fn_accepts1, dummyeq) == true
  args = (1.0,)
  obtained = fn_accepts1(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # multiple @accepts statements
  @with_signature function fn_accepts2(eq::DummyEquation)
    @accepts u1
    @accepts v1, v2
    rhs_u1 = 2.0 * u1
    rhs_v1 = -3.0 * v1
    rhs_v2 = -3.0 * v2
    @returns rhs_u1, rhs_v1, rhs_v2
  end
  @test accepts(fn_accepts2, dummyeq) == ((:u1,),(:v1,:v2))
  @test returns(fn_accepts2, dummyeq) == (:rhs_u1,:rhs_v1,:rhs_v2)
  @test state_indices(fn_accepts2, dummyeq) == ()
  @test has_signature(fn_accepts2, dummyeq) == true
  args1 = (1.0,)
  args2 = (2.0,3.0)
  obtained = fn_accepts2(args1, args2, dummyeq)
  expected = (2.0,-6.0,-9.0)
  @test obtained == expected

  # Prefix @accepts variables
  @with_signature function fn_prefix1(eq::DummyEquation)
    @accepts Prefix(tmp), u1
    rhs_u1 = 2.0 * tmp_u1
    @returns rhs_u1
  end
  @test accepts(fn_prefix1, dummyeq) == ((:u1,),)
  @test returns(fn_prefix1, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_prefix1, dummyeq) == ()
  @test has_signature(fn_prefix1, dummyeq) == true
  args = (1.0,)
  obtained = fn_prefix1(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # Prefix multiple @accepts variables
  @with_signature function fn_prefix2(eq::DummyEquation)
    @accepts Prefix(lhs), u1, u2
    @accepts Prefix(rhs), u1, u2
    @accepts n1, n2
    flx_u1 = n1 * lhs_u1 + n2 * rhs_u1
    flx_u2 = n1 * lhs_u2 + n2 * rhs_u2
    @returns flx_u1, flx_u2
  end
  @test accepts(fn_prefix2, dummyeq) == ((:u1,:u2),(:u1,:u2),(:n1,:n2))
  @test returns(fn_prefix2, dummyeq) == (:flx_u1,:flx_u2)
  @test state_indices(fn_prefix2, dummyeq) == ()
  @test has_signature(fn_prefix2, dummyeq) == true
  args1 = (1.0,-1.0)
  args2 = (2.0,-2.0)
  args3 = (1.0,2.0)
  obtained = fn_prefix2(args1, args2, args3, dummyeq)
  expected = (5.0,-5.0)
  @test obtained == expected

  # @accepts State variables
  @with_signature function fn_state1(eq::DummyEquation)
    @accepts State(u1, u2)
    rhs_u1, rhs_u2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns rhs_u1, rhs_u2
  end
  @test accepts(fn_state1, dummyeq) == ((:u1, :u2),)
  @test returns(fn_state1, dummyeq) == (:rhs_u1, :rhs_u2)
  @test state_indices(fn_state1, dummyeq) == (1,2)
  @test has_signature(fn_state1, dummyeq) == true
  args = (1.0, 2.0)
  obtained = fn_state1(args, dummyeq)
  expected = (4.0, 5.0)
  @test obtained == expected

  # @accepts single State variable
  @with_signature function fn_state2(eq::DummyEquation)
    @accepts State(u1)
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn_state2, dummyeq) == ((:u1,),)
  @test returns(fn_state2, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_state2, dummyeq) == (1,)
  @test has_signature(fn_state2, dummyeq) == true
  args = (1.0,)
  obtained = fn_state2(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # @accepts mix of variables and State variables
  @with_signature function fn_state3(eq::DummyEquation)
    @accepts State(u1), u2, State(u3, u4)
    rhs_u1 = 2.0 * u1 + u2 + u3 + u4
    @returns rhs_u1
  end
  @test accepts(fn_state3, dummyeq) == ((:u1,:u2,:u3,:u4),)
  @test returns(fn_state3, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_state3, dummyeq) == (1,3,4)
  @test has_signature(fn_state3, dummyeq) == true
  args = (1.0,2.0,3.0,4.0)
  obtained = fn_state3(args, dummyeq)
  expected = (11.0,)
  @test obtained == expected

  # no variable name for dispatch type required
  @with_signature function fn_noargname(#= missing name here =#::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn_noargname, dummyeq) == ((:u1,),)
  @test returns(fn_noargname, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_noargname, dummyeq) == ()
  @test has_signature(fn_noargname, dummyeq) == true
  args = (1.0,)
  obtained = fn_noargname(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # function with unused variable rhs_u2 in body throws normal UndefVarError
  @with_signature function fn_undefvar(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1, rhs_u2
  end
  args = (1.0,)
  @test_throws UndefVarError fn_undefvar(args, dummyeq)

  # missing function body error
  @test_throws LoadError @eval @with_signature function fn(eq::DummyEquation) end

  # @accepts/@returns must appear at the start/end of the function body
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    println("this line is not allowed")
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
    println("this line is not allowed")
  end

  # @accepts/@returns can only appear once
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1, u2
    rhs_u1 = 2.0 * u1
    rhs_u2 = 2.0 * u2
    @returns rhs_u1, rhs_u2
    @returns rhs_u1, rhs_u2
  end

  # can't mix State and Prefix
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts Prefix(u1), State(u2)
    rhs_u1 = 2.0 * u1
    rhs_u2 = 2.0 * u2
    @returns rhs_u1, rhs_u2
  end
  # can't mix State and Prefix
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts State(u1), u2, Prefix(u3)
    rhs_u1 = 2.0 * u1
    rhs_u2 = 4.0 * u2
    rhs_u3 = 5.0 * u3
    @returns rhs_u1, rhs_u2, rhs_u3
  end

  # Prefix must appear as the first argument in a @accepts line
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1, Prefix(lhs)
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    @accepts u2, Prefix(lhs)
    rhs_u1 = 2.0 * u1
    rhs_u2 = 4.0 * u2
    @returns rhs_u1, rhs_u2
  end
  # Prefix requires at least one variable following it
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts Prefix(lhs)
    @accepts u2
    rhs_u1 = 2.0
    rhs_u2 = 4.0 * u2
    @returns rhs_u1, rhs_u2
  end

  # State can only be used with one @accepts statement
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts State(u1), u2
    @accepts u3
    rhs_u1 = 2.0 * u1
    rhs_u2 = 4.0 * u2
    rhs_u3 = 5.0 * u3
    @returns rhs_u1, rhs_u2, rhs_u3
  end

  # @accepts and @returns require each at least one argument
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts
    rhs_u1 = 1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns
  end

  # function definition requires exactly one argument
  @test_throws LoadError @eval @with_signature function fn_w_error(#=no argument=#)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation, too_much)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end

  # the only argument must be type annotated
  @test_throws LoadError @eval @with_signature function fn_w_error(eq)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end

  # fallback interface methods error when called
  function fn_with_no_signature(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws ErrorException accepts(fn_with_no_signature,dummyeq)
  @test_throws ErrorException returns(fn_with_no_signature,dummyeq)
  @test_throws ErrorException state_indices(fn_with_no_signature,dummyeq)
  # expect the trait function is just false
  has_signature(fn_with_no_signature, dummyeq) == false

end


function make_data(N, nin, nout)
  args = Tuple( rand(N) for _ = 1:nin )
  rets = Tuple( rand(N) for _ = 1:nout )
  return args, rets, 1:N, 1:N
end


@testset "@with_signature" begin

  N = 100

  struct DummyEquation end
  dummyeq = DummyEquation()

  # general usage
  @with_signature [legacy=false] function fn_general(eq::DummyEquation)
    @accepts u1, u2
    ru1, ru2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns ru1, ru2
  end
  @test accepts(fn_general, dummyeq, nothing) == (:u1, :u2)
  @test returns(fn_general, dummyeq, nothing) == (:ru1, :ru2)
  @test has_signature(fn_general, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 2, 2)
  fn_general(dummyeq, args, rets, vidxs, bidxs)
  u1, u2 = args
  ru1, ru2 = deepcopy(rets)
  @. ru1 = 2.0 * u1 + u2
  @. ru2 = u2 * 3.0 - u1
  expected = (ru1, ru2)
  @test rets == expected


  # @accepts and @returns with only one argument each
  @with_signature [legacy=false] function fn_accepts1(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test accepts(fn_accepts1, dummyeq, nothing) == (:u1,)
  @test returns(fn_accepts1, dummyeq, nothing) == (:ru1,)
  @test has_signature(fn_accepts1, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_accepts1(dummyeq, args, rets, vidxs, bidxs)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1
  expected = (ru1,)
  @test rets == expected

  # multiple @accepts statements
  @with_signature [legacy=false] function fn_accepts2(eq::DummyEquation)
    @accepts u1
    @accepts v1, v2
    ru1 = 2.0 * u1
    rv1 = -3.0 * v1
    rv2 = -3.0 * v2
    @returns ru1, rv1, rv2
  end
  @test accepts(fn_accepts2, dummyeq, nothing) == (:u1,:v1,:v2)
  @test returns(fn_accepts2, dummyeq, nothing) == (:ru1,:rv1,:rv2)
  @test has_signature(fn_accepts2, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 3, 3)
  fn_accepts2(dummyeq, args, rets, vidxs, bidxs)
  u1, v1, v2 = args
  ru1, rv1, rv2 = deepcopy(rets)
  @. ru1 = 2.0 * u1
  @. rv1 = -3.0 * v1
  @. rv2 = -3.0 * v2
  expected = (ru1, rv1, rv2)
  @test rets == expected

  # Prefix @accepts variables
  @with_signature [legacy=false] function fn_prefix1(eq::DummyEquation)
    @accepts [Prefix(tmp)] u1
    ru1 = 2.0 * tmp_u1
    @returns ru1
  end
  @test accepts(fn_prefix1, dummyeq, nothing) == (:u1,)
  @test returns(fn_prefix1, dummyeq, nothing) == (:ru1,)
  @test has_signature(fn_prefix1, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_prefix1(dummyeq, args, rets, vidxs, bidxs)
  u1,  = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1
  expected = (ru1,)
  @test rets == expected

  # Prefix multiple @accepts variables
  @with_signature [legacy=false] function fn_prefix2(eq::DummyEquation)
    @accepts [Prefix(lhs)] u1, u2
    @accepts [Prefix(rhs)] u1, u2
    @accepts n1, n2
    flx_u1 = n1 * lhs_u1 + n2 * rhs_u1
    flx_u2 = n1 * lhs_u2 + n2 * rhs_u2
    @returns flx_u1, flx_u2
  end
  @test accepts(fn_prefix2, dummyeq, nothing) == (:u1,:u2,:u1,:u2,:n1,:n2)
  @test returns(fn_prefix2, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_prefix2, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 6, 2)
  fn_prefix2(dummyeq, args, rets, vidxs, bidxs)
  lhs_u1, lhs_u2, rhs_u1, rhs_u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * lhs_u1 + n2 * rhs_u1
  @. flx_u2 = n1 * lhs_u2 + n2 * rhs_u2
  expected = (flx_u1, flx_u2)
  @test rets == expected

  # no variable name for dispatch type required
  @with_signature [legacy=false] function fn_noargname(#= missing name here =#::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test accepts(fn_noargname, dummyeq, nothing) == (:u1,)
  @test returns(fn_noargname, dummyeq, nothing) == (:ru1,)
  @test has_signature(fn_noargname, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_noargname(dummyeq, args, rets, vidxs, bidxs)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1
  expected = (ru1,)
  @test rets == expected

  # using only bdry variables
  @with_signature [legacy=false] function fn_bdry1(eq::DummyEquation)
    @accepts [bdry] u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_bdry1, dummyeq, nothing) == (:u1,:u2,:n1,:n2)
  @test returns(fn_bdry1, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_bdry1, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  fn_bdry1(dummyeq, args, rets, vidxs, bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1 + n2 * u1
  @. flx_u2 = n1 * u2 + n2 * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # using volume and bdry variables with prefixes
  @with_signature [legacy=false] function fn_bdry2(eq::DummyEquation)
    @accepts u1
    @accepts [Prefix(test)] u2
    @accepts [Prefix(foo), bdry] n1, n2
    bar_flx_u1 = foo_n1 * u1 + foo_n2 * u1
    foo_flx_u2 = foo_n1 * test_u2 + foo_n2 * test_u2
    @returns [Prefix(bar)] flx_u1
    @returns [bdry,Prefix(foo)] flx_u2
  end
  @test accepts(fn_bdry2, dummyeq, nothing) == (:u1,:u2,:n1,:n2)
  @test returns(fn_bdry2, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_bdry2, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  fn_bdry1(dummyeq, args, rets, vidxs, bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1 + n2 * u1
  @. flx_u2 = n1 * u2 + n2 * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # using only bdry variables, but with shuffled indices
  # simple version first: compute @. flx_u1[rng_bidxs] = u1[rng_bidxs], which
  # does not alter the ordering of values, but instead just computes in different order
  @with_signature [legacy=false] function fn_bdry3(eq::DummyEquation)
    @accepts [bdry] u1
    flx_u1 = u1
    @returns [bdry] flx_u1
  end
  @test accepts(fn_bdry3, dummyeq, nothing) == (:u1,)
  @test returns(fn_bdry3, dummyeq, nothing) == (:flx_u1,)
  @test has_signature(fn_bdry3, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(10, 1, 1)
  rng_bidxs = Random.shuffle(bidxs)
  fn_bdry3(dummyeq, args, rets, vidxs, rng_bidxs)
  u1, = args
  flx_u1, = deepcopy(rets)
  @. flx_u1 = u1
  expected = (flx_u1,)
  @test rets == expected

  # using only bdry variables, but with shuffled indices
  @with_signature [legacy=false] function fn_bdry4(eq::DummyEquation)
    @accepts [bdry] u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_bdry4, dummyeq, nothing) == (:u1,:u2,:n1,:n2)
  @test returns(fn_bdry4, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_bdry4, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_bidxs = Random.shuffle(bidxs)
  fn_bdry4(dummyeq, args, rets, vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1 + n2 * u1
  @. flx_u2 = n1 * u2 + n2 * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use volume and bdry variables with shuffled bdry indices
  @with_signature [legacy=false] function fn_volbdry1(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry1, dummyeq, nothing) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry1, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry1, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_bidxs = Random.shuffle(bidxs)
  fn_volbdry1(dummyeq, args, rets, vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1[rng_bidxs] = n1[rng_bidxs] * u1 + n2[rng_bidxs] * u1
  @. flx_u2[rng_bidxs] = n1[rng_bidxs] * u2 + n2[rng_bidxs] * u2
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use volume and bdry variables with shuffled volume indices
  @with_signature [legacy=false] function fn_volbdry2(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry2, dummyeq, nothing) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry2, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry2, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_vidxs = Random.shuffle(vidxs)
  fn_volbdry2(dummyeq, args, rets, rng_vidxs, bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1 = n1 * u1[rng_vidxs] + n2 * u1[rng_vidxs]
  @. flx_u2 = n1 * u2[rng_vidxs] + n2 * u2[rng_vidxs]
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use volume and bdry variables with shuffled volume and bdry indices
  @with_signature [legacy=false] function fn_volbdry3(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry3, dummyeq, nothing) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry3, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry3, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 4, 2)
  rng_vidxs = Random.shuffle(vidxs)
  rng_bidxs = Random.shuffle(bidxs)
  fn_volbdry3(dummyeq, args, rets, rng_vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1[rng_bidxs] = n1[rng_bidxs] * u1[rng_vidxs] + n2[rng_bidxs] * u1[rng_vidxs]
  @. flx_u2[rng_bidxs] = n1[rng_bidxs] * u2[rng_vidxs] + n2[rng_bidxs] * u2[rng_vidxs]
  expected = (flx_u1,flx_u2)
  @test rets == expected

  # use longer volume and bdry variables than (shuffled) volume and bdry index ranges
  @with_signature [legacy=false] function fn_volbdry4(eq::DummyEquation)
    @accepts u1, u2
    @accepts [bdry] n1, n2
    flx_u1 = n1 * u1 + n2 * u1
    flx_u2 = n1 * u2 + n2 * u2
    @returns [bdry] flx_u1, flx_u2
  end
  @test accepts(fn_volbdry3, dummyeq, nothing) == (:u1,:u2,:n1,:n2)
  @test returns(fn_volbdry3, dummyeq, nothing) == (:flx_u1,:flx_u2)
  @test has_signature(fn_volbdry3, dummyeq, nothing) == true
  args, rets, _, _ = make_data(N, 4, 2)
  rng_vidxs = Random.shuffle(1:div(N,2))
  rng_bidxs = Random.shuffle(rng_vidxs) .+ 10 # reshuffled and offset
  fn_volbdry3(dummyeq, args, rets, rng_vidxs, rng_bidxs)
  u1, u2, n1, n2 = args
  flx_u1, flx_u2 = deepcopy(rets)
  @. flx_u1[rng_bidxs] = n1[rng_bidxs] * u1[rng_vidxs] + n2[rng_bidxs] * u1[rng_vidxs]
  @. flx_u2[rng_bidxs] = n1[rng_bidxs] * u2[rng_vidxs] + n2[rng_bidxs] * u2[rng_vidxs]
  expected = (flx_u1[rng_bidxs],flx_u2[rng_bidxs])
  obtained = (rets[1][rng_bidxs],rets[2][rng_bidxs])
  @test obtained == expected


  # test options
  # debug=true disables all 'dangerous' optimizations like @inbounds, @simd
  @with_signature [legacy=false,debug=true] function fn_debug1(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1 + 1.0
    @returns ru1
  end
  @test accepts(fn_debug1, dummyeq, nothing) == (:u1,)
  @test returns(fn_debug1, dummyeq, nothing) == (:ru1,)
  @test has_signature(fn_debug1, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_debug1(dummyeq, args, rets, vidxs, bidxs)
  (u1,), (ru1,) = args, deepcopy(rets)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1 + 1.0
  expected = (ru1,)
  @test rets == expected

  # test options
  @with_signature [legacy=false,simd=false] function fn_debug2(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1 + 1.0
    @returns ru1
  end
  @test accepts(fn_debug2, dummyeq, nothing) == (:u1,)
  @test returns(fn_debug2, dummyeq, nothing) == (:ru1,)
  @test has_signature(fn_debug2, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_debug2(dummyeq, args, rets, vidxs, bidxs)
  (u1,), (ru1,) = args, deepcopy(rets)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1 + 1.0
  expected = (ru1,)
  @test rets == expected


  # test options
  @with_signature [legacy=false,inbounds=false] function fn_debug3(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1 + 1.0
    @returns ru1
  end
  @test accepts(fn_debug3, dummyeq, nothing) == (:u1,)
  @test returns(fn_debug3, dummyeq, nothing) == (:ru1,)
  @test has_signature(fn_debug3, dummyeq, nothing) == true
  args, rets, vidxs, bidxs = make_data(N, 1, 1)
  fn_debug3(dummyeq, args, rets, vidxs, bidxs)
  (u1,), (ru1,) = args, deepcopy(rets)
  u1, = args
  ru1, = deepcopy(rets)
  @. ru1 = 2.0 * u1 + 1.0
  expected = (ru1,)
  @test rets == expected


  ##### error handling

  # error on invalid options syntax
  @test_throws LoadError @eval @with_signature { legacy=true,legacy=false } function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  @test_throws LoadError @eval @with_signature legacy=true function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end

  # error on duplicated @with_signature options
  @test_throws LoadError @eval @with_signature [legacy=true,legacy=false] function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  @test_throws LoadError @eval @with_signature [skip_debug_info=true,legacy=false,legacy=false] function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  # error on invalid type for option
  @test_throws LoadError @eval @with_signature [legacy=1] function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  @test_throws LoadError @eval @with_signature [unknown_option=true,legacy=false] function fn_unknown_option(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # function with unused variable ru2 in body throws normal UndefVarError
  @with_signature [legacy=false] function fn_undefvar(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1, undef_var
  end
  args, rets, vidxs, bidxs = make_data(N, 1, 2)
  @test_throws UndefVarError fn_undefvar(dummyeq, args, rets, vidxs, bidxs)

  # missing function body error
  @test_throws LoadError @eval @with_signature [legacy=false] function fn(eq::DummyEquation) end

  # @accepts/@returns must appear only at the start/end of the function body
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    println("this line is not allowed here")
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    println("this line is not allowed")
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    println("this line is not allowed here")
    @accepts u1
    @accepts u2
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    println("this line is not allowed")
    @returns ru2
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    ru1 = 3.0 * u1
    @returns ru2
  end

  # @accepts and @returns require each at least one argument
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts
    ru1 = 1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts
    @accepts u1
    ru1 = 1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
    @returns
  end

  # error on duplicated options for @accepts/@returns
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts [bdry,bdry] u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts [Prefix(lhs),bdry,Prefix(lhs)] u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation)
    @accepts [blabla] u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # function definition requires exactly one argument
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(#=no argument=#)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq::DummyEquation, too_much)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # the only argument must be type annotated
  @test_throws LoadError @eval @with_signature [legacy=false] function fn_w_error(eq)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end

  # fallback interface methods error when called
  function fn_with_no_signature(eq::DummyEquation)
    @accepts u1
    ru1 = 2.0 * u1
    @returns ru1
  end
  @test_throws ErrorException accepts(fn_with_no_signature,dummyeq,nothing)
  @test_throws ErrorException returns(fn_with_no_signature,dummyeq,nothing)
  # except the trait function is just false
  has_signature(fn_with_no_signature, dummyeq, nothing) == false

end
